/**
	$Id: DummyJobExecutor.java 4478 2008-09-05 15:15:54Z jari $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.clients.jobagent.executors;

import java.util.Collections;

import net.sf.basedb.clients.jobagent.Agent;
import net.sf.basedb.clients.jobagent.JobExecutor;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.JobAgentSettings;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.signal.SignalReceiver;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.Values;

/**
 	This is dummy job executor implementation which just sets the status
 	of each job to {@link net.sf.basedb.core.Job.Status#DONE} without actually
 	executing the job. This class is usful for debugging the job agent application.
 	<p>
 	<table>
 	<tr>
	 	<th>Configuration parameter</th>
	 	<th>Default value</th>
	 	<th>Description</th>
 	</tr>
 	
 	<tr>
 		<td>agent.executor.dummy.wait</td>
 		<td></td>
 		<td>
 			Number of seconds it takes to execute a job. The implementation simly
 			waits the specified number of seconds. If not specified the execution will
 			return immediately.
 		</td>
 	</tr>
 	</table>

	@author nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:15:54 +0200 (Fri, 05 Sep 2008) $
*/
public class DummyJobExecutor
	implements JobExecutor
{

	/**
		Log job agent events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.jobagent.executors.DummyJobExecutor");
	
	
	private int wait;
	
	public DummyJobExecutor()
	{}
	
	/*
		From the JobRunner interface
		-------------------------------------------
	*/
	public void init(Agent agent)
	{
		wait = Values.getInt(agent.getProperty("agent.executor.dummy.wait"), -1);
	}
	
	public void executeJob(SessionControl sc, Agent agent, Job job, JobAgentSettings settings,
		Job.ExecutionTime usedSlot)
	{
		log.info("Executing job: " + job);
		DbControl dc = null;
		boolean aborted = false;
		Throwable error = null;
		try
		{
			dc = sc.newDbControl();
			job = Job.getById(dc, job.getId());
			job.start("Not really, but used for testing the job agent", 
				agent.getServerName(), agent.getJobAgent(dc));
			if (wait > 0)
			{
				SignalReceiver signalReceiver = agent.getSignalReceiver();
				job.setSignalTransporter(signalReceiver.getSignalTransporterClass(), 
					signalReceiver.registerSignalHandler(new ThreadSignalHandler()));
				job.setProgress(50, "Halfway; waiting " + wait + " seconds");
				dc.commit();
				try
				{
					synchronized (Thread.currentThread())
					{
						Thread.currentThread().wait(wait * 1000);
					}
				}
				catch (InterruptedException ex)
				{
					aborted = true;
				}
				catch (Throwable t)
				{
					log.error(t.getMessage(), t);
					error = t;
				}
				dc  = sc.newDbControl();
				dc.reattachItem(job);
			}
			if (aborted)
			{
				job.doneError("Aborted by user");
			}
			else if (error != null)
			{
				job.doneError(error.getMessage(), Collections.singleton(error));
			}
			else
			{
				job.doneOk("Not really, but used for testing job agent");
			}
			dc.commit();
			log.info("Done executing: " + job);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	public void close()
	{}
	// -------------------------------------------
}
