/**
	$Id: ThreadJobExecutor.java 4478 2008-09-05 15:15:54Z jari $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.clients.jobagent.executors;

import java.util.Arrays;
import java.util.List;

import net.sf.basedb.clients.jobagent.Agent;
import net.sf.basedb.clients.jobagent.CmdLine;
import net.sf.basedb.clients.jobagent.JobExecutor;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.JobAgentSettings;
import net.sf.basedb.core.PluginExecutionRequest;
import net.sf.basedb.core.PluginResponse;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.signal.SignalReceiver;
import net.sf.basedb.core.signal.SocketSignalReceiver;
import net.sf.basedb.util.SocketUtil;
import net.sf.basedb.util.Values;

/**
	This is a job executor which executes jobs in the current thread.
	This has the disadvantage that a misbehaving job can
	interfere with other jobs and the job agent. For example, a
	call to {@link java.lang.System#exit(int)} would result in
	all other jobs as well as the job agent getting killed.
	<p>
	The advantage is that there is no overhead in time or memory.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:15:54 +0200 (Fri, 05 Sep 2008) $
*/
public class ThreadJobExecutor
	implements JobExecutor
{

	/**
		Log job agent events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.jobagent.executors.ThreadJobExecutor");
	
	private SignalReceiver signalReceiver;
	
	public ThreadJobExecutor()
	{}
	
	/*
		From the JobRunner interface
		-------------------------------------------
	*/
	public void init(Agent agent)
	{
		this.signalReceiver = agent == null ? null : agent.getSignalReceiver();
	}
	
	public void executeJob(SessionControl sc, Agent agent, Job job, JobAgentSettings settings,
		Job.ExecutionTime usedSlot)
	{
		log.info("Executing job: " + job);
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			
			// Load a fresh copy of the job item
			job = Job.getById(dc, job.getId());
			PluginExecutionRequest exec = null;
			
			try
			{
				exec = job.execute(null, Application.getHostName(), settings);
			}
			catch (Throwable t)
			{
				//	The job could not be started, set error status and send message to owner
				job.doneError(t.getMessage(), Arrays.asList(t));
				log.error("Error executing job: " + job, t);
			}
			exec.registerSignalReceiver(signalReceiver);
			dc.commit();

			PluginResponse response = exec.invoke();
			if (response.getStatus() == Response.Status.ERROR)
			{
				// There was an error, job status and message is already sent to the user
				log.error("Error executing job: " + job + "; " + response.getMessage());
				List<? extends Throwable> errors = response.getErrorList();
				if (errors != null)
				{
					for (Throwable t : errors)
					{
						log.error(t.getMessage(), t);
					}
				}
			}
			else
			{
				log.info("Done executing: " + job);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	public void close()
	{}
	// -------------------------------------------
	
	private void setSignalReceiver(SignalReceiver signalReceiver)
	{
		this.signalReceiver = signalReceiver;
	}
	
	public static void main(String[] args)
	{
		int exitCode = 0;
		CmdLine cmdLine = new CmdLine(args);
		int jobId = Values.getInt(cmdLine.getOption("-j"));
		int settingsId = Values.getInt(cmdLine.getOption("-s"));
		String login = cmdLine.getOption("-u");
		String password = cmdLine.getOption("-p");
		String proxyPort = cmdLine.getOption("-x");
		int threadPriority = Values.getInt(cmdLine.getOption("-t"), Thread.NORM_PRIORITY);

		String loginComment = "Job agent running on host " + 
			Application.getHostName() + " executing job with ID = " + jobId;
		
		log.info("Running job with ID: " + jobId);
		DbControl dc = null;
		SessionControl sc = null;
		SessionControl impersonated = null;
		Application.Pinger pinger = null;
		try
		{
			ThreadJobExecutor executor = new ThreadJobExecutor();
			
			if (proxyPort != null)
			{
				SocketSignalReceiver signalReceiver = new SocketSignalReceiver();
				signalReceiver.init("proxy=" + Application.getHostName() + ":" + proxyPort);
				executor.setSignalReceiver(signalReceiver);
				Runtime.getRuntime().addShutdownHook(new Thread(new ShutdownHook(signalReceiver)));
			}
			
			sc = Application.newSessionControl("net.sf.basedb.clients.jobagent", 
				SocketUtil.getLocalHost().toString(), null);
			sc.login(login, password, loginComment, false);
			dc = sc.newDbControl();
			Job job = Job.getById(dc, jobId);
			JobAgentSettings settings = JobAgentSettings.getById(dc, settingsId);
			dc.close();
			
			impersonated = sc.impersonateLogin(job, "Running job: " + job.getName());
			if (job.getActiveProjectId() != 0)
			{
				dc = impersonated.newDbControl();
				Project activeProject = Project.getById(dc, job.getActiveProjectId());
				impersonated.setActiveProject(activeProject);
				dc.close();
			}
			
			Thread.currentThread().setPriority(threadPriority);
			pinger = Application.newPinger(sc);
			executor.executeJob(impersonated, null, job, settings, null);
		}
		catch (Throwable t)
		{
			log.error("Error executing job with ID: " + jobId, t);
			exitCode = 1;
		}
		finally
		{
			if (pinger != null) pinger.stop();
			if (dc != null) dc.close();
			if (impersonated != null && impersonated.isLoggedIn()) impersonated.logout();
			if (sc != null && sc.isLoggedIn()) sc.logout();
		}
		System.exit(exitCode);
	}
	
	private static class ShutdownHook
		implements Runnable
	{
		private SignalReceiver signalReceiver;
		
		private ShutdownHook(SignalReceiver signalReceiver)
		{
			this.signalReceiver = signalReceiver;
		}
		public void run()
		{
			signalReceiver.close(5000);
		}
	}
}
