/*
	$Id: HybridizationTransfer.java 4478 2008-09-05 15:15:54Z jari $

	Copyright (C) 2005 Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Hakkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioMaterialEvent;
import net.sf.basedb.core.CommonItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.LabeledExtract;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
<dl>
<dt><b>Base 1 tables</b></dt>
<dd>Hybridization</dd>
<dt><b>Unused columns</b></dt>
<dd>addedDate, hybridizationDate</dd>
<dt><b>Notes</b></dt>
<dd>
<ul>
<li>Should hybridizationDate be an annotation?</li>
</ul>
</dd>
</dl>

@author Gregory
@version 2.0
@base.modified $Date: 2008-09-05 17:15:54 +0200 (Fri, 05 Sep 2008) $
*/

public class HybridizationTransfer extends CommonItemTransfer
{
	private ArraySlideTransfer arraySlideT = null;
	private LabeledExtractTransfer labeledExtractT = null;
	private ProtocolTransfer protocolT = null;
	
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.HybridizationTransfer");

	/**
	Constructs a new HybridizationTransfer.
	
	@param userT A finished {@link UserTransfer}
	@param groupT A finished {@link GroupTransfer}
	@param arraySlideT A finished {@link ArraySlideTransfer}
	 */
	HybridizationTransfer(UserTransfer userT, GroupTransfer groupT,
			ArraySlideTransfer arraySlideT, LabeledExtractTransfer labeledExtractT, ProtocolTransfer protocolT)
	{
		super(userT, groupT);
		log.debug("Checking that necessary transfers have been done before proceeding with Hybridization");
		if (arraySlideT.isDone() && labeledExtractT.isDone() && protocolT.isDone())
		{
			this.arraySlideT = arraySlideT;
			this.labeledExtractT = labeledExtractT;
			this.protocolT = protocolT;
		}
		else
		{
			if (!arraySlideT.isDone()) log.debug("Dependency: ArraySlides haven't been transfered");
			if (!labeledExtractT.isDone()) log.debug("Dependency: Labeled Extract haven't been transfered");
			if (!protocolT.isDone()) log.debug("Dependency: Protocols haven't been transfered");
			dependencyFailure();
		}
	}

	public void start()
	{
		log.info("Transferring Hybridizations...");
		int count = getInt("countHybridizations");
		initIdMap(count);
		startProgress(count, "Hybridizations");
		run(prepareStatementFromFile("selectHybridizations"), 0, getBatchSize());
		done();
		log.info("Transferring Hybridizations...DONE");
	}

	/**
	Creates a {@link Hybridization} from the current position in
	<code>ResultSet</code>.
	
	@param dc {@link DbControl} used to transfer this item
	@param rs <code>ResultSet</code> containing the data for this item.
	@return <code>Hybridization</code>
	@throws BaseException If a new hybridization cannot be created.
	@throws SQLException If current ResultSet position is invalid or number
	        of columns in the select statement is to low.
	 */
	protected Hybridization createItem(DbControl dc, ResultSet rs)
			throws BaseException, SQLException
	{		
		log.debug("transfer Hybridization:" + rs.getString(6) + " with id " + rs.getInt(1) + " from BASE1");
		Hybridization hybridization = Hybridization.getNew(dc);
		setCommonItemData(dc, (CommonItem) hybridization, rs);
		int slideId = rs.getInt(11);
		if (slideId != 0)
		{
			ArraySlide arraySlide = arraySlideT.getArraySlideById(dc, slideId);
			if (arraySlide != null)
			{
				hybridization.setArraySlide(arraySlide);
			}
			else
			{
				log.debug("ArraySlide with id: " + slideId + " couldn't be found");
			}
		}
		BioMaterialEvent creationEvent = hybridization.getCreationEvent();
		creationEvent.setEventDate(rs.getDate(9));
		creationEvent.setProtocol(protocolT.getProtocolById(dc, rs.getInt(10)));
		connectToLabeledExtracts(dc, hybridization, rs.getInt(ID));
		if (log.isDebugEnabled())
		{
			log.debug("Transfered Hybridization: " + hybridization.getName() + "\tBASE1 id(" + rs.getInt(1) + ")");
			log.debug("--------------------------------------------------------------------------");
		}
		return hybridization;
	}

	/**
	Connects labled extracts to a hybridization
	
	@param dc <code>DbControl</code>
	@param hybridization The hybridization
	@param base1HybridizationId The corresponding base 1 id for this hybridization
	 */
	private void connectToLabeledExtracts(DbControl dc, Hybridization hybridization,
			int base1HybridizationId)
	{
		ResultSet rs = null;
		try
		{
			PreparedStatement ps = prepareStatementFromFile("selectLabeledExtractsByHybridization");
			ps.setInt(1, base1HybridizationId);
			rs = ps.executeQuery();
			BioMaterialEvent creationEvent = hybridization.getCreationEvent();
			while (rs.next())
			{
				LabeledExtract le = labeledExtractT.getLabeledExtractById(dc, rs.getInt(1));
				if (le != null)
				{
					creationEvent.addSource(le, rs.getFloat(2));
					if (log.isDebugEnabled())
					{
						log.debug("Connected LabelExtract:" + le + " to Hybridization:" + hybridization);
					}
				}
				else
				{
					log.debug("No LabelExtract could be found with id: " + rs.getInt(1));
				}
			}
		}
		catch (Exception e)
		{
			log.error("FAILED connecting LabelExtract to Hybridization:" + hybridization, e);
			stop();
		}
	}
	
	
	/**
	Returns a Hybridization from the given base 1 slide id.
	
	@param dc {@link DbControl} used to get this hybridization.
	@param base1HybridizationId
	@return <code>Hybridization</code> or null if the Hybridization does
	        not exsist.
	 */
	public Hybridization getHybridizationById(DbControl dc,
			int base1HybridizationId)
	{
		Integer base2Id = getBase2Id(base1HybridizationId);
		return base2Id == null ? null : Hybridization.getById(dc, base2Id);
	}
}
