/*
	$Id: RoleTransfer.java 4478 2008-09-05 15:15:54Z jari $

	Copyright (C) 2005 Jari Hakkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Hakkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Role;
import net.sf.basedb.core.RoleKey;
import net.sf.basedb.core.SystemItems;

import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
This class provides methods to transfer Roles from base 1 to BASE 2.

@author Gregory
@version 2.0
@base.modified $Date: 2008-09-05 17:15:54 +0200 (Fri, 05 Sep 2008) $
*/

public class RoleTransfer extends Transfer
{
	// BASE 1 access masks
	private final int SUPERUSER = (1 << 0); // Almighty
	private final int SUPERREADER = (1 << 30); // Full read access
	private final int SHARE_WORLD = (1 << 25); // Share with 'world'
	private final int ADMIN_USER = (1 << 1); // Edit users
	private final int USER_READ = (1 << 2); // Read user list
	private final int NEWS_EDIT = (1 << 3); // Edit site news
	private final int USER_READ_MORE = (1 << 4); // Read more user info
	private final int ARRAY_READ = (1 << 6); // Read array and plate info
	private final int ARRAY_EDIT = (1 << 7); // Edit arrays and plates
	private final int WARE_EDIT = (1 << 8); // Edit hard/software
	private final int PROTOCOL_EDIT = (1 << 9); // Edit protocols
	private final int PROGRAM_EDIT = (1 << 10); // Edit plugins
	private final int BIO_READ = (1 << 11); // Read sample/extract/label
	private final int BIO_EDIT = (1 << 12); // Edit sample/extract/label
	private final int BIO_TISSUE = (1 << 13); // Edit sample origins
	private final int BIO_ANNOT = (1 << 14); // Edit sample annotation types
	private final int PROGRAM_READ = (1 << 15); // Read plugins
	private final int HYB_READ = (1 << 16); // Read hybridizations
	private final int HYB_EDIT = (1 << 17); // Edit hybridizations
	private final int UPLOAD = (1 << 18); // Upload files
	private final int WIZZZARD_EDIT = (1 << 19); // Edit file formats
	private final int EXPER_READ = (1 << 20); // Read experiments
	private final int EXPER_EDIT = (1 << 21); // Edit experiments
	private final int EXPER_CREATE = (1 << 22); // Create experiments
	private final int EXPER_MAGEML = (1 << 23); // Create MAGE-ML for
												// experiments
	private final int BROADCAST = (1 << 24); // Chat in left frame
	private final int REPORTER_READ = (1 << 26); // Read reporters
	private final int REPORTER_EDIT = (1 << 27); // Edit reporters

	// The access masks corresponding to BASE 2 predefined roles
	private final int GUEST = UPLOAD | REPORTER_READ;
	private final int USER = GUEST | BIO_EDIT | HYB_EDIT | EXPER_CREATE;
	private final int POWER_USER = USER | WIZZZARD_EDIT | REPORTER_EDIT | WARE_EDIT | BIO_ANNOT | ARRAY_EDIT | PROTOCOL_EDIT;
	private final int SUPERVISOR = SUPERREADER;
	private final int ADMINISTRATOR = SUPERUSER;

	// The meaningless access masks are not used in BASE 2
	private final int MEANINGLESS = BROADCAST | BIO_TISSUE | EXPER_EDIT | EXPER_MAGEML | USER_READ_MORE;

	// Access masks that corresponds to a DENIED permission in BASE 2 if missing
	private final int DENIED = PROGRAM_READ | ARRAY_READ | USER_READ | BIO_READ | HYB_READ | EXPER_READ;
	
	// Maps access mask to roles (for the new roles only)
	private Map<Integer, Role> maskMap = null;
	
	/**
	 	Logger
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.RoleTransfer");
	
	RoleTransfer()
	{}

	public void start()
	{
		log.info("Transferring roles...");
		// The number of roles created is static
		startProgress(21, "Roles");
		log.debug("Administrator has mask " + ADMINISTRATOR);
		log.debug("Supervisor has mask " + SUPERVISOR);		
		log.debug("Power user has mask " + POWER_USER);
		log.debug("User has mask " + USER);
		log.debug("Guest has mask " + GUEST);
		maskMap = new HashMap<Integer, Role>();
		DbControl dc = newDbControl();
		try
		{
			log.info("Create roles in BASE 2 that corresponds to BASE 1");
			// Permissions that are used for the new roles
			Set<Permission> create = EnumSet.of(Permission.CREATE);
			Set<Permission> admin = EnumSet.of(Permission.CREATE, Permission.READ, Permission.USE, Permission.WRITE, Permission.DELETE);
			Set<Permission> denied = EnumSet.of(Permission.DENIED);
			Set<Permission> use = EnumSet.of(Permission.READ, Permission.USE);

			String migrationInfo = "This role was created by the migration tool and corresponds to the BASE 1 ";
			createRole(dc, "Administer users",  migrationInfo + "\"Administer users\" permission.",
				ADMIN_USER,
				new ItemPermission[] 
				{
					new ItemPermission(Item.USER, admin),
					new ItemPermission(Item.GROUP, admin),
					new ItemPermission(Item.ROLE, admin),
					new ItemPermission(Item.QUOTA, admin),
					new ItemPermission(Item.QUOTATYPE, use),
					new ItemPermission(Item.DIRECTORY, create)
				}
			);
			
			createRole(dc, "Add plugins",  migrationInfo + "\"Add plug-ins\" permission.",
				PROGRAM_EDIT,
				new ItemPermission[] 
   				{
   					new ItemPermission(Item.PLUGINDEFINITION, create),
   					new ItemPermission(Item.PLUGINCONFIGURATION, create),
   					new ItemPermission(Item.PLUGINTYPE, admin)
   				}
   			);
			
			createRole(dc, "News administrator",  migrationInfo + "\"Add news\" permission.",
				NEWS_EDIT,
				new ItemPermission[]
				{
					new ItemPermission(Item.NEWS, admin)
				}
			);

			createRole(dc, "Add protocols",  migrationInfo + "\"Add protocols\" permission.",
				PROTOCOL_EDIT,
				new ItemPermission[]
				{
					new ItemPermission(Item.PROTOCOL, create),
					new ItemPermission(Item.PROTOCOLTYPE, use)
				}
			);

			createRole(dc, "Add arrays/plates",  migrationInfo + "\"Add arrays/plates\" permission.",
				ARRAY_EDIT,
				new ItemPermission[]
				{
					new ItemPermission(Item.PLATETYPE, create),
					new ItemPermission(Item.PLATEGEOMETRY, use),
					new ItemPermission(Item.PLATE, create),
					new ItemPermission(Item.PLATEMAPPING, use),
					new ItemPermission(Item.ARRAYDESIGN, create),
					new ItemPermission(Item.ARRAYBATCH, create),
					new ItemPermission(Item.ARRAYSLIDE, create),
					new ItemPermission(Item.PROTOCOLTYPE, use)
				}
			);
			
			createRole(dc, "Add annotation types",  migrationInfo + "\"Add annotation types\" permission.",
				BIO_ANNOT,
				new ItemPermission[]
				{
					new ItemPermission(Item.ANNOTATIONTYPE, create),
					new ItemPermission(Item.ANNOTATIONTYPECATEGORY, create)
				}
			);

			createRole(dc, "Add hardware/software",  migrationInfo + "\"Add hardware/software\" permission.",
				WARE_EDIT,
				new ItemPermission[]
				{
					new ItemPermission(Item.HARDWARE, create),
					new ItemPermission(Item.HARDWARETYPE, use),
					new ItemPermission(Item.SOFTWARE, create),
					new ItemPermission(Item.SOFTWARETYPE, use)
				}
			);
			
			createRole(dc, "Add reporters",  migrationInfo + "\"Add reporters\" permission.",
				REPORTER_EDIT,
				new ItemPermission[]
				{
					new ItemPermission(Item.REPORTER, admin),
					new ItemPermission(Item.REPORTERTYPE, admin),
					new ItemPermission(Item.REPORTERLIST, create)
				}
			);
			
			createRole(dc, "Add file formats",  migrationInfo + "\"Add file formats\" permission.",
				WIZZZARD_EDIT,
				new ItemPermission[]
				{
					new ItemPermission(Item.PLUGINCONFIGURATION, create)
				}
			);
			
			createRole(dc, "Denied plugins",  migrationInfo + "\"Read plugins\" permission.",
				PROGRAM_READ,
				new ItemPermission[]
				{
					new ItemPermission(Item.PLUGINCONFIGURATION, denied),
					new ItemPermission(Item.PLUGINDEFINITION, denied)
				}
			);
			
			createRole(dc, "Share with world",  migrationInfo + "\"Share with world\" permission.",
				SHARE_WORLD,
				new ItemPermission[]
				{
					new ItemPermission(Item.SYSTEM, EnumSet.of(Permission.SHARE_TO_EVERYONE))
				}
			);
			
			createRole(dc, "Add biomaterials",  migrationInfo + "\"Add biomaterials\" permission.",
				BIO_EDIT,
				new ItemPermission[]
				{
					new ItemPermission(Item.BIOSOURCE, create),
					new ItemPermission(Item.SAMPLE, create),
					new ItemPermission(Item.EXTRACT, create),
					new ItemPermission(Item.LABELEDEXTRACT, create),
					new ItemPermission(Item.LABEL, create)
				}
			);

			createRole(dc, "Add hybridizations",  migrationInfo + "\"Add hybridizations\" permission.",
				HYB_EDIT,
				new ItemPermission[]
				{
					new ItemPermission(Item.HYBRIDIZATION, create),
					new ItemPermission(Item.SCAN, create),
					new ItemPermission(Item.RAWBIOASSAY, create)
				}
			);
			
			createRole(dc, "Create experiments",  migrationInfo + "\"Create experiments\" permission.",
				EXPER_CREATE,
				new ItemPermission[]
				{
					new ItemPermission(Item.EXPERIMENT, create)
				}
			);

			createRole(dc, "Upload files",  migrationInfo + "\"Upload files\" permission.",
				UPLOAD,
				new ItemPermission[]
				{
					new ItemPermission(Item.FILE, create),
					new ItemPermission(Item.FILETYPE, use),
					new ItemPermission(Item.DIRECTORY, create)
				}
			);
			
			createRole(dc, "Denied arrays/plates",  migrationInfo + "\"Read arrays/plates\" permission.",
				ARRAY_READ,
				new ItemPermission[]
				{
					new ItemPermission(Item.PLATETYPE, denied),
					new ItemPermission(Item.PLATEGEOMETRY, denied),
					new ItemPermission(Item.PLATEMAPPING, denied),
					new ItemPermission(Item.PLATE, denied),
					new ItemPermission(Item.ARRAYDESIGN, denied),
					new ItemPermission(Item.ARRAYBATCH, denied),
					new ItemPermission(Item.ARRAYSLIDE, denied)
				}
			);
			
			createRole(dc, "Denied users",  migrationInfo + "\"Read user info\" permission.",
				USER_READ,
				new ItemPermission[]
				{
					new ItemPermission(Item.USER, denied)
				}
			);
			
			createRole(dc, "Denied biomaterials",  migrationInfo + "\"Read biomaterials\" permission.",
				BIO_READ,
				new ItemPermission[]
				{
					new ItemPermission(Item.BIOSOURCE, denied),
					new ItemPermission(Item.SAMPLE, denied),
					new ItemPermission(Item.EXTRACT, denied),
					new ItemPermission(Item.LABELEDEXTRACT, denied),
					new ItemPermission(Item.LABEL, denied)
				}
			);
			
			createRole(dc, "Denied hybridizations",  migrationInfo + "\"Read hybridizations\" permission.",
				HYB_READ,
				new ItemPermission[]
				{
					new ItemPermission(Item.HYBRIDIZATION, denied),
					new ItemPermission(Item.SCAN, denied),
					new ItemPermission(Item.RAWBIOASSAY, denied)
				}
			);

			createRole(dc, "Denied experiments",  migrationInfo + "\"Read experiments\" permission.",
				EXPER_READ,
				new ItemPermission[]
				{
					new ItemPermission(Item.EXPERIMENT, denied)
				}
			);
			
			createRole(dc, "Read reporters",  migrationInfo + "\"Read reporters\" permission.",
				REPORTER_READ,
				new ItemPermission[]
				{
					new ItemPermission(Item.REPORTER, use),
					new ItemPermission(Item.REPORTERTYPE, use)
				}
			);
			dc.commit();
			progress.stop();
		}
		catch (Exception e)
		{
			log.error("FAILED  when creating roles", e);
			stop();
		}
		done();
		log.info("Transferring roles...DONE");
	}

	/**
	Returns a list of roles matching a base 1 accessMask
	
	@param accessMask A base 1 accessMask
	@return A list of roles matching a given base 1 accessMask
	 */
	public Set<Integer> getRoleIds(Integer accessMask)
	{
		int mask = accessMask.intValue();		
		
		Set<Integer> roleIds = new HashSet<Integer>();

		// Check if BASE1 superuser
		if ((mask & ADMINISTRATOR) == ADMINISTRATOR)
		{
			roleIds.add(SystemItems.getId(Role.ADMINISTRATOR));
			return roleIds;
		}

		// Remove BASE 1 permission that have no meaining in BASE 2
		mask &= ~MEANINGLESS;
		
		// All "Read Item X" permissions are converted to DENIED permissions
		mask ^= DENIED;
		
		if ((mask & SUPERVISOR) == SUPERVISOR)
		{			
			roleIds.add(SystemItems.getId(Role.SUPERVISOR));
			mask &= ~SUPERVISOR;
			mask &= ~DENIED;
		}
		
		// We use either POWER_USER, USER or GUEST as the basis for 
		if ((mask & POWER_USER) == POWER_USER)
		{			
			roleIds.add(SystemItems.getId(Role.POWER_USER));
			mask &= ~POWER_USER;
		}
		else if ((mask & USER) == USER)
		{			
			roleIds.add(SystemItems.getId(Role.USER));
			mask &= ~USER;
		}
		else if ((mask & GUEST) == GUEST)
		{
			roleIds.add(SystemItems.getId(Role.GUEST));
			mask &= ~GUEST;
		}
		
		if (mask != 0)
		{
			// The remaining bits in the access mask didn't fit the predefined roles
			// Use the newly created roles to map the last bits
			for (Map.Entry<Integer, Role> entry : maskMap.entrySet())
			{
				int roleMask = entry.getKey();
				if ((mask & roleMask) == roleMask)
				{
					roleIds.add(entry.getValue().getId());
					mask &= ~roleMask;
				}
			}
		}
		
		// If there are more bits in the mask it is a bug
		// We must either map or ignore (see MEANINGLESS) all bits
		if (mask != 0)
		{
			BaseException be = new BaseException("Access mask could not be completely converted to roles: " + 
				"original mask=" + accessMask + "; end mask="+mask+"; assigned roles=" + roleIds.toString());
			log.error("FAILED when converting access mask to roles", be);
			throw be;
		}
		return roleIds;
	}

	private Role createRole(DbControl dc, String name, String description, int base1Mask, ItemPermission[] permissions)
	{		
		Role role = Role.getNew(dc);
		role.setName("BASE 1.2: " + name);
		role.setDescription(description);
		for (ItemPermission ip : permissions)
		{
			RoleKey key = RoleKey.getByItemType(dc, ip.getItem());
			key.setPermissions(role, ip.getPermissions());
		}
		maskMap.put(base1Mask, role);
		dc.saveItem(role);
		log.debug("Created  \"" + name + "\" role");
		return role;
	}
	
	static class ItemPermission
	{
		private final Item item;
		private final Set<Permission> permissions;
		
		ItemPermission(Item item, Set<Permission> permissions)
		{
			this.item = item;
			this.permissions = permissions;
		}
		
		public Item getItem()
		{
			return item;
		}
		
		public Set<Permission> getPermissions()
		{
			return permissions;
		}
		
	}
	
}
