/**
	$Id: AbstractJspFactory.java 4478 2008-09-05 15:15:54Z jari $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.clients.web.extensions;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.util.extensions.xml.PathSetter;
import net.sf.basedb.util.extensions.xml.VariableSetter;

/**
	An abstract BASE class for action and renderer factories. This class
	provides support for adding scripts and stylesheets to the rendered
	JSP page.
	<p>
	Use the {@link #setScript(String)} and {@link #setStylesheet(String)}
	methods to add references to script and stylesheet files to the factory. 

	<p>
	Note! The naming of these methods may be a bit misleading, since each 
	call ADD the argument to a set. To remove items from the sets use the 
	{@link #getScripts()} and {@link #getStylesheets()} methods.

	<p>
	Scripts and stylesheets that has been added to the factory will 
	be propagated to the {@link JspContext#addScript(String)}
	and {@link JspContext#addStylesheet(String)} methods in
	the {@link #prepareContext(JspContext)}.

	@author nicklas
	@version 2.7
	@base.modified $Date: 2008-09-05 17:15:54 +0200 (Fri, 05 Sep 2008) $
*/
public abstract class AbstractJspFactory
{

	private Set<String> scripts;
	private Set<String> stylesheets;
	
	/**
		Create a new factory instance.
	*/
	public AbstractJspFactory()
	{}
	
	/**
		Prepare the JSP context by adding all registered script
		and stylesheet references to it.
		@param jspContext The context to add the scripts to
	*/
	protected void prepareContext(JspContext jspContext)
	{
		if (getScripts() != null) 
		{
			for (String script : getScripts())
			{
				jspContext.addScript(script);
			}
		}
		if (getStylesheets() != null) 
		{
			for (String stylesheet : getStylesheets())
			{
				jspContext.addStylesheet(stylesheet);
			}
		}
	}
	
	/**
		Add a reference to a javascript file to the factory.
		@param script The script to add
	*/
	@VariableSetter
	@PathSetter
	public void setScript(String script)
	{
		if (scripts == null) scripts = new HashSet<String>();
		scripts.add(script);
	}
	
	/**
		Add a reference to a CSS stylesheet file to the factory.
		@param stylesheet The stylesheet to add
	*/
	@VariableSetter
	@PathSetter
	public void setStylesheet(String stylesheet)
	{
		if (stylesheets == null) stylesheets = new HashSet<String>();
		stylesheets.add(stylesheet);
	}
	
	/**
		Get a set with all script references.
		@return A set or null if no scripts has been added
	*/
	public Set<String> getScripts()
	{
		return scripts;
	}
	
	/**
		Get a set with all stylesheet references.
		@return A set or null if no stylesheets has been added
	*/
	public Set<String> getStylesheets()
	{
		return stylesheets;
	}
}
