/**
	$Id: ScanResults.java 4478 2008-09-05 15:15:54Z jari $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.clients.web.extensions;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

/**
	Holds information about the results of a scan for new/deleted/updated
	extensions.

	@author nicklas
	@version 2.7
	@base.modified $Date: 2008-09-05 17:15:54 +0200 (Fri, 05 Sep 2008) $
	@see ExtensionsControl#installAndUpdateExtensions(boolean)
*/
public class ScanResults
{

	private final boolean manualScan;
	private final boolean forceUpdate;
	private final long startTime;

	private Map<ExtensionsFile, FileResults> fileResults;
	private boolean hasError;
	private int numErrorFiles;
	private long endTime;
	private String summary;
	
	/**
		Create a new scan results object.
		@param manualScan TRUE if this is a manual scan, 
			FALSE if it is an automatic scan
		@param forceUpdate TRUE if the scan is forcing update of already
			installed extensions, FALSE otherwise
	*/
	public ScanResults(boolean manualScan, boolean forceUpdate)
	{
		this.manualScan = manualScan;
		this.forceUpdate = forceUpdate;
		this.fileResults = new TreeMap<ExtensionsFile, FileResults>();
		this.hasError = false;
		this.numErrorFiles = 0;
		this.startTime = System.currentTimeMillis();
	}
	
	/**
		Set the end time of the scan
	*/
	void setEnded()
	{
		this.endTime = System.currentTimeMillis();
	}
	
	/**
		Get the start time of the scan as a millisecond value.
		@see System#currentTimeMillis()
	*/
	public long getStartTime()
	{
		return startTime;
	}
	
	/**
		Get the end time of the scan as a millisecond value.
		@see System#currentTimeMillis()
	*/
	public long getEndTime()
	{
		return endTime;
	}
	
	/**
		Get a summary of the scan results as a textual description.
	*/
	public String getSummary()
	{
		return summary;
	}
	
	void setSummary(String summary)
	{
		this.summary = summary;
	}
	
	/**
		@return TRUE if the scan was manual, FALSE if it was automatic
	*/
	public boolean wasManual()
	{
		return manualScan;
	}
	
	/**
		@return TRUE if the scan forced update of already installed extensions,
			FALSE otherwise
	*/
	public boolean wasForceUpdate()
	{
		return forceUpdate;
	}
	
	/**
		Check if the scan was successful or not. For more detailed
		information use {@link #getResults(ExtensionsFile)}.
		@return TRUE if there was any problems, FALSE if everything was 
			successful
	*/
	public boolean hasError()
	{
		return hasError;
	}	
	
	/**
		The number of extension files that could not be completely loaded
		because of an error.
	*/
	public int getNumErrorFiles()
	{
		return numErrorFiles;
	}
	
	/**
		Get all extension files that was handled in the last scan. The
		collection includes both new, deleted and updated extension files.
		@return A collections with {@link ExtensionsFile}:s
	*/
	public Collection<ExtensionsFile> getFiles()
	{
		return fileResults.keySet();
	}
	
	/**
		Get the results for a specific extensions file.
		@param extFile The extensions file to get the results for
		@return A {@link FileResults} object or null if the
			extensions file is not found
	*/
	public FileResults getResults(ExtensionsFile extFile)
	{
		return fileResults.get(extFile);
	}
	
	/**
		Sets the short status message for an entire file. Example:
		"Installed", "Failed", etc.
		@param extFile The file to set the status on
		@param status The status message
		@see #addMessage(ExtensionsFile, String)
		@see #addErrorMessage(ExtensionsFile, String)
	*/
	void setStatus(ExtensionsFile extFile, String status)
	{
		FileResults m = fileResults.get(extFile);
		if (m == null) 
		{
			m = new FileResults();
			fileResults.put(extFile, m);
		}
		m.status = status;
	}
	
	/**
		Adds a detail message indicating a successful operation 
		to a file.
		@param extFile The file to add the message to
		@param message The message
		@see #addErrorMessage(ExtensionsFile, String)
	*/
	void addMessage(ExtensionsFile extFile, String message)
	{
		addMessage(extFile, message, false);
	}
	
	/**
		Adds a detail message inidcating a failed operation 
		to a file.
		@param extFile The file to add the message to
		@param message The message
	*/
	void addErrorMessage(ExtensionsFile extFile, String message)
	{
		addMessage(extFile, message, true);
	}
	
	
	private void addMessage(ExtensionsFile extFile, String message, boolean error)
	{
		FileResults m = fileResults.get(extFile);
		if (m == null) 
		{
			m = new FileResults();
			fileResults.put(extFile, m);
		}
		m.messages.add(message);
		if (error) 
		{
			if (!m.hasError) numErrorFiles++;
			hasError = true;
			m.hasError = true;
			m.status = "Error";
		}
	}
	

	/**
		Scan results for a single XML or JAR file.
	
	*/
	public static class FileResults
	{
		List<String> messages = new LinkedList<String>();
		boolean hasError = false;
		String status = null;
		
		/**
			If there was an error in this file or not.
			@return TRUE if there was an error, FALSE if successful
		*/
		public boolean hasError()
		{
			return hasError;
		}
		
		/**
			A short status message indicating the completed
			results on this file.
			@see #getMessages()
		*/
		public String getStatus()
		{
			return status;
		}
		
		/**
			A list of detailed messages, indicating the results of
			individual operation on this file.
		*/
		public List<String> getMessages()
		{
			return messages;
		}
		
	}
	
}
