/*
	$Id: FileUploadProgress.java 4478 2008-09-05 15:15:54Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Johan Enell

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.clients.web.fileupload;

/**
	Objects of this class holds information about the progress
	of an upload. Use the {@link FileUpload#getProgress()} to get 
	the progress that is associated with a particular file upload.
	<p>
	Use the {@link #getTotalBytes()} and
	{@link #getTransferredBytes()} to get information
	about how the upload is progressing.

	@author Nicklas
	@version 2.0
	@see FileUpload
	@base.modified $Date$
*/
public class FileUploadProgress
{
	private final long totalBytes;
	private final long startTime;
	private long transferredBytes;
	private String currentFilename;
	private Throwable exception;
	private boolean abort = false;

	/**
		Create a new <code>FileUploadProgress</code> object.
		This constructor is package private. Use {@link
		FileUpload#getProgress()} to get the progress that is
		associated with a particular file upload.

		@param totalBytes The total number of bytes that should be uploaded.
	*/
	FileUploadProgress(long totalBytes)
	{
		this.totalBytes = totalBytes;
		this.startTime = System.currentTimeMillis();
		this.transferredBytes = 0;
	}
	
	/**
		Get the total number of bytes expected to upload.
		@return The total number of bytes to upload
	*/
	public long getTotalBytes()
	{
		return totalBytes;
	}

	/**
		Get the number of bytes that have been uploaded so far.
		@return The number of bytes
	*/
	public long getTransferredBytes()
	{
		return transferredBytes;
	}
	/**
		Sets the number of transferred bytes.
		@param bytes The number of bytes
	*/
	void setTransferredBytes(long bytes)
	{
		transferredBytes = bytes;
	}
	/**
		Adds a value to the number of transfered bytes.
		@param bytes The value to add to the number of transfered bytes
	*/
	void addTransferredBytes(long bytes)
	{
		transferredBytes += bytes;
	}
	
	/**
		Checks if the transfer is complete or not. A transfer is complete
		when the transferred bytes equals the total bytes.
		@return TRUE if the transfer has been completed, FALSE otherwise
	*/
	public boolean transferIsComplete()
	{
		return (getTotalBytes() == getTransferredBytes());
	}

	/**
		Get the name of the file that is currently beeing transferred.
		The value returned is the name of the file on the client computer.
		@return The name of the file
	*/
	public String getCurrentFilename()
	{
		return currentFilename;
	}
	/**
		Sets the name of the file that is currently uploading
		@param filename The name of the file
	*/
	void setCurrentFilename(String filename)
	{
		currentFilename = filename;
	}

	/**
		Get the exception that caused the upload to fail.
		@return A <code>Throwable</code> or null if no exception has occured
	*/
	public Throwable getException()
	{
		return exception;
	}
	/**
		Set the exception that caused the upload to fail.
	*/
	void setException(Throwable exception)
	{
		this.exception = exception;	
	}

	/**
		Get the number of milliseconds since the start of the upload.
	*/
	public long getMillisecondsSinceStart()
	{
		return System.currentTimeMillis() - startTime;
	}

	/**
		Request that the upload should be aborted. The associated {@link
		FileUpload} object will check this property at regular intervals and if set
		it will close the upload stream as soon as possible, and then throw an 
		{@link UploadAbortedException} which will also be available from the
		{@link #getException()} method.
	*/
	public void setAbort()
	{
		this.abort = true;
	}

	/**
		Check if the upload has been requested to be aborted. Use the
		{@link #getException()} method to actually check if the upload
		has been aborted.
	*/
	public boolean hasAborted()
	{
		return abort;
	}
	
}
