/**
	$Id: PlainTextTemplate.java 4478 2008-09-05 15:15:54Z jari $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.clients.web.plugins;

import java.io.IOException;
import java.io.Writer;
import java.util.List;
import java.util.regex.Pattern;

import net.sf.basedb.core.Item;
import net.sf.basedb.util.Values;

/**
 	A plain text template that exports data in tab-delimited format.
 	 
	@author nicklas
	@version 2.2
	@base.modified $Date: 2008-09-05 17:15:54 +0200 (Fri, 05 Sep 2008) $
*/
public class PlainTextTemplate
	implements ExportTemplate
{

	private final Writer exportStream;
	private final List<ExportedProperty> properties;
	private int colNum;
	
	/**
		Create a new template.
		@param exportStream The stream to write to
		@param properties All exported properties
		@param itemType The type of items we are exporting
	*/
	public PlainTextTemplate(Writer exportStream, List<ExportedProperty> properties, Item itemType)
	{
		this.exportStream = exportStream;
		this.properties = properties;
	}
	
	/**
		@return text/plain
	*/
	public String getMimeType()
	{
		return "text/plain";
	}

	/**
		Do nothing.
	*/
	public void beginDocument() 
		throws IOException
	{}

	/**
		Write the headers on a single line. 
	*/
	public void writeHeaders() 
		throws IOException
	{
		for (ExportedProperty ep : properties)
		{
			exportStream.write(escape(ep.title));
			exportStream.write("\t");
		}
		exportStream.write("\n");
	}

	/**
		Reset column number.
	*/
	public void beginItem(int id) 
		throws IOException
	{
		colNum = 0;
	}

	/**
		Join the annotation values into a single string with a comma.
	*/
	@SuppressWarnings("unchecked")
	public void writeAnnotations(ExportedProperty ep, List<?> values)
		throws IOException
	{
		if (colNum > 0) exportStream.write("\t");
		colNum++;
		exportStream.write(escape(Values.getString(values, ",", true, ep.formatter)));
	}

	/**
		Write the data object.
	*/
	@SuppressWarnings("unchecked")
	public void writeProperty(ExportedProperty ep, Object data)
		throws IOException
	{
		if (colNum > 0) exportStream.write("\t");
		colNum++;
		exportStream.write(escape(ep.formatter.format(data)));
	}

	/**
		Write a newline.
	*/
	public void endItem() 
		throws IOException
	{
		exportStream.write("\n");
	}

	/**
		Do nothing.
	*/
	public void endDocument() 
		throws IOException
	{}

	
	private static final Pattern UNSAFE = Pattern.compile("(\\r?\\n\\r?|\\t)");
	/**
		Replace newlines and tabs with a space. Null values are converted to empty strings.
	*/
	private String escape(String data)
	{
		if (data == null) 
		{
			return "";
		}
		else
		{
			return UNSAFE.matcher(data).replaceAll(" ");
		}
	}
}
