/*
	$Id: Page.java 4483 2008-09-07 21:59:52Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006, 2008 Johan Enell, Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.clients.web.taglib;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.clients.web.Base;
import net.sf.basedb.util.Values;

import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.TagSupport;
import javax.servlet.http.HttpServletRequest;

/**
	This tag should be used together with the {@link Head &lt;base:head&gt;} and
	{@link Body &lt;base:body&gt;} tags on every JSP page in BASE.
	<p>
	<b>Example:</b>
	<pre>
&lt;%@ taglib prefix="base" uri="/WEB-INF/base.tld" %&gt;
&lt;base:page type="default" title="Welcome to BASE"&gt;
&lt;base:head&gt;
   &lt;!-- scripts and other head things... --&gt;
&lt;/base:head&gt;
&lt;base:body&gt;
   Page contents go here....
&lt;/base:body&gt;
&lt;/base:page&gt;
</pre>

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:page
      type=default|popup|frameset|include
		title=...
		menu=...
   &gt;
</pre>
	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>type</td>
		<td>default</td>
		<td>no</td>
		<td>The type of the page. Three values are defined:
			<ul>
			<li>default: The default if no other value is specified.
			<li>popup: Use this value for a popup dialogue.
			<li>frameset: Use this for pages containing a frameset.
			<li>include: Use this for pages that are included as part of another page
				(skips generation of &lt;html&gt;, &lt;body&gt;, etc. tags).
			</ul>
		</td>
	</tr>
	<tr>
		<td>title</td>
		<td>-</td>
		<td>yes</td>
		<td>The title of the page.</td>
	</tr>
	<tr>
		<td>menu</td>
		<td>standard</td>
		<td>no</td>
		<td>
			The name of the menu to use on the page. This value is only used
			for pages with type=default. See the <code>/include/menu.jsp</code>
			file for defined menu names.
		</td>
	</tr>
	</table>
	
	@author Nicklas
	@version 2.0
*/
public class Page
	extends TagSupport
{

	/**
		
	*/
	private static final long serialVersionUID = -5693587214280114326L;

	/**
		This is the default page type.
	*/
	static final int PAGE_TYPE_DEFAULT = 0;
	
	/**
		The page is a popup dialouge.
	*/
	static final int PAGE_TYPE_POPUP = 1;
	
	/**
		The page is a frameset.
	*/
	static final int PAGE_TYPE_FRAMESET = 2;
	
	/**
		This page is an included page.
	*/
	static final int PAGE_TYPE_INCLUDE = 3;

	/**
		The title of the page.
	*/
	private String title = null;
	
	/**
		The name of the menu that should be displayed on the current page.
		This setting is only used if the type is PAGE_TYPE_DEFAULT
	*/
	private String menu = "auto";

	/**
		The type of the current page.
	*/
	private int type = PAGE_TYPE_DEFAULT;

	/**
		The name of this webserver.
	*/
	private static String SERVER_NAME = null;

	/**
		The path to the root directory of this webserver.
	*/
	private static String ROOT = null;

	/**
		The version of BASE.
	*/
	private static String BASE_VERSION = null;
	
	private static String MAX_URL_LENGTH = null;

	/**
		The SessionControl object in use for this request.
	*/
	private SessionControl sc;


	/*
		Taglib initialization methods
	*/
	public void setMenu(String menu)
	{
		this.menu = menu;
	}
	public String getMenu()
	{
		return menu;
	}
	
	public void setTitle(String title)
	{
		this.title = title;
	}
	public String getTitle()
	{
		return title;
	}

	public void setType(String type)
	{
		if ("popup".equalsIgnoreCase(type))
		{
			this.type = PAGE_TYPE_POPUP;
		}
		else if ("frameset".equalsIgnoreCase(type))
		{
			this.type = PAGE_TYPE_FRAMESET;
		}
		else if ("include".equalsIgnoreCase(type))
		{
			this.type = PAGE_TYPE_INCLUDE;
		}
		else
		{
			this.type = PAGE_TYPE_DEFAULT;
		}
	}
	public int getTypeCode()
	{
		return type;
	}
	
	/*
		Server methods
	*/
	public String getRoot()
	{
		return ROOT;
	}
	/**
		@since 2.8
	*/
	public String getMaxUrlLength()
	{
		return MAX_URL_LENGTH;
	}
	public String getServerName()
	{
		if (SERVER_NAME == null) SERVER_NAME = Application.getHostName();
		return SERVER_NAME;
	}
	public String getBaseVersion()
	{
		if (BASE_VERSION == null) 
		{
			BASE_VERSION = "BASE " + Application.getMajorVersion() + "." + 
				Application.getMinorVersion() + "." + 
				Application.getMaintenanceVersion();
		}
		return BASE_VERSION;
	}
	public SessionControl getSessionControl()
	{
		return sc;
	}

	
	public void setPageContext(PageContext pageContext)
	{
		super.setPageContext(pageContext);
		if (ROOT == null) 
		{
			ROOT = ((HttpServletRequest)pageContext.getRequest()).getContextPath()+"/";
			int maxUrlLength = Values.getInt(pageContext.getServletContext().getInitParameter("max-url-length"));
			MAX_URL_LENGTH = maxUrlLength > 0 ? String.valueOf(maxUrlLength) : null;
		}
		try
		{
			sc = Base.getSessionControl(pageContext, false);
		}
		catch (Exception ex)
		{}
	}

	public int doStartTag()
		throws JspException
	{
		if (type != PAGE_TYPE_INCLUDE)
		{
			StringBuilder sb = new StringBuilder();
			sb.append("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">\n");
			sb.append("<html>\n");
			try
			{
				pageContext.getOut().print(sb.toString());
			}
			catch (Exception ex)
			{
				throw new JspTagException(ex.getMessage());
			}
		}
		return EVAL_BODY_INCLUDE;
	}

	public int doEndTag() 
		throws JspException
	{
		if (type != PAGE_TYPE_INCLUDE)
		{
			StringBuilder sb = new StringBuilder();
			sb.append("</html>\n");
			try
			{
				pageContext.getOut().print(sb.toString());
			}
			catch (Exception ex)
			{
				throw new JspTagException(ex.getMessage());
			}
		}
		return EVAL_PAGE;
	}
}



