/*
	$Id: PresetSelector.java 4478 2008-09-05 15:15:54Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Johan Enell

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.clients.web.taglib.table;

import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.data.RemovableData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.data.OwnableData;
import net.sf.basedb.core.data.SessionData;
import net.sf.basedb.core.data.ProjectData;
import net.sf.basedb.util.Enumeration;

import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.util.Values;

import java.util.List;
import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.TagSupport;

/**

	Special tag for generating a list of saved contexts and other settings.
	It uses the SessionControl and Item from the table to get the current
	context and a list of available contexts.

	@author Nicklas
	@version 2.0
*/
public class PresetSelector
	extends TagSupport
{

	/**
		
	*/
	private static final long serialVersionUID = 6787200020881710014L;

	/**
		The parent &lt;tbl:table&gt; tag.
	*/
	private Table table;
	
	/**
		Optional <code>class</code> attribute
	*/
	private String clazz = "header";
	
	/**
		Optional <code>colspan</code> attribute
	*/
	private String colspan = null;
	
	/**
		Optional <code>style</code> attribute
	*/
	private String style = null;
	
	/**
		If the preset selector should be visible or not.
	*/
	private boolean visible = true;
	
	/**
		If the 'removed' option should be included in the preset selector or not.
		If the list contains items that are not removable, the option is never included.
	*/
	private boolean disableRemoved = false;
	
	/**
		If the 'owned by me' option should be included in the preset selector or not.
		If the list contains items that are not ownable, the option is never included.
	*/
	private boolean disableOwnedBy = false;

	/**
		If the 'shared to me' option should be included in the preset selector or not.
		If the list contains items that are not shareable, the option is never included.
	*/
	private boolean disableSharedTo = false;

	/**
		If the 'in project' option should be included in the preset selector or not.
		If the list contains items that are not removable, the option is never included.
	*/
	private boolean disableInProject = false;

	/**
		If the 'owned by others' option should be included in the preset selector or not.
		If the list contains items that are not ownable, the option is never included.
	*/
	private boolean disableOthers = false;	
	
	private String onChange = "presetOnChange()";
	
	/*
		Taglib initialization methods
	*/
	public void setClazz(String clazz)
	{
		this.clazz = clazz;
	}
	public String getClazz()
	{
		return clazz;
	}

	public void setStyle(String style)
	{
		this.style = style;
	}
	public String getStyle()
	{
		return style;
	}
	
	public void setColspan(String colspan)
	{
		this.colspan = colspan;
	}
	public String getColspan()
	{
		return colspan;
	}

	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}

	public void setOnchange(String onChange)
	{
		this.onChange = onChange;
	}
	public String getOnchange()
	{
		return onChange;
	}

	public void setDisableremoved(boolean disable)
	{
		this.disableRemoved = disable;
	}
	public boolean getDisableremoved()
	{
		return disableRemoved;
	}
	
	public void setDisableownedby(boolean disable)
	{
		this.disableOwnedBy = disable;
	}
	public boolean getDisableownedby()
	{
		return disableOwnedBy;
	}
	
	public void setDisablesharedto(boolean disable)
	{
		this.disableSharedTo = disable;
	}
	public boolean getDisablesharedto()
	{
		return disableSharedTo;
	}
	
	public void setDisableinproject(boolean disable)
	{
		this.disableInProject = disable;
	}
	public boolean getDisableinproject()
	{
		return disableInProject;
	}
	
	public void setDisableothers(boolean disable)
	{
		this.disableOthers = disable;
	}
	public boolean getDisableothers()
	{
		return disableOthers;
	}
	
	/*
		Display methods
	*/
	public int doStartTag()
		throws JspException
	{
		table = (Table)findAncestorWithClass(this, Table.class);
		if (!isVisible()) return SKIP_BODY;
		
		table.addHeaderColumn(Values.getInt(getColspan(), 1));
		
		SessionControl sc = table.getSc();
		Item itemType = table.getItem();
		String subContext = table.getSubcontext();
		
		ItemContext cc = sc.getCurrentContext(itemType, subContext);
		StringBuilder sb = new StringBuilder();
		sb.append("<th");
		if (getClazz() != null) sb.append(" class=\"").append(getClazz()).append("\"");
		if (getStyle() != null) sb.append(" style=\"").append(getStyle()).append("\"");
		if (getColspan() != null) sb.append(" colspan=\"").append(getColspan()).append("\"");
		sb.append(">");
		
		sb.append("<table border=\"0\" cellspaing=\"0\" cellpadding=\"0\"><tr><td>");
		sb.append("<select name=\"presetselector\" onchange=\"").append(getOnchange()).append("\">\n");
		sb.append("<option value=\"\">- view / presets -\n");
		
		// Saved context names
		Enumeration<Integer, String> contextNames = sc.getContextNames(itemType, subContext);
		if (contextNames.size() > 0)
		{
			sb.append("<option value=\"\">------------------\n");
		}
		for (int i = 0;  i < contextNames.size(); ++i)
		{
			String name = contextNames.getValue(i);
			if (!ItemContext.DEFAULT_NAME.equals(name))
			{
				String fullName = HTML.encodeTags(name);
				String shortName = HTML.encodeTags(Values.trimString(name, 20));
				sb.append("<option value=\"load:").append(contextNames.getKey(i)).append("\"");
				sb.append(" title=\"Load preset: ").append(fullName).append("\">").append(shortName).append("\n");
			}
		}
		
		// Column presets
		sb.append("<option value=\"\">------------------\n");
		sb.append("<option value=\"option:columns:all\" title=\"Show all columns\">All columns\n");
		sb.append("<option value=\"option:columns:\" title=\"Show only required columns\">Required columns\n");
		String defaultColumns = (String)cc.getObject("defaultColumns");
		if (defaultColumns != null)
		{
			sb.append("<option value=\"option:columns:").append(defaultColumns).append("\" title=\"Show only default columns\">Default columns\n");
		}
		sb.append("<option value=\"managecolumns:\" title=\"Show, hide and re-order columns\">Other...\n");

		// Include options
		sb.append("<option value=\"\">------------------\n");
		int numOptions = 0;
		Class dataClass = itemType.getDataClass();
		if (dataClass != null)
		{
			if (RemovableData.class.isAssignableFrom(dataClass) && !disableRemoved)
			{
				boolean current = cc.getInclude().contains(Include.REMOVED);
				sb.append("<option value=\"option:removed:").append(current ? "0" : "1").append("\"");
				sb.append(" title=\"Show/hide removed items\">");
				sb.append(current ? "&nbsp;" : "&nbsp;&nbsp;").append("Removed\n");
				table.addHidden("removed", Boolean.toString(current));
				numOptions++;
			}
			if (OwnableData.class.isAssignableFrom(dataClass) && !disableOwnedBy)
			{
				boolean current = cc.getInclude().contains(Include.MINE);
				sb.append("<option value=\"option:owned:").append(current ? "0" : "1").append("\"");
				sb.append(" title=\"Show/hide items owned by me\">");
				sb.append(current ? "&nbsp;" : "&nbsp;&nbsp;").append("Owned by me\n");
				table.addHidden("owned", Boolean.toString(current));
				numOptions++;
			}
			if (ShareableData.class.isAssignableFrom(dataClass) 
					|| ProjectData.class.isAssignableFrom(dataClass)
				)
			{
				if (!disableSharedTo)
				{
					boolean current = cc.getInclude().contains(Include.SHARED);
					sb.append("<option value=\"option:shared:").append(current ? "0" : "1").append("\"");
					sb.append(" title=\"Show/hide items shared to me\">");
					sb.append(current ? "&nbsp;" : "&nbsp;&nbsp;").append("Shared to me\n");
					table.addHidden("shared", Boolean.toString(current));
					numOptions++;
				}
				
				if (sc.getActiveProjectId() != 0 && !ProjectData.class.isAssignableFrom(dataClass) && !disableInProject)
				{
					boolean current = cc.getInclude().contains(Include.IN_PROJECT);
					sb.append("<option value=\"option:inproject:").append(current ? "0" : "1").append("\"");
					sb.append(" title=\"Show/hide items in current project\">");
					sb.append(current ? "&nbsp;" : "&nbsp;&nbsp;").append("In current project\n");
					table.addHidden("inproject", Boolean.toString(current));
					numOptions++;
				}
			}
	
			if (sc.hasPermission(Permission.READ, itemType) && !disableOthers && 
					(
						OwnableData.class.isAssignableFrom(dataClass) || 
						SessionData.class.isAssignableFrom(dataClass)
					)
				)
			{
				boolean current = cc.getInclude().contains(Include.OTHERS);
				sb.append("<option value=\"option:others:").append(current ? "0" : "1").append("\"");
				sb.append(" title=\"Show/hide items owned by others\">");
				sb.append(current ? "&nbsp;" : "&nbsp;&nbsp;").append("Owned by others\n");
				table.addHidden("others", Boolean.toString(current));
				numOptions++;
			}
		}

		// Context managemenet operations
		if (numOptions > 0)
		{
			sb.append("<option value=\"\">------------------\n");
			numOptions = 0;
		}
		sb.append("<option value=\"clearfilter:\" title=\"Clear all filters on visible columns\">Clear filter\n");
		sb.append("<option value=\"saveas:"+subContext+"\" title=\"Save current settings\">Save as...\n");
		sb.append("<option value=\"manage:"+subContext+"\" title=\"Manage saved contexts\">Manage...\n");
		sb.append("</select></td>");
		
		List<String> hiddenFilteredColumns = table.getHiddenFilteredColumns();
		if (hiddenFilteredColumns != null && hiddenFilteredColumns.size() > 0)
		{
			String columns = Values.getString(hiddenFilteredColumns, ",", true);
			sb.append("<td>&nbsp;</td>");
			sb.append("<td><a href=\"javascript:Table.showColumns('").append(table.getId()).append("', '").append(columns).append("')\"");
			sb.append(" title=\"There are hidden columns with a filter. Click here to show those columns.\">");
			sb.append("<img src=\"").append(table.getPage().getRoot()).append("images/hiddenfilter.gif\" border=\"0\">");
			sb.append("</a></td>");
		}
		sb.append("</tr></table>");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_BODY_INCLUDE;
	}

	public int doEndTag()
		throws JspException
	{
		if (!isVisible()) return EVAL_PAGE;
		StringBuilder sb = new StringBuilder();
		sb.append("</th>\n");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_PAGE;
	}

}


