/*
	$Id: BlockInfo.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

/**
	A <code>BlockInfo</code> object is used to represent the uniqe identifiers of an 
	{@link net.sf.basedb.core.ArrayDesignBlock ArrayDesignBlock}. It can either be the 
	block number or the metagrid coordinates. If block number is used metagrid can be 
	added but they are not used to identify an arraydesignblock and it is possible to 
	create two blocks with the same metagrid as long as the blocknumbers are uniqe.
	
	@author Enell, Nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public final class BlockInfo
{
	/**
		The block number.
	*/
	private final int block;

	/**
		The Y coordinate of the block.
	*/
	private final int metaGridY;
	
	/**
		The X coordinate of the block.
	*/
	private final int metaGridX;
	
	/**
		Create a new block info using the meta coordinates. Block is set to 0.
		@param metaGridX The row or x meta coordinate
		@param metaGridY The column or y meta coordinate
		@throws InvalidDataException If any of the coordinates is 0 or negative
	*/
	public BlockInfo(int metaGridX, int metaGridY)
		throws InvalidDataException
	{
		if (metaGridX <= 0) throw new NumberOutOfRangeException("metaGridX", metaGridX, 1, false);
		if (metaGridY <= 0) throw new NumberOutOfRangeException("metaGridY", metaGridY, 1, false);
		this.block = 0;
		this.metaGridX = metaGridX;
		this.metaGridY = metaGridY;
	}
	
	/**
		Create a new block info using meta coordinates.
		@param block The block number which must be greater or equal than 0
		@param metaGridX The row or x meta coordinate, use 0 if unknown
		@param metaGridY The column or y meta coordinate, use 0 if unknown
		@throws InvalidDataException If the block number is 0 or negative
	*/
	public BlockInfo(int block, int metaGridX, int metaGridY)
		throws InvalidDataException
	{
		if (block < 0) throw new NumberOutOfRangeException("block", block, 0, false);
		this.block = block;
		this.metaGridX = metaGridX;
		this.metaGridY = metaGridY;
	}

	/**
		Check if this object is equal to another <code>BlockInfo</code>
		object. They are equal if they have the same block number &gt; 0, or the same meta
		coordinates.
	*/
	@Override
	public final boolean equals(Object o)
	{
		if (!(o instanceof BlockInfo)) return false;
		BlockInfo bi = (BlockInfo) o;
		if (this.block > 0 && bi.block > 0) return this.block == bi.block;
		return this.metaGridX == bi.metaGridX && this.metaGridY == bi.metaGridY;
	}
	
	/**
		Calculate the hash code for the object.
	*/
	@Override
	public final int hashCode()
	{
		return 59 * block + 31 * metaGridX + 67 * metaGridY;
	}
	
	/**
		Get the coordinate as [number;metaGridX,metaGridY]
	*/
	@Override
	public final String toString()
	{
		return "["+block+";"+metaGridX+","+metaGridY+"]";
	}

	/**
		Get the row or x meta coordinate.
	*/
	public int getMetaGridX()
	{
		return metaGridX;
	}

	/**
		Get the column or y meta coordinate.
	*/
	public int getMetaGridY()
	{
		return metaGridY;
	}

	/**
		Get the block number. If the block number is 0 it is considerd unknown and 
		meta coordinates are used instead.
	*/
	public int getBlock()
	{
		return block;
	}
}
