/**
	$Id: Coloring.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.data.ColoringData;

/**
	This class keeps coloring information for items which can be colored when values
	are displayed in a client application. Color values are calculated by a 
	{@link net.sf.basedb.util.ColorGenerator} based on this coloring information and the colors
	specified by user settings.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
	@see net.sf.basedb.util.ColorGenerator
*/
public class Coloring
	implements AccessControlled
{
	private final ColoringData data;
	private final AccessControlled item;
	
	/**
		Create Coloring object which links to an existing item.
		@param data The coloring data
		@param item The access controlled item
	*/
	Coloring(ColoringData data, AccessControlled item)
	{
		this.data = data;
		this.item = item;
	}

	/**
	 	Create a new independent coloring object.
	*/
	public Coloring()
	{
		this.data = new ColoringData();
		this.item = null;
	}
	
	/*
		From the AccessControlled interface
		-------------------------------------------
	*/
	public void checkPermission(Permission permission) 
		throws PermissionDeniedException
	{
		if (item != null) item.checkPermission(permission);
	}
	public Set<Permission> getPermissions()
	{
		return item == null ? null : item.getPermissions();
	}
	public boolean hasPermission(Permission permission)
	{
		return item == null ? true : item.hasPermission(permission);
	}
	// -------------------------------------------

	/**
		Check that all values has been specified and that
		minValue &lt;= midValue &gt;= maxValue.
	*/
	void validate()
		throws InvalidDataException
	{
		if (isUsingColors())
		{
			Float minValue = getMinValue();
			Float midValue = getMidValue();
			Float maxValue = getMaxValue();
			if (minValue == null) throw new InvalidUseOfNullException("minValue");
			if (midValue == null) throw new InvalidUseOfNullException("midValue");
			if (maxValue == null) throw new InvalidUseOfNullException("maxValue");
			if (minValue > midValue) 
			{
				throw new NumberOutOfRangeException("minValue", minValue, midValue, true);
			}
			if (midValue > maxValue) 
			{
				throw new NumberOutOfRangeException("midValue", midValue, maxValue, true);
			}
		}
	}
	
	/**
		If this item should be colored or not. If this method returns FALSE,
		the rest of the properties on this object are ignored.
		@return TRUE if colors should be used, FALSE otherwise
	*/
	public boolean isUsingColors()
	{
		return data.isUsingColors();
	}
	/**
		Specifiy if colors should be used when displaying values for this
		item or not.
		@param useColors TRUE to use colors, FALSE otherwise
	*/
	public void setUsingColors(boolean useColors)
	{
		checkPermission(Permission.WRITE);
		data.setUsingColors(useColors);
	}
	
	/**
		If we should take the logarithm of values before the color is 
		generated or not.
		@return TRUE to calculate the logarithm, FALSE to leave the value as is
	*/
	public boolean isLogarithmic()
	{
		return data.isLogarithmic();
	}
	
	/**
		Specify if we should take the logarithm of values before the color is 
		generated or not.
		@param logarithmic TRUE to calculate the logarithm, FALSE to leave the value as is
	*/
	public void setLogarithmic(boolean logarithmic)
	{
		checkPermission(Permission.WRITE);
		data.setLogarithmic(logarithmic);
	}
	
	/**
		Get the value that represents the minimum value. All values below this will get
		the same color.
		@return The min value
	*/
	public Float getMinValue()
	{
		return data.getMinValue();
	}
	
	/**
		Set the value that represents the minimum value. All values below this will get
		the same color.
	 	@param minValue Value to set. 
	*/
	public void setMinValue(Float minValue)
	{
		checkPermission(Permission.WRITE);
		data.setMinValue(minValue);
	}

	/**
		Get the value that represents the midpoint value. 
		@return The midpoint value
	*/
	public Float getMidValue()
	{
		return data.getMidValue();
	}
	
	/**
		Set the value that represents the midpoint value. 
		@param midValue The midpoint value
	*/
	public void setMidValue(Float midValue)
	{
		checkPermission(Permission.WRITE);
		data.setMidValue(midValue);
	}

	/**
		Get the value that represents the maximum value. All values above this will get
		the same color.
		@return The max value
	*/
	public Float getMaxValue()
	{
		return data.getMaxValue();
	}
	
	/**
		Set the value that represents the maximum value. All values above this will get
		the same color.
	 	@param maxValue The value to set. 
	*/
	public void setMaxValue(Float maxValue)
	{
		checkPermission(Permission.WRITE);
		data.setMaxValue(maxValue);
	}

}
