/*
	$Id: DataQuery.java 4479 2008-09-05 15:16:27Z jari $
	
	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import org.hibernate.StatelessSession;

import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.signal.SignalException;

/**
	An implementation of the {@link Query} interface that returns data
	objects. This type of query is only used for batchable items,
	like reporters and raw data.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class DataQuery<I extends BasicData>
	extends AbstractEntityQuery
{

	/**
		The class of the data objects that are returned.
	*/
	private final Class<I> dataClass;

	/**
		Create a new query for the specified item, using the default optional
		runtime filter.
		@param dataClass The class of the data objects that are returned
	*/
	DataQuery(Class<I> dataClass, String entityName)
	{
		super(Item.fromDataClass(dataClass), entityName, true);
		this.dataClass = dataClass;
	}
	
	/**
		Create a query that may have different return type and root entity.
		For example, to return reporters from a query rooted at features:
		<code>new DataQuery(ReporterData.class, Item.FEATURE, null, "reporter")</code>.
		This will be translated to something like (in HQL):
		<code>SELECT ftr.reporter FROM FeatureData ftr ...</code>

		@param dataClass The class of the data objects that are returned 
		@param rootType The root item type of the query
		@param rootName The root enitity name, or null to use
			the class name of the root type
		@param select The HQL property that we should select on the root
			type to get to the return type
		@since 2.8
	*/
	DataQuery(Class<I> dataClass, Item rootType, String rootName, String select)
	{
		super(Item.fromDataClass(dataClass), rootType, rootName, select, true, 
			QueryRuntimeFilterFactory.getOptionalFilter(rootType));
		this.dataClass = dataClass;
	}
	
	/**
		Create a new query for the specified item, using a non-default optional
		runtime filter.
		@param dataClass The class of the data objects that are returned
		@param optionalFilter A runtime filter replacing the default optional filter
			or null to not use any optional filter
	*/
	DataQuery(Class<I> dataClass, String entityName, QueryRuntimeFilter optionalFilter)
	{
		super(Item.fromDataClass(dataClass), entityName, true, optionalFilter);
		this.dataClass = dataClass;
	}

	/**
		Execute the query and return the results as an iterator.
		@param dc The <code>DbControl</code> used to access the database
			and check permissions
		@throws BaseException If there is an error
	*/
	public DataResultIterator<I> iterate(DbControl dc)
		throws BaseException
	{
		long totalCount = -1;
		ScrollIterator<I> result = null;
		SessionControl sc = dc.getSessionControl();
		enableFilters(dc);
		try
		{
			StatelessSession session = dc.getStatelessSession();
			// Load the total count if it is requested
			if (isReturningTotalCount())
			{
				totalCount = QueryExecutor.loadData(Long.class, getCountHqlQuery(dc), session, sc);
			}
			
			// Load the results
			result = 
				QueryExecutor.loadIterator(dataClass, getMainHqlQuery(dc), session, sc);
		}
		catch (InterruptedException ex)
		{
			throw new SignalException("Aborted by user", ex);
		}
		finally
		{
			disableFilters(dc);
		}
		
		return new DataResultIterator<I>(result, 
			dc.getSessionControl(), null, dataClass, totalCount);
	}

}
