/*
  $Id: Directory.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.data.DirectoryData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

import java.util.LinkedList;
import java.util.Set;

/**
	This is used to represent directory items. Directories can
	be used to group {@link File}:s into more manageable
	compartments. Directories can be organised into an arbitrarily
	deep tree-like structure. For each directory no two sub-directories
	may have the same name, and no two files may have the same name.
	However, it is legal for a file to have the same name as a directory.
	
	@author enell
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class Directory
	extends SharedItem<DirectoryData>
	implements Nameable, Removable, SystemItem
{

	/**
		The type of item represented by this class.
		@see Item#DIRECTORY
		@see #getType()
	*/
	public static final Item TYPE = Item.DIRECTORY;
	
	/**
		The id for the <code>Directory</code> item representing
		the root.
	*/
	public static final String ROOT = "net.sf.basedb.core.Directory.ROOT";

	/**
		The id for the <code>Directory</code> item representing
		the home directory. The home directory is the parent directory for
		the user's home directories.
	*/
	public static final String HOME = "net.sf.basedb.core.Directory.HOME";

	/**
		The id for the <code>Directory</code> item representing
		the template directory. The template directory is used to create
		the directory structure for new users.
	*/
	public static final String TEMPLATE = "net.sf.basedb.core.Directory.TEMPLATE";


	/**
		Create a new <code>Directory</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param parent The parent directory for this directory. It cannot be 
			null 
		@return The new <code>Directory</code> item
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have {@link Permission#CREATE CREATE}
			permission for this type of items
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public static Directory getNew(DbControl dc, Directory parent)
		throws BaseException
	{
		Directory d = dc.newItem(Directory.class);
		d.setName("New directory");
		d.setParent(parent);
		d.setAutoCompress(parent.getAutoCompress());
		if (parent.getAutoShare())
		{
			ShareableUtil.copyPermissions(dc, parent, d);
			d.setAutoShare(true);
		}
		return d;
	}

	/**
		Create a new <code>Directory</code> from a path. Every directory that doesn't 
		exist is created. 
		
		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param path The path to create
		@return The directory representing the path
	*/
	public static Directory getNew(DbControl dc, Path path)
	{
		Directory currentDir = null;
		if (path.getUserLogin() != null)
		{
			currentDir = getByPath(dc, new Path("~" + path.getUserLogin(), Path.Type.DIRECTORY));
		}
		else
		{
			currentDir = getByPath(dc, new Path("/", Path.Type.DIRECTORY));
		}
		
		String currentPath = currentDir.getPath().toString();
		
		for (int i = 0; i < path.getDirectoryCount(); i++)
		{
			currentPath += "/" + path.getDirectory(i);
			Path p = new Path(currentPath, Path.Type.DIRECTORY);
			Directory subDir = null;
			if (Directory.exists(dc, currentDir, path.getDirectory(i)))
			{
				subDir = Directory.getByPath(dc, p);
			}
			else
			{
				subDir = currentDir.newSubDirectory();
				subDir.setName(path.getDirectory(i));
				dc.saveItem(subDir);
			}
			currentDir = subDir;
		}
		return currentDir;
	}
	
	/**
		Get a <code>Directory</code> item when you know the ID.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Directory</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Directory getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Directory d = dc.loadItem(Directory.class, id);
		if (d == null) throw new ItemNotFoundException("Directory[id="+id+"]");
		return d;
	}
	
	/**
		Get a <code>Directory</code> item when you know the path.
		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param path The path to the directory to load
		@return The <code>Directory</code> item
		@throws ItemNotFoundException If an item with the specified path is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Directory getByPath(DbControl dc, Path path)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		int directoryId = getIdFromPath(dc, path);
		return getById(dc, directoryId);
	}
	
	/**
		Get the id of a directory when you know the path.
		
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param path The path to the directory to load
		@return The id of the <code>Directory</code> item
		@throws ItemNotFoundException If an item with the specified path is not found
		@throws BaseException If there is another error
	*/
	public static int getIdFromPath(DbControl dc, Path path)
		throws BaseException
	{
		DirectoryData current = null;
		org.hibernate.Session session = dc.getHibernateSession();
		if (path.getUserLogin() != null)
		{
			UserData userData = null;
			if ("".equals(path.getUserLogin()))
			{
				// Represents the current user
				int userId = dc.getSessionControl().getLoggedInUserId();
				userData = HibernateUtil.loadData(dc.getHibernateSession(), UserData.class, userId);
				if (userData == null)
				{
					throw new ItemNotFoundException("User[id="+userId+"]");
				}
			}
			else
			{
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), "GET_USER_FOR_LOGIN");
				query.setString("login", path.getUserLogin());
				userData = HibernateUtil.loadData(UserData.class, query);
				if (userData == null)
				{
					throw new ItemNotFoundException("User[login="+path.getUserLogin()+"]");
				}
			}
			current = userData.getHomeDirectory();
		}
		else
		{
			current = HibernateUtil.loadData(session, DirectoryData.class, SystemItems.getId(Directory.ROOT));
		}
		if (current == null) throw new ItemNotFoundException("Directory[path="+path+"]");
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_SUBDIRECTORY_WITH_NAME");
		for (int i = 0; i < path.getDirectoryCount(); i++)
		{
			query.setEntity("parent", current);
			query.setString("name", path.getDirectory(i));
			current = HibernateUtil.loadData(DirectoryData.class, query);
			if (current == null) 
			{
				throw new ItemNotFoundException("Directory[path="+path+"]");
			}
		}
		return current.getId();
	}

	/**
		Checks if a subdirectory with the specified name exists in a
		directory.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param directory The directory to check in
		@param subdirectoryName The name of the subdirectory to look for
		@return TRUE if the file exists, FALSE otherwise
		@throws BaseException If there is an error
	*/
	public static boolean exists(DbControl dc, Directory directory, String subdirectoryName)
		throws BaseException
	{
		if (subdirectoryName != null) subdirectoryName = subdirectoryName.trim();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), "GET_SUBDIRECTORY_WITH_NAME");
		/*
			SELECT dir
			FROM DirectoryData dir
			WHERE dir.name = :name
			AND dir.parent = :parent
		*/
		query.setInteger("parent", directory.getData().getId());
		query.setString("name", subdirectoryName);
		DirectoryData subdirectory = HibernateUtil.loadData(DirectoryData.class, query);
		return subdirectory != null;
	}

	/**
		Creates a homedirectory for a specified user. The Directory will be placed in
		the predefined home directory, {@link Directory#HOME}.
		
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param user The user whose home directry should be created.
		@param copyTemplate TRUE if the new home directory shold copy the directory 
			structure in the template directory. 
		@return The new hom directory.
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#WRITE} or {@link Permission#CREATE} permission 
			to the item.
		@throws BaseException If there is another error.
	 */
	public static Directory createHomeDirectory(DbControl dc, User user, boolean copyTemplate)
		throws PermissionDeniedException, BaseException
	{
		SessionControl sc = dc.getSessionControl();
		user.checkPermission(Permission.WRITE);
		if (!sc.hasPermission(Permission.CREATE, Item.DIRECTORY))
		{
			throw new PermissionDeniedException(Permission.CREATE, "Directory");	
		}

		Directory allHome = Directory.getById(dc, SystemItems.getId(Directory.HOME));
		Directory userHome = Directory.getNew(dc, allHome);
		userHome.setName(Path.makeSafeFilename(user.getLogin(), ""));
		userHome.setOwner(user);
		userHome.setProjectKey(null);
		dc.saveItem(userHome);
		user.setHomeDirectory(userHome);
		
		if (copyTemplate)
		{
			Directory template = Directory.getById(dc, SystemItems.getId(Directory.TEMPLATE));
			copySubdirectories(dc, template, userHome, user);
		}
		return userHome;
	}

	/**
		This method copies a directory strucure to the destination directory.
		
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param from The source directory to copy
		@param to The destination directory where the new directory structure is created
		@param owner The owner of the destination directory
		@throws BaseException If there is an error
	*/
	public static void copySubdirectories(DbControl dc, Directory from, Directory to, User owner)
		throws BaseException
	{
		ItemQuery<Directory> query = Directory.getQuery();
		query.include(Include.MINE, Include.OTHERS, Include.SHARED);
		query.restrict(
			Restrictions.eq(
				Hql.property("parent"),
				Hql.entity(from)
			)
		);

		ItemResultIterator<Directory> children = query.iterate(dc);
		while (children.hasNext())
		{
			Directory child = children.next();
			Directory copy = Directory.getNew(dc, to);
			copy.setOwner(owner);
			copy.setProjectKey(null);
			copy.setName(child.getName());
			copy.setDescription(child.getDescription());
			copy.setAutoCompress(child.getAutoCompress());
			dc.saveItem(copy);
			copySubdirectories(dc, child, copy, owner);
		}
	}

	/**
		Get a query configured to retrieve directories.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Directory> getQuery()
	{
		return new ItemQuery<Directory>(Directory.class);
	}
	
	/**
 		Creates a new directory item from the given data.
 	
 		@param data the data.
	*/
	Directory(DirectoryData data)
	{
		super(data);
	}
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}

	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (!Path.isValidName(name)) throw new InvalidDataException("A directory cannot contain the following characters "+Path.INVALID_CHARACTERS);
		NameableUtil.setName(getData(), name);
	}

	public String getDescription()
	{
		return getData().getDescription();
	}

	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}

	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the SystemItem interface
		-------------------------------------------
	*/
	public String getSystemId()
	{
		return getData().getSystemId();
	}
	public boolean isSystemItem()
	{
		return getSystemId() != null;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check that:
		<ul>
		<li>no files are in the directory
		<li>no subdirectories are in the directory
		<li>no user has this directory as the home directory
		</ul>
		
		@return TRUE if the directory is used.
		@throws BaseException If there is an error.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		long count;
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FILES_IN_DIRECTORY", "count(*)");
		/*
			SELECT {1}
			FROM FileData f
			WHERE f.directory = :directory
		*/
		query.setEntity("directory", getData());
		count = HibernateUtil.loadData(Long.class, query); 
		if (count == 0)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_SUBDIRECTORIES_IN_DIRECTORY", "count(*)");
			/*
				SELECT {1}
				FROM DirectoryData d
				WHERE d.parent = :directory
			*/
			query.setEntity("directory", getData());
			count = HibernateUtil.loadData(Long.class, query);
		}
		if (count == 0)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_HOME_DIRECTORIES", "count(*)");
			/*
				SELECT {1}
				FROM UserData usr
				WHERE usr.homeDirectory = :directory
			*/
			query.setEntity("directory", getData());
			count = HibernateUtil.loadData(Long.class, query);
		}
		if (count == 0)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_EXPERIMENT_DIRECTORIES", "count(*)");
			/*
				SELECT {1}
				FROM ExperimentData xpr
				WHERE xpr.directory = :directory
			*/
			query.setEntity("directory", getData());
			count = HibernateUtil.loadData(Long.class, query);
		}
		return count > 0 || super.isUsed();
	}

	/**
		Get all:
		<ul>
		<li>{@link File}:s in this directory
		<li>Sub-{@link Directory}:s in this directory
		<li>{@link User}:s having this directory as their home-directory
		<li>{@link Experiment}:s having this directory as the plugin directory
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Files
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FILES_IN_DIRECTORY", "f.id");
			/*
				SELECT {1}
				FROM FileData f
				WHERE f.directory = :directory
			*/
		query.setEntity("directory", getData());
		addUsingItems(using, Item.FILE, query);

		// Sub-directories
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SUBDIRECTORIES_IN_DIRECTORY", "d.id");
			/*
				SELECT {1}
				FROM DirectoryData d
				WHERE d.parent = :directory
			*/
		query.setEntity("directory", getData());
		addUsingItems(using, Item.DIRECTORY, query);

		// Users with home-directories
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_HOME_DIRECTORIES", "usr.id");
			/*
				SELECT {1}
				FROM UserData usr
				WHERE usr.homeDirectory = :directory
			*/
		query.setEntity("directory", getData());
		addUsingItems(using, Item.USER, query);

		// Experiments
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_EXPERIMENT_DIRECTORIES", "xpr.id");
			/*
				SELECT {1}
				FROM ExperimentData xpr
				WHERE xpr.directory = :directory
			*/
		query.setEntity("directory", getData());
		addUsingItems(using, Item.EXPERIMENT, query);
		return using;
	}
	
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (isSystemItem())
		{
			denied |= Permission.deny(Permission.DELETE, Permission.CREATE);
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------
	/**
		Get the parent directory of this directory.
		
		@return The parent <code>Directory</code>, or null if
			this directory is a root directory
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have {@link Permission#READ READ}
			permission to the items
		@throws BaseException If there is another error
	*/
	public Directory getParent() 
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Directory.class, getData().getParent());
	}
	
	/**
		Set the parent directory of this directory.
		@param parent The new parent directory. Cant be null.
		@throws PermissionDeniedException If the logged in user doesn't
			have {@link Permission#WRITE} permission to this directory or
			{@link Permission#USE} permission to the parent directory
			permission for the parent directory
		@throws InvalidDataException If parent is null or is a subdirectory
			to this directory
	*/
	public void setParent(Directory parent)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (isSystemItem()) throw new PermissionDeniedException(Permission.WRITE, toString());
		if (parent == null) throw new InvalidUseOfNullException("parent");
		parent.checkPermission(Permission.USE);

		DirectoryData dd = parent.getData();
		while (dd != null)
		{
			String path = getPath().toString();
			String parentPath = parent.getPath().toString();
			if (getData().equals(dd)) 
				throw new InvalidDataException("The parent '"+parentPath+"' is already a subdirectory to this directory '"+path+"'.");
			dd = dd.getParent();
		}
		getData().setParent(parent.getData());
	}
	
	/**
		If files that are uploaded to this directory should be
		compressed or not. This setting only affects files when they
		are uploaded. It doesn't affect files that are moved between
		directories.
		@return TRUE to automatically compress files uploaded to this directory
		@since 2.5
	*/
	public boolean getAutoCompress()
	{
		return getData().getAutoCompress();
	}
	
	/**
		Set if files of should automatically
		be compressed when they are uploaded to this directory.
		@param autoCompress TRUE to enable auto-compression
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@since 2.5
	*/
	public void setAutoCompress(boolean autoCompress)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setAutoCompress(autoCompress);
	}
	
	/**
		If new files and subdirectories that are created in this
		directory automatically should be shared wtih the same
		permissions as this directory. This setting only affects
		new files/directories.

		@return TRUE to automatically share new files and subdreictories
		@since 2.8
	*/
	public boolean getAutoShare()
	{
		return getData().getAutoShare();
	}
	
	/**
		Set if new files and subdirectories of should automatically
		be shared with the same permissions as this directory. If a
		project is active and this directory isn't already shared
		to the project, the permissions will be extended to also
		share the new item to the project.
		
		@param autoShare TRUE to enable auto-sharing
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@since 2.8
	*/
	public void setAutoShare(boolean autoShare)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setAutoShare(autoShare);
	}
	
	/**
		Returns TRUE if this directories parent equals null and therfore is a root directory.
		
		@return TRUE if this directory is root directory.
	*/
	public boolean isRootDirectory()
	{
		return getData().getParent() == null;
	}
	
	/**
		Returns TRUE if this directory is a subdirectory to {@link Directory#HOME}.
		
		@return TRUE is this directory is a home directory.
	*/
	public boolean isHomeDirectory()
	{
		DirectoryData parent = getData().getParent();
		return parent != null && parent.getId() == SystemItems.getId(HOME);
	}
	
	/**
		Get the complete path to this directory.
		
		@return A {@link Path} object
	*/
	public Path getPath() 
	{
		DirectoryData dd = getData();
		LinkedList<String> directories = new LinkedList<String>();
		int rootId = SystemItems.getId(Directory.ROOT);
		while (dd != null && dd.getId() != rootId)
		{
			directories.addFirst(dd.getName());
			dd = dd.getParent();
		}
		return new Path(null, directories.toArray(new String[directories.size()]), null);
	}
	
	/**
		Create a new subdirectory to this directory.
		
		@return The new <code>Directory</code> item
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have {@link Permission#CREATE CREATE}
			permission for directory items, or {@link Permission#WRITE WRITE}
			permission for this directory
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public Directory newSubDirectory()
		throws PermissionDeniedException, BaseException
	{
		Directory sub = Directory.getNew(getDbControl(), this);
		return sub;
	}
	
	/**
		Get a query that return subdirectories of this directory.
		
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<Directory> getSubDirectories()
	{
		ItemQuery<Directory> query = Directory.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("parent"), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Get a query that return all files in this directory.
		
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<File> getFiles()
	{
		ItemQuery<File> query = File.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("directory"), 
				Hql.entity(this)
			)
		);
		return query;
	}
}
