/*
  $Id: FeatureIdentificationMethod.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.core.data.FeatureData;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.dbengine.DbEngine;

/**
	This enumeration defines constants that are used to specify 
	which method to use for identifying features on an array design.

	@author Nicklas
	@version 2.6
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
 */
public enum FeatureIdentificationMethod
{
	
	/**
		None, because the array design is a file-only platform.
	 */
	NONE(0, false, "None")
	{
		/**
			@return Always null
		*/
		Object getIdentifier(FeatureData feature)
		{
			return null;
		}
		/**
			@return Always null
		*/
		Object getIdentifier(RawData rawData, String externalFeatureId)
		{
			return null;
		}
	},
	
	/**
		Use coordinates (block, metaGridX, metaGridY, row, column)
	*/
	COORDINATES(1, false, "Coordinates")
	{
		/**
			@return A {@link FeatureCoordinate} object
		*/
		Object getIdentifier(FeatureData feature)
		{
			return Feature.getFeatureCoordinate(feature);
		}
		/**
			@return A {@link FeatureCoordinate} object created from
				coordinate values in the raw data object
		*/
		Object getIdentifier(RawData raw, String externalFeatureId)
		{
			return new FeatureCoordinate(raw.getBlock(), raw.getMetaGridX(),
				raw.getMetaGridY(), raw.getRow(), raw.getColumn());
		}
	},
	
	/**
		Use position number.
	*/
	POSITION(2, false, "Position")	
	{
		/**
			@return The position number
		*/
		Object getIdentifier(FeatureData feature)
		{
			return feature.getPosition();
		}
		/**
			@return The position value
		*/
		Object getIdentifier(RawData rawData, String externalFeatureId)
		{
			return rawData.getPosition();
		}
	},

	/**
		Use the external feature ID.
	*/
	FEATURE_ID(3, true, "Feature ID")
	{
		/**
			@return The external ID.
		*/
		Object getIdentifier(FeatureData feature)
		{
			return feature.getExternalId();
		}
		/**
			@return The externalFeatureId
		*/
		Object getIdentifier(RawData rawData, String externalFeatureId)
		{
			return externalFeatureId;
		}
	};
	
	/**
		Maps an integer to a location.
	*/
	private static final Map<Integer, FeatureIdentificationMethod> valueMapping = 
		new HashMap<Integer, FeatureIdentificationMethod>();
	
	static
	{
		for(FeatureIdentificationMethod method : FeatureIdentificationMethod.values())
		{
			FeatureIdentificationMethod l = valueMapping.put(method.getValue(), method);
			assert l == null : "Another feature identification method with the value "
				+method.getValue()+" already exists";
		}
	}
	
	/**
		The integer value of this method. 
	*/
	private final int value;
	/**
		Returned by the toString method.
	*/
	private final String displayValue;
	
	/**
		If the feature identifiers may be affect by case sensitivity at the database
		level or not.
	*/
	private final boolean caseInsensitive;

	/**
	 	Creates a new feature identification method. Used internal only.
	 	
		@param value The integer value of this location.
	*/
	private FeatureIdentificationMethod(int value, boolean caseInsensitive, String displayValue)
	{
		this.value = value;
		this.displayValue = displayValue;
		this.caseInsensitive = caseInsensitive;
	}
	
	
	public String toString()
	{
		return displayValue;
	}
	
	/**
		Get the <code>FeatureIdentificationMethod</code> object when you know the integer code.
		
		@param value The integer value.
		@return The feature identification method for the integer value.
	*/
	public static FeatureIdentificationMethod fromValue(int value)
	{
		FeatureIdentificationMethod method = valueMapping.get(value);
		assert method != null : "method == null for value "+value;
		return method;
	}
	
	/**
		Get the integer value that is used when storing the feature identification method
		to the database.
		@return The integer value for this method.
	*/
	public int getValue()
	{
		return value;
	}
	
	/**
		Return TRUE if this identification method can be case insensitive. Ie. 
		because the identifier is a string object. In this case we must check if
		the database is case sensitive or not. If FALSE, case is not in issue,
		ie. because the identifier is not a string object.
		@see DbEngine#caseInsensitiveComparison()
	*/
	public boolean caseInsensitive()
	{
		return caseInsensitive;
	}
	
	/**
		Get the feature identifier for this method.
		@param feature The feature to get the identifier for
	*/
	abstract Object getIdentifier(FeatureData feature);
	
	/**
		Get a feature identifier using this method
		@param raw The raw data to pick coordinates/position from
		@param externalFeatureId The external feature ID, null can be used if not
			needed by this particular identification method
	*/
	abstract Object getIdentifier(RawData raw, String externalFeatureId);
	
}
