/*
	$Id: Job.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Johan Enell, Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg, Martin Svensson, Gregory Vincic
	Copyright (C) 2007 Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ExperimentData;
import net.sf.basedb.core.data.JobData;
import net.sf.basedb.core.data.MessageData;
import net.sf.basedb.core.data.ParameterValueData;
import net.sf.basedb.core.data.PluginConfigurationData;
import net.sf.basedb.core.data.PluginDefinitionData;

import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.plugin.Plugin.MainType;
import net.sf.basedb.core.signal.Signal;
import net.sf.basedb.core.signal.SignalTransporter;

import java.util.Collection;
import java.util.Date;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.Collections;
import java.util.Arrays;
import java.io.PrintWriter;
import java.io.StringWriter;

/**
	This class represents the execution of a job. Normally a job is executed by
	a plugin, but it can also be executed by an external program. To find the
	type of the job check the {@link #getJobType()} method. Many of the methods
	in this class only apply to jobs of one of the types. The following methods can
	only be used on jobs that are executed by a plugin:
	
	<ul>
	<li>{@link #configure(GuiContext)}: Starts the configuration sequence of the job
	<li>{@link #execute(ProgressReporter, String)}: Starts executing the job
	</ul>
	
	The following methods can only be used by jobs executed by an external program:
	<ul>
	<li>{@link #start(String, String)}: Tell the core that job has started
	<li>{@link #setProgress(int, String)}: Update the progress status of the job
	<li>{@link #doneOk(String)}: Tell the core that the job finished successfully
	<li>{@link #doneError(String)}: Thell the core that the job finished with an error
	</ul>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class Job
	extends OwnedItem<JobData>
	implements Nameable, Removable, Transactional
{
	/**
		The type of item represented by this class.
		@see Item#JOB
		@see #getType()
	*/
	public static final Item TYPE = Item.JOB;

	/**
		Create a new job.
		@deprecated Use {@link #getNew(DbControl, PluginDefinition, PluginConfiguration, Experiment)}
			instead
	*/
	public static Job getNew(DbControl dc, PluginDefinition plugin, PluginConfiguration config)
		throws InvalidDataException, BaseException
	{
		return getNew(dc, plugin, config, null);
	}
	
	/**
		Create a new <code>Job</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
	 	@param plugin The plugin this job should use. 
		@param config The plugin configuration this job should use
			or null if the job is executed by an external program
		@param experiment The experiment the job is a part of, or null
			if the job is not part of an experiment
		@return The new <code>Job</code> item
		@throws InvalidDataException If the specified plugin requires a configuration but the
			config parameter is null
		@throws BaseException If there is an error
		@since 2.7
	*/
	public static Job getNew(DbControl dc, PluginDefinition plugin, 
		PluginConfiguration config, Experiment experiment)
		throws InvalidDataException, BaseException
	{
		Job j = dc.newItem(Job.class);
		JobData jobData = j.getData();
		if (config == null && plugin == null)
		{
			jobData.setType(Type.OTHER.getValue());
			jobData.setPriority(5);
		}
		else
		{
			if (plugin == null) plugin = config.getPluginDefinition();
			
			if (plugin.requiresConfiguration() && config == null)
			{
				throw new InvalidDataException("Plugin " + plugin.getName() + " requires a configuration");
			}
			
			j.setPluginDefinition(plugin);
			j.setPluginConfiguration(config);
			jobData.setType(Type.RUN_PLUGIN.getValue());
			Plugin.MainType pluginType = plugin.getMainType();
			if (pluginType == Plugin.MainType.IMPORT)
			{
				jobData.setPriority(4);
			}
			else
			{
				jobData.setPriority(5);
			}
		}

		j.setName("New job");
		jobData.setStatus(Status.UNCONFIGURED.getValue());
		jobData.setCreated(new Date());
		jobData.setSendMessage(true);
		jobData.setRemoveJobWhenFinished(false);
		if (experiment != null) jobData.setExperiment(experiment.getData());
		return j;
	}

	/**
		Get a <code>Job</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Job</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static Job getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Job j = dc.loadItem(Job.class, id);
		if (j == null) throw new ItemNotFoundException("Job[id="+id+"]");
		return j;
	}

	/**
		Get a query configured to retrieve <jobs.

		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Job> getQuery()
	{
		return new ItemQuery<Job>(Job.class);
	}

	private JobSignalTransporter signalTransporter;
	
	Job(JobData jobData)
	{
		super(jobData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Return TRUE if the status is EXECUTING or:
		<ul>
		<li>a transformation is linked to this job
		<li>a raw bioassay is linked to this job
		<li>an array design is linked to this job
		</ul>
		FALSE otherwise.
	*/
	public boolean isUsed()
		throws BaseException
	{
		boolean used = getStatus() == Status.EXECUTING;
		if (!used)
		{
			org.hibernate.Session session = getDbControl().getHibernateSession();
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_TRANSFORMATIONS_FOR_JOB", "count(*)");
			/*
				SELECT {1}
				FROM TransformationData trf
				WHERE trf.job = :job
			*/
			query.setEntity("job", this.getData());
			used = HibernateUtil.loadData(Long.class, query) > 0;
		}
		if (!used)
		{
			org.hibernate.Session session = getDbControl().getHibernateSession();
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_RAWBIOASSAYS_FOR_JOB", "count(*)");
			/*
				SELECT {1}
				FROM RawBioAssayData rba
				WHERE rba.job = :job
			*/
			query.setEntity("job", this.getData());
			used = HibernateUtil.loadData(Long.class, query) > 0;
		}
		if (!used)
		{
			org.hibernate.Session session = getDbControl().getHibernateSession();
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_ARRAYDESIGNS_FOR_JOB", "count(*)");
			/*
				SELECT {1}
				FROM ArrayDesignData ad
				WHERE ad.job = :job
			*/
			query.setEntity("job", this.getData());
			used = HibernateUtil.loadData(Long.class, query) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get the transformation and raw bioassays using this job.
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_TRANSFORMATIONS_FOR_JOB", "trf.id");
			/*
				SELECT {1}
				FROM TransformationData trf
				WHERE trf.job = :job
			*/
		query.setEntity("job", this.getData());
		addUsingItems(using, Item.TRANSFORMATION, query);
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_RAWBIOASSAYS_FOR_JOB", "rba.id");
			/*
				SELECT {1}
				FROM RawBioAssayData rba
				WHERE rba.job = :job
			*/
		query.setEntity("job", this.getData());
		addUsingItems(using, Item.RAWBIOASSAY, query);
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_ARRAYDESIGNS_FOR_JOB", "ad.id");
			/*
				SELECT {1}
				FROM ArrayDesignData ad
				WHERE ad.job = :job
			*/
		query.setEntity("job", this.getData());
		addUsingItems(using, Item.ARRAYDESIGN, query);
		return using;
	}
	/**
		Set's the project id to the currently active project.
	*/
	@Override
	void onBeforeCommit(Action action)
		throws NotLoggedInException, BaseException
	{
		super.onBeforeCommit(action);
		int projectId = getSessionControl().getActiveProjectId();
		if (projectId != 0)
		{
			getData().setActiveProjectId(projectId);
		}
	}
	
	/**
		If the {@link Signal#ABORT} signal is waiting to be sent send it.
		@since 2.6.1
	*/
	@Override
	void onAfterCommit(Action action)
	{
		super.onAfterCommit(action);
		if (signalTransporter != null) signalTransporter.onAfterCommit();
	}

	/**
		Grant read permission if the logged in user has read permission to the 
		experiment the job belongs to.
	*/
	@Override
	void initPermissions(int granted, int denied)
	{
		ExperimentData parent = getData().getExperiment();
		if (parent != null)
		{
			int parentPermission = 0;
			if (parent.getId() != 0)
			{
				parentPermission = getSessionControl().getAllPermissions(parent);
			}
			else
			{
				BasicItem<?> basicParent = getDbControl().getItem(BasicItem.class, parent);
				parentPermission = Permission.grant(basicParent.getPermissions());
			}
			if (Permission.hasPermission(parentPermission, Permission.READ))
			{
				granted |= Permission.grant(Permission.READ);
			}
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/**
		Get the type of job, ie. if it is a job executed by a plugin
		or not.
	*/
	public Job.Type getJobType()
	{
		return Type.fromValue(getData().getType());
	}
	
	/**
		Get the {@link PluginConfiguration} defining the plugin this job should execute.
		
		@return The <code>PluginConfiguration</code> item, or null if this job 
			isn't executed by a plugin
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the plugin configuration
		@throws BaseException If there is another error
	*/
	public PluginDefinition getPluginDefinition()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PluginDefinition.class, getData().getPluginDefinition()); 
	}
	
	/**
		Set the plugin configuration of this job.
		@throws PermissionDeniedException If the logged in user
			doesn't have use permission for the plugin configuration
	*/
	private void setPluginDefinition(PluginDefinition plugin)
		throws PermissionDeniedException
	{
		if (plugin == null) throw new InvalidUseOfNullException("plugin");
		plugin.checkPermission(Permission.USE);
		getData().setPluginDefinition(plugin.getData());
		getData().setPluginVersion(plugin.getVersionString());
	}

	
	/**
		Get the {@link PluginConfiguration} defining the plugin this job should execute.
		
		@return The <code>PluginConfiguration</code> item, or null if this job 
			isn't executed by a plugin
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the plugin configuration
		@throws BaseException If there is another error
	*/
	public PluginConfiguration getPluginConfiguration()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PluginConfiguration.class, getData().getPluginConfiguration()); 
	}
	
	/**
		Set the plugin configuration of this job.
		@throws PermissionDeniedException If the logged in user
			doesn't have use permission for the plugin configuration
	*/
	private void setPluginConfiguration(PluginConfiguration pluginConfiguration)
		throws PermissionDeniedException
	{
		if (pluginConfiguration != null) 
		{
			pluginConfiguration.checkPermission(Permission.USE);
			getData().setPluginConfiguration(pluginConfiguration.getData());
			getData().setParameterVersion(pluginConfiguration.getParameterVersion());
		}
	}
	
	/**
		Get the experiment this job is a part of.
		@return The <code>Experiment</code> item, or null if this job 
			isn't part of an experiment
		@since 2.4
	*/
	public Experiment getExperiment()
	{
		return getDbControl().getItem(Experiment.class, getData().getExperiment()); 		
	}
	/**
		Set the experiment this job is part of. Once an experiment has been set, it
		can't be changed.
		@since 2.4
	*/
	void setExperiment(Experiment experiment)
	{
		ExperimentData current = getData().getExperiment();
		ExperimentData next = experiment == null ? null : experiment.getData();
		if (current == null)
		{
			getData().setExperiment(next);
		}
		else if (!current.equals(next))
		{
			throw new PermissionDeniedException("A job can only be part of one experiment: " + this);
		}
	}
	
	/**
		Get the version of the plugin used to execute this job at the time the
		job was executed or registered.
		@return The version string, or null if not known
		@since 2.8
	*/
	public String getPluginVersion()
	{
		return getData().getPluginVersion();
	}
	
	/**
		Get the version number of the configuration parameters used for this job.
	*/
	public int getParameterVersion()
	{
		return getData().getParameterVersion();
	}

	/**
		If a message should be sent to the owner once the job has finished
		or not.
	*/
	public boolean getSendMessage()
	{
		return getData().getSendMessage();
	}

	/**
		Set if a message should be sent to owner when the job finishes.
	 	@param sendMessage TRUE if a message should be sent, FALSE otherwise. 
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
	*/
	public void setSendMessage(boolean sendMessage)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setSendMessage(sendMessage);
	}
	
	/**
		If the job should flagged for deletion after it's done successfully.
	 	@return TRUE if job is flagged to be removed, FALSE otherwise 
	*/
	public boolean getRemoveJobWhenFinished()
	{
		return getData().getRemoveJobWhenFinished();
	}
	
	/**
	 	Set if the job should be deleted after it has finished successfully.
	 	@param removeJobWhenFinished TRUE if the job should be 
	 		flagged for deletion when finished, FALSE otherwise. 
	 	@throws PermissionDeniedException If the logged in user doesn't have 
	 		write permission
	*/
	public void setRemoveJobWhenFinished(boolean removeJobWhenFinished)
	{
		checkPermission(Permission.WRITE);
		getData().setRemoveJobWhenFinished(removeJobWhenFinished);
	}

	/**
		Get the estimated execution time of the job. This value can be
		used by job queue managers to decide which job that should be
		executed at when.
		@return A <code>ExecutionTime</code> object
	*/
	public ExecutionTime getEstimatedExecutionTime()
	{
		return ExecutionTime.fromValue(getData().getEstimatedExecutionTime());
	}
	
	/**
		Get the status of the job.
	*/
	public Status getStatus()
	{
		return Status.fromValue(getData().getStatus());
	}
	
	/**
		Check if the job should be/was executed as a dry-run job.
		A dry-run job is a job that does all the things a normal job would 
		do except committing the changes to the database. The dry-run job 
		may output a log file with detailed information about what should 
		have happened if it hadn't been a dry-run.
		@since 2.8
	*/
	public boolean isDryRun()
	{
		return getData().isDryRun();
	}
		
	/**
		The maximum allowed length of the status message.
	*/
	public static final int MAX_STATUS_MESSAGE_LENGTH = JobData.MAX_STATUS_MESSAGE_LENGTH;
	/**
		Get a message about the current status of the job.
	*/
	public String getStatusMessage()
	{
		return getData().getStatusMessage();
	}
	
	/**
		The maximum allowed length of the stack trace message.
	*/
	public static final int MAX_STACK_TRACE_LENGTH = JobData.MAX_STACK_TRACE_LENGTH;
	/**
		Get the stack trace in case the job ended with an error
	*/
	public String getStackTrace()
	{
		return getData().getStackTrace();
	}

	/**
		An estimate of the percentage of the work currently completed by 
		an executing job. The value is always 0 for a job that hasn't started
		executing, and 100 for a job that has finised (successfully or with an error).
	*/
	public int getPercentComplete()
	{
		return getData().getPercentComplete();
	}
	
	/**
		Get the priority of the job. The priority is a value between 1 and 10,
		with 1 as the highest priority.
	*/
	public int getPriority()
	{
		return getData().getPriority();
	}
	
	/**
		Set the priority of the job. The priority is a value between 1 and 10,
		with 1 as the highest priority.
	 	@param priority The priority to set. 
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission or the job's status isn't <code>WAITING</code>
		@throws InvalidDataException If the priority isn't between 1 and 10
	*/
	public void setPriority(int priority)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (getStatus() != Status.WAITING) 
		{
			throw new PermissionDeniedException("Cannot set the priority for job with status '"+getStatus()+"': "+toString());
		}
		// TODO - do we need a global permission for setting this property ?
		if (priority < 1 || priority > 10) throw new NumberOutOfRangeException("priority", priority, 1, 10);
		getData().setPriority(priority);
	}
	
	/**
	 	The ID of the project that should be made the active project while running
	 	this job. If the project can't be loaded the job should be executed without
	 	an active project.
		@return The ID of the active project, or 0 if no project should be active
	 */
	public int getActiveProjectId()
	{
		return getData().getActiveProjectId();
	}
	
	/**
		Get the date and time the job was registered in the database.
		@return A date
	*/
	public Date getCreated()
	{
		return DateUtil.copy(getData().getCreated());
	}

	/**
		Get the date and time the job was started.
		@return A date, or null if the job hasn't started
		@see #start(String, String)
	*/
	public Date getStarted()
	{
		return DateUtil.copy(getData().getStarted());
	}

	/**
		The maximum allowed length of the server name.
	*/
	public static final int MAX_SERVER_LENGTH = JobData.MAX_SERVER_LENGTH;
	/**
		Get the name of the server where the job is executing.
	*/
	public String getServer()
	{
		return getData().getServer();
	}
	
	/**
		Checks if this job has saved information about the context that was active
		when this job was created. If context information is available it is possible
		to re-configure a failed job.
		@see #configure(GuiContext)
		@see #getLastGuiContext()
		@return TRUE if context information is available, FALSE otherwise
		@since 2.8
	*/
	public boolean hasContext()
	{
		return getData().getContextItem() != null && getData().getContextType() != null;
	}
	
	/**
		Get the gui-context that was active when this job was created.
		@return The active context, or null if not known
		@since 2.8
	*/
	public GuiContext getLastGuiContext()
	{
		GuiContext context = null;
		if (hasContext())
		{
			context = new GuiContext(Item.fromValue(getData().getContextItem()), 
					GuiContext.Type.fromValue(getData().getContextType()), getData().getSubContext());
		}
		return context;
	}
	
	
	/**
		Get the job agent this job should run on, is running on or 
		was running on.
		@return A job agent or null if not known
		@since 2.8
	*/
	public JobAgent getJobAgent()
	{
		Integer id = getData().getJobAgentId();
		JobAgent agent = null;
		if (id != null)
		{
			try
			{
				agent = JobAgent.getById(getDbControl(), id);
			}
			catch (ItemNotFoundException ex)
			{}
		}
		return agent;
	}
	
	/**
		Set the job agent this job should run on. If the job is already
		running or has ended, this call is ignored.
		<p>
		NOTE! This method doesn't check that the specified job agent
		has actually been configured to run the plug-in used by this job.
		Setting this parameter to an incorrect value may result in the
		job never being executed. 
		
		@param agent The job agent, or null to let any job agent run
			the job
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission on this job, or use permission on the
			job agent or is lacking the {@link Permission#SELECT_JOBAGENT}
			permission
		@since 2.8
	*/
	public void setJobAgent(JobAgent agent)
	{
		Job.Status status = getStatus();
		if (status != Job.Status.WAITING) return;
		checkPermission(Permission.WRITE);
		if (agent != null) 
		{
			if (!getSessionControl().hasSystemPermission(Permission.SELECT_JOBAGENT))
			{
				throw new PermissionDeniedException(Permission.SELECT_JOBAGENT);
			}
			agent.checkPermission(Permission.USE);
		}
		getData().setJobAgentId(agent == null ? null : agent.getId());
	}
	
	/**
		Set information needed to create a signal transporter object that can be used
		to send signals to the plug-in that is currently executing this job.
		@param clazz The signal transporter class
		@param initParams Initialisation parameters that will be passed
			to {@link SignalTransporter#init(String)}
		@throws PermissionDeniedException If the user doesn't have write
			permission for the job
		@since 2.6
	*/
	public void setSignalTransporter(Class<? extends SignalTransporter> clazz, String initParams)
	{
		checkPermission(Permission.WRITE);
		if (clazz == null) throw new InvalidUseOfNullException("clazz");
		if (initParams == null) initParams = "";
		getData().setSignalTransporter(clazz.getName() + ":" + initParams);
	}
	
	/**
		Checks if a signal transporter has been registered for this job or not.
		@return TRUE if a signal transporter is registered, FALSE otherwise
		@since 2.6
	*/
	public boolean hasSignalTransporter()
	{
		return getData().getSignalTransporter() != null;
	}
	
	/**
		Create and initialise a signal transporter object that can be used
		to send signal to the plug-in running this job. 
		<p>
		Note 1! Depending on the transport mechanism used, it is not certain that 
		the created transporter can actually send the signal. For example, some transporters
		only work if the plug-in is running in the same virtual machine, while others
		supports sending signals to other JVM:s on the same or a different
		computer. 
		<p>
		Note 2! The returned transporter will delay the {@link Signal#ABORT} signal
		until {@link DbControl#commit()} is called and will only send the signal if
		the transaction is completed succesfully. If the transaction fails or is rolled
		back the signal is not sent. Other signals are sent immediately.
		<p>
		Note 3! This method always returns the same transporter instance if
		it is called multiple times.
		
		@return An initialised signal transporter object, or null
			if no signal transporter has been registered
		@throws PermissionDeniedException If the user doesn't have write
			permission for the job
		@since 2.6
	*/
	public SignalTransporter getSignalTransporter()
	{
		checkPermission(Permission.WRITE);
		if (signalTransporter == null)
		{
			String tmp = getData().getSignalTransporter();
			if (tmp != null)
			{
				try
				{
					String[] tmp2 = tmp.split(":", 2);
					String transporterClass = tmp2[0];
					String initParams = tmp2.length > 1 ? tmp2[1] : null;
					SignalTransporter transporter = (SignalTransporter)Class.forName(transporterClass).newInstance();
					transporter.init(initParams);
					signalTransporter = new JobSignalTransporter(transporter);
				}
				catch (Exception ex)
				{
					throw new BaseException(ex);
				}
			}
		}
		return signalTransporter;
	}
	
	/**
		Get the date and time the job ended.
		@return A date, or null if the job hasn't ended
	*/
	public Date getEnded()
	{
		return DateUtil.copy(getData().getEnded());
	}

	/**
	 	@deprecated Use {@link #setPrepared(String, JobAgent)} instead
	*/
	public void setPrepared(String server)
		throws PermissionDeniedException
	{
		setPrepared(server, null);
	}
	
	/**
		Set the job's status to {@link Status#PREPARED}. This status is
		used to signal that a job is about to be executed, but the actual excecution
		hasn't started yet. This status prevents two job agents from trying to
		start the same job twice.
		
	 	@param server Name of the server which this job will executed on
		@param agent The job agent that is going to execute the
			job or null if it is executed by something else, eg. the
			internal job queue
	 	@throws PermissionDeniedException If the logged in user doesn't have write permission.
		@since 2.8
	*/
	public void setPrepared(String server, JobAgent agent)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (getStatus() != Status.WAITING && getStatus() != Status.UNCONFIGURED) 
		{
			throw new PermissionDeniedException("Can't prepare a job with status '"+getStatus()+"': "+toString());
		}
		JobData data = getData();
		data.setStarted(new Date());
		data.setStatus(Status.PREPARED.getValue());
		data.setServer(StringUtil.setNullableString(server, "server", MAX_SERVER_LENGTH));
		getData().setJobAgentId(agent == null ? null : agent.getId());
	}

	/**
		@deprecated Use {@link #start(String, String, JobAgent)} instead
	*/
	public void start(String statusMessage, String server)
		throws PermissionDeniedException, InvalidDataException
	{
		start(statusMessage, server, null);
	}
	
	/**
 		Register the job as started. Note! This method doesn't start the actual 
		execution of the job. This method should be called by an job runner application
		to tell the core that the job runner has started to execute the job.
		This method can't be used for jobs that are executed by plugins.
		
		@param statusMessage A message
		@param server The name of the server where the job is started
		@param agent The job agent that is going to execute the
			job or null if it is executed by something else, eg. the
			internal job queue
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the status message is too long
		@since 2.8
	 */
	public void start(String statusMessage, String server, JobAgent agent)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		JobData data = getData();
		data.setStatusMessage(StringUtil.setNullableString(statusMessage, "statusMessage", MAX_STATUS_MESSAGE_LENGTH));
		data.setServer(StringUtil.setNullableString(server, "server", MAX_SERVER_LENGTH));
		data.setStatus(Status.EXECUTING.getValue());
		data.setStarted(new Date());
		data.setPercentComplete(0);
		data.setJobAgentId(agent == null ? null : agent.getId());
		PluginDefinitionData plugin = data.getPluginDefinition();
		data.setPluginVersion(plugin == null ? null : plugin.getVersionString());
	}

	/**
		Get a progress reporter that reports progress by updating the information
		in the database.
		
		@param chained An optional chained progress reporter to which all
			updates will be forwarded
		@return A progress reporter object
		@since 2.4
	*/
	public ProgressReporter getProgressReporter(ProgressReporter chained)
	{
		return new ProgressReporterImpl(this, null, null, chained);
	}
	
	/**
		Set the progress of the job. The job must be in the {@link Status#EXECUTING} or
		{@link Status#PREPARED} status.
		@param percentComplete The number of percent completed
		@param statusMessage A message
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission of the job isn't in the <code>EXECUTING</code> status
		@throws InvalidDataException If the status message is too long or the
			percentage value isn't between 0 and 100
	*/
	public void setProgress(int percentComplete, String statusMessage)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (getStatus() != Status.EXECUTING && getStatus() != Status.PREPARED) 
		{
			throw new PermissionDeniedException("Cannot set progress for a job with status '"+getStatus()+"': "+toString());
		}
		if (percentComplete < 0 || percentComplete > 100) 
		{
			throw new NumberOutOfRangeException("percentComplete", percentComplete, 0, 100);
		}
		JobData data = getData();
		data.setStatusMessage(StringUtil.setNullableString(statusMessage, "statusMessage", MAX_STATUS_MESSAGE_LENGTH));
		data.setPercentComplete(percentComplete);
	}

	/**
		Register the job as successfully completed. 
		
		@param statusMessage A message
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the status message is too long
	*/
	public void doneOk(String statusMessage)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		JobData data = getData();
		data.setStatusMessage(StringUtil.setNullableString(statusMessage, "statusMessage", MAX_STATUS_MESSAGE_LENGTH));
		data.setPercentComplete(100);
		data.setStatus(Status.DONE.getValue());
		data.setEnded(new Date());
		data.setStackTrace(null);
		data.setSignalTransporter(null);
		if (getSendMessage()) sendMessage();
	}

	/**
		Register the job as completed with an error.
		
		@param statusMessage A message
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the status message is too long
	*/
	public void doneError(String statusMessage)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		JobData data = getData();
		data.setStatusMessage(StringUtil.setNullableString(statusMessage, "statusMessage", MAX_STATUS_MESSAGE_LENGTH));
		data.setPercentComplete(100);
		data.setStatus(Status.ERROR.getValue());
		data.setStackTrace(null);
		data.setEnded(new Date());
		data.setSignalTransporter(null);
		if (getSendMessage()) sendMessage();
	}
	
	/**
		Register the job as completed with an error.
		
		@param statusMessage A short message describing the error
		@param errors A list of one or more errors that caused the job to fail, the
			complete stack trace for those errors are stored in the {@link #getStackTrace()}
			property
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the status message is too long
	*/
	public void doneError(String statusMessage, Collection<? extends Throwable> errors)
	{
		checkPermission(Permission.WRITE);
		JobData data = getData();
		data.setStatusMessage(StringUtil.setNullableString(statusMessage, "statusMessage", MAX_STATUS_MESSAGE_LENGTH));
		data.setPercentComplete(100);
		data.setStatus(Status.ERROR.getValue());
		data.setSignalTransporter(null);
		if (errors != null)
		{
			StringWriter sw = new StringWriter();
			PrintWriter pw = new PrintWriter(sw);
			for (Throwable t : errors)
			{
				t.printStackTrace(pw);
			}
			pw.close();
			data.setStackTrace(sw.toString());
		}
		data.setEnded(new Date());
		if (getSendMessage()) sendMessage();
	}
	
	/**
		Retry a job which finished with an error.
		@param useLatestConfiguration If the job should use the latest configuration
			parameters in case these has changed since the job was created
	 	@throws PermissionDeniedException If the logged in user doesn't have write permission
	 	@deprecated Use {@link #retry(boolean, boolean)} instead
	*/
	public void retry(boolean useLatestConfiguration)
		throws PermissionDeniedException
	{
		retry(useLatestConfiguration, false);
	}
	
	/**
		Retry a job which finished with an error or a successful dry-run job.
		@param useLatestConfiguration If the job should use the latest configuration
			parameters in case these has changed since the job was created
		@param clearDryRun If the job was a dry-run job, clear the dry-run flag
			to indicate that it should be executed for real the next time
	 	@throws PermissionDeniedException If the logged in user doesn't have write permission
	*/
	public void retry(boolean useLatestConfiguration, boolean clearDryRun)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		JobData data = getData();
		data.setStatus(Job.Status.WAITING.getValue());
		data.setPercentComplete(0);
		data.setStarted(null);
		data.setEnded(null);
		data.setStatusMessage(null);
		data.setStackTrace(null);
		data.setServer(null);
		data.setJobAgentId(null);
		if (useLatestConfiguration)
		{
			PluginConfigurationData config = data.getPluginConfiguration();
			if (config != null)
			{
				data.setParameterVersion(config.getParameterVersion());
			}
		}
		if (clearDryRun) data.setDryRun(false);
	}

	
	/**
		Start the configuration sequence for a job. Do not
		call {@link DbControl#commit()} until the configuration sequence
		is done. See {@link net.sf.basedb.core.PluginRequest} for more  information. This method changes
		the status of the job to {@link Status#WAITING}.
	 
	 	@param context GUI context from where the configuration is called.
		@return A <code>PluginRequest</code> object
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws BaseException If there is another error
	*/
	public PluginConfigurationRequest configure(GuiContext context)
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (getJobType() != Type.RUN_PLUGIN)
		{
			throw new PermissionDeniedException("Not allowed for a job of type '"+getJobType()+"': "+toString());
		}
		if (getStatus() == Status.EXECUTING) 
		{
			throw new PermissionDeniedException("Cannot configure an executing job: "+toString());
		}
		PluginDefinition pd = getPluginDefinition();
		if (!pd.isInteractive())
		{
			throw new BaseException("Plugin can't be configured interactively: " + pd);
		}
		
		PluginConfiguration config = getPluginConfiguration();
		InteractivePlugin plugin = (InteractivePlugin)pd.newInstance(); 
		SessionControl sc = getSessionControl();
		ParameterValuesImpl parameters = new ParameterValuesImpl(this, false);
		
		PluginConfigurationRequest request = 
			new PluginConfigurationRequest(
				sc, plugin, Request.COMMAND_CONFIGURE_JOB,
				config, config == null ? null : config.getParameterValuesImpl(getParameterVersion()),
				this, parameters, context, pd, pd.getAllowImmediateExecution()
			);
		request.setProgressReporter(new ProgressReporterImpl(this, null, null, null));
		return request;
	}

	/**
		Start the execution sequence for a job. Call {@link DbControl#commit()}
		before calling {@link net.sf.basedb.core.PluginRequest#invoke()}. 
		See {@link net.sf.basedb.core.PluginRequest} for more  information. 
	
		@param progress The {@link ProgressReporter} where the plugin can report its progress. If 
			<code>null</code> {@link net.sf.basedb.core.Job.ProgressReporterImpl} is used
		@param server The name of the server executing the plugin
		@return A <code>Plugin.Request</code> object
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission or the job isn't in the {@link Status#WAITING} status
		@throws InvalidDataException If the server name is too long
		@throws BaseException If there is another error
	*/
	public PluginExecutionRequest execute(ProgressReporter progress, String server)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		return execute(progress, server, null);
	}
	
	
	/**
		Start the execution sequence for a job. Call {@link DbControl#commit()}
		before calling {@link net.sf.basedb.core.PluginRequest#invoke()}. 
		See {@link net.sf.basedb.core.PluginRequest} for more  information. 

		@param progress The {@link ProgressReporter} where the plugin can report its progress. If 
			<code>null</code> {@link net.sf.basedb.core.Job.ProgressReporterImpl} is used
		@param server The name of the server executing the plugin
		@param settings The job agent settings to use or null to use default settings
		@return A <code>Plugin.Request</code> object
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission or the job isn't in the {@link Status#WAITING} status
		@throws InvalidDataException If the server name is too long
		@throws BaseException If there is another error
	*/
	public PluginExecutionRequest execute(ProgressReporter progress, String server, JobAgentSettings settings)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (getJobType() != Type.RUN_PLUGIN)
		{
			throw new PermissionDeniedException("Not allowed for a job of type '"+getJobType()+"': "+toString());
		}
		if (!getStatus().isExecutable()) 
		{
			throw new PermissionDeniedException("Cannot execute a job with status '"+getStatus()+"': "+toString());
		}
		if (server != null && server.length() > MAX_SERVER_LENGTH)
		{
			throw new StringTooLongException("server", server, MAX_SERVER_LENGTH);
		}
		
		PluginDefinition pd = getPluginDefinition();
		PluginConfiguration config = getPluginConfiguration();
		Plugin plugin = pd.newInstance(settings == null ? null : settings.getJarPath());
		JobAgent agent = settings == null ? null : settings.getJobAgent();
		SessionControl sc = getSessionControl();
		ParameterValuesImpl parameters = new ParameterValuesImpl(this, true);
		ProgressReporterImpl jobProgress = 
			new ProgressReporterImpl(this, server, agent, progress);
		
		PluginExecutionRequest request = 
			new PluginExecutionRequest(
				sc, plugin, Request.COMMAND_EXECUTE,
				config, config == null ? null : config.getParameterValuesImpl(getParameterVersion()),
				this, parameters, pd
			);
		request.setProgressReporter(jobProgress);
		return request;
	}

	/**
		Get the names of all parameters for the job.
		@return A <code>Set</code> containing the names of the parameters or an
			empty set if no parameters has been defined
	*/
	public Set<String> getParameterNames()
	{
		return Collections.unmodifiableSet(getData().getParameters().keySet());
	}
	
	/**
		Get the value of a named parameter. If the parameter has mulitple
		values only the first is returned.
		
		@param name The name of the parameter
		@return The value as an object, or null if the 
			parameter doesn't exist
		@throws PermissionDeniedException If the parameter contain values that
			the logged in user doesn't have read permission for
		@throws BaseException If there is another error
	*/
	public Object getParameterValue(String name)
		throws PermissionDeniedException, BaseException
	{
		List<?> values = getParameterValues(name);
		return values == null || values.size() == 0 ? null : values.get(0);
	}
	
	/**
		Get the values of a named parameter.
		@param name The name of the parameter
		@return A list of objects containing the values, or null if the 
			parameter doesn't exist
		@throws PermissionDeniedException If the parameter contain values that
			the logged in user doesn't have read permission for
		@throws BaseException If there is another error
	*/
	public List<?> getParameterValues(String name)
		throws PermissionDeniedException, BaseException
	{
		ParameterValueData<?> parameterValue = getData().getParameters().get(name);
		if (parameterValue == null) return null;
		return Collections.unmodifiableList(Values.getItemValues(getDbControl(), parameterValue.getValues()));
	}
	
	/**
		Get extra information about a parameter. The information includes the 
		label, description and values of the parameter.
		@param name The name of the parameter
		@return A ParameterInfo object or null if no parameter with the 
			specified name exists
		@throws PermissionDeniedException If the parameter contain values that
			the logged in user doesn't have read permission for
		@throws BaseException If there is another error
	*/
	public ParameterInfo getParameterInfo(String name)
		throws PermissionDeniedException, BaseException
	{
		ParameterValueData<?> parameterValue = getData().getParameters().get(name);
		if (parameterValue == null) return null;
		return new ParameterInfo(getDbControl(), parameterValue);
	}
	
	/**
		Set the value of a configuration parameter. This method can only be called
		if the job isn't already executing. This
		method also sets the status of the job to {@link Status#WAITING}.

		@param name The name of the parameter
		@param parameterType The type of the parameter
		@param value The value of the parameter
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission or if the job is executing
		@throws InvalidDataException If name is null or the new value doesn't
			validate against the parameter type
		@throws BaseException If there is another error
	*/
	public void setParameterValue(String name, ParameterType<?> parameterType, Object value)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		setParameterValue(name, null, null, parameterType, value);
	}
	
	/**
		Set the value of a configuration parameter. This method can only be called
		if the job isn't already executing. This
		method also sets the status of the job to {@link Status#WAITING}.
	
		@param name The name of the parameter
		@param label The label of the parameter (optional)
		@param description A description of the parameter (optional)
		@param parameterType The type of the parameter
		@param value The value of the parameter
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission or if the job is executing
		@throws InvalidDataException If name is null or the new value doesn't
			validate against the parameter type
		@throws BaseException If there is another error
	*/
	public void setParameterValue(String name, String label, String description, ParameterType<?> parameterType, Object value)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (name == null) throw new InvalidUseOfNullException("name");
		if (getStatus() == Status.EXECUTING) 
		{
			throw new PermissionDeniedException("Cannot configure an executing job: "+toString());
		}
		getData().setStatus(Status.WAITING.getValue());
		setParameterValuesInternal(name, label, description, parameterType, Arrays.asList(value), true);
	}
	
	/**
		Set the value of a configuration parameter. This method can only be called
		if the job isn't already executing. This
		method also sets the status of the job to {@link Status#WAITING}.
		
		@param name The name of the parameter
		@param parameterType The type of the parameter
		@param values A list containing the values
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission or if the job is
			executing
		@throws InvalidDataException If name is null or the new value doesn't
			validate against the parameter type
		@throws BaseException If there is another error
	*/
	public void setParameterValues(String name, ParameterType<?> parameterType, List<?> values)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		setParameterValues(name, null, null, parameterType, values);
	}
	
	/**
		Set the value of a configuration parameter. This method can only be called
		if the job is executed by an external program, and isn't already executing. This
		method also sets the status of the job to {@link Status#WAITING}.
		
		@param name The name of the parameter
		@param label The label of the parameter (optional)
		@param description A description of the parameter (optional)
		@param parameterType The type of the parameter
		@param values A list containing the values
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission or if the job is
			executing
		@throws InvalidDataException If name is null or the new value doesn't
			validate against the parameter type
		@throws BaseException If there is another error
	*/
	public void setParameterValues(String name, String label, String description, ParameterType<?> parameterType, List<?> values)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (name == null) throw new InvalidUseOfNullException("name");
		if (getStatus() == Status.EXECUTING) 
		{
			throw new PermissionDeniedException("Cannot configure an executing job: "+toString());
		}
		getData().setStatus(Status.WAITING.getValue());
		setParameterValuesInternal(name, label, description, parameterType, values, true);
	}
	
	/**
		Copy all parameter values from another job. Existing parameters with the same
		name will be overwritten.
		
		@param copyFrom The job to copy parameters from
		@throws InvalidDataException If the copyFrom parameter is null
		@throws PermissionDeniedException If the logged in user doesn't have write
			permission for this configuration
		@throws BaseException If there is another error
	*/
	@SuppressWarnings("unchecked")
	public void copyParametersFrom(Job copyFrom)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (copyFrom == null) throw new InvalidUseOfNullException("copyFrom");
		Map<String, ParameterValueData<?>> to = getData().getParameters();
		Map<String, ParameterValueData<?>> from = copyFrom.getData().getParameters();
	
		try
		{
			for (Map.Entry<String, ParameterValueData<?>> fromEntry : from.entrySet())
			{
				String toName = fromEntry.getKey();
				ParameterValueData<?> fromParameter = fromEntry.getValue();
				Class<? extends ParameterValueData> dataClass = fromParameter.getClass();
				if (org.hibernate.proxy.HibernateProxy.class.isAssignableFrom(dataClass))
				{
					// The 'fromParameter' is a proxy, we need the real data class
					dataClass = 
						(Class<? extends ParameterValueData>)org.hibernate.Hibernate.getClass(fromParameter);
				}
				ParameterValueData<?> toParameter = dataClass.newInstance();
				toParameter.setLabel(fromParameter.getLabel());
				toParameter.setDescription(fromParameter.getDescription());
				toParameter.replaceValues(fromParameter.getValues());
				to.put(toName, toParameter);
			}
		}
		catch (IllegalAccessException ex)
		{
			throw new BaseException(ex);
		}
		catch (InstantiationException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Set the values of a configuration parameter.
		@param name The name of the configuration parameter
		@param label The label of the parameter (optional)
		@param description A description of the parameter (optional)
		@param parameterType The type of the parameter
		@param values A list containing the new values, null or empty to remove the 
			configuration values
		@param validate If validation by {@link ParameterType#validate(String, List)}
			is needed or not
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If name is null or the new value doesn't
			validate against the parameter type
		@throws BaseException If there is another error
	*/
	void setParameterValuesInternal(String name, String label, String description, 
		ParameterType<?> parameterType, List<?> values, boolean validate)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		ParameterValueData old = null;
		if (values == null || values.size() == 0)
		{
			old = getData().getParameters().remove(name);
		}
		else
		{
			if (parameterType == null) throw new InvalidUseOfNullException("parameterType");
			if (validate) parameterType.validate(name, values);
			ParameterValueData<?> parameterValue = parameterType.newParameterValueData();
			parameterValue.setLabel(label);
			parameterValue.setDescription(description);
			parameterValue.replaceValues(Values.getDataValues(values));
			old = getData().getParameters().put(name, parameterValue);
		}
		if (old != null) HibernateUtil.deleteData(getDbControl().getHibernateSession(), old);
	}
	
	/**
		Get a write-protected ParameterValuesImpl object.
	*/
	ParameterValuesImpl getParameterValuesImpl()
		throws BaseException
	{
		return new ParameterValuesImpl(this, true);
	}

	
	/**
		Send a message to the owner of the job when the job has finished
	*/
	private void sendMessage()
	{
		MessageData message = new MessageData();
		JobData data = getData();
		message.setTo(data.getOwner());
		message.setFrom("SYSTEM");
		message.setTimeSent(new Date());
		message.setJob(data);
		message.setDescription(data.getStatusMessage());
		
		if (getStatus() == Job.Status.ERROR)
		{
			message.setName(net.sf.basedb.util.Values.trim("Error: " + getName(), 
				Message.MAX_NAME_LENGTH));
		}
		else
		{
			message.setName(net.sf.basedb.util.Values.trim("Done: " + getName(), 
				Message.MAX_NAME_LENGTH));
		}
		HibernateUtil.saveData(getDbControl().getHibernateSession(), message);
	}
	
	/**
		Implements the ProgressReporter interface and
		stores the percentage completed for the current job.
	*/
	static class ProgressReporterImpl
		implements ProgressReporter
	{
		/**
			Number of miliseconds between updates to the database.
		*/
		private static final long UPDATE_INTERVAL = 5 * 1000;

		/**
	 		Min percentage difference between updates to the database.
		*/
		private static final int UPDATE_PERCENT = 5;
		
		private final SessionControl sc;
		private final int jobId;
		private final String server;
		private final Job job;
		private final JobAgent agent;
		private final ProgressReporter chained;
		private long lastUpdate;
		
		private int lastValue;
		private int offset;
		
		ProgressReporterImpl(Job job, String server, JobAgent agent, ProgressReporter chained)
			throws BaseException
		{
			this.sc = job.getSessionControl();
			this.job = job;
			this.jobId = job.getId();
			this.server = server;
			this.agent = agent;
			this.chained = chained;
			this.lastUpdate = 0;
			this.lastValue = 0;
			this.offset = 0;
		}
		
		/*
			From the ProgressReporter interface
			---------------------------------------------
		*/
		public void display(int percent, String message)
		{
			// Do not update unless at least :
			// * some time has passed
			// * the plug-in has finished (percent == 100)
			// * progress has increased by at least 5%
			if (System.currentTimeMillis() - lastUpdate >= UPDATE_INTERVAL || 
					percent == 100 || percent - lastValue >= UPDATE_PERCENT)
			{
				// If the percent is negative and less then the last value, an offset will be calculated.
				// This will adjust the percent to a realistic value. 
				//Notice: It is only possible to give a value close to what expected. 
				if (percent < lastValue && percent < 0)
				{	
					if (Math.abs(percent) < Math.abs(lastValue))
					{
						offset = offset + 2*Math.abs(lastValue);
					}
					else
					{	
						offset = offset + 2*Math.abs(percent);
					}
				}
				lastValue = percent;		
				percent = percent + offset;
				
				lastUpdate = System.currentTimeMillis();
				DbControl dc = null;
				try
				{
					dc = sc.newDbControl();
					updateProgress(dc.getHibernateSession(), percent, message);
					dc.commit();
				}
				catch (Throwable ex)
				{
					ex.printStackTrace();
				}
				finally
				{
					if (dc != null) dc.close();
				}
			}
			if (chained != null) chained.display(percent, message);
		}
		
		public void append(String message)
		{
			// Do not update unless at least some time has passed
			if (System.currentTimeMillis() - lastUpdate > UPDATE_INTERVAL)
			{
				lastUpdate = System.currentTimeMillis();
				DbControl dc = null;
				try
				{
					dc = sc.newDbControl();
					appendProgress(dc.getHibernateSession(), message);
					dc.commit();
				}
				catch (Throwable ex)
				{}
				finally
				{
					if (dc != null) dc.close();
				}
			}
			if (chained != null) chained.append(message);
		}
		// -------------------------------------------
		
		private void updateProgress(org.hibernate.Session session, int percent, String statusMessage)
		{
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "UPDATE_JOB_STATUS");
			query.setInteger("percent", percent);
			query.setString("statusMessage", statusMessage);
			query.setInteger("id", jobId);
			HibernateUtil.executeUpdate(query);
		}
		
		private void appendProgress(org.hibernate.Session session, String statusMessage)
		{
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "APPEND_JOB_STATUS");
			query.setString("statusMessage", statusMessage);
			query.setInteger("id", jobId);
			HibernateUtil.executeUpdate(query);
		}
		
		void setConfigured(PluginResponse response)
		{
			JobData data = job.getData();
			data.setStatus(Job.Status.WAITING.getValue());
			data.setPercentComplete(0);
			data.setStarted(null);
			data.setEnded(null);
			data.setStatusMessage(null);
			data.setStackTrace(null);
			data.setServer(null);
			data.setJobAgentId(null);
			data.setDryRun(response.isDryRun());
			data.setEstimatedExecutionTime(response.getEstimatedExecutionTime().getValue());
		}
		
		void setStarted(Class<? extends SignalTransporter> signalTransporter, String signalId)
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				Job job = Job.getById(dc, jobId);
				if (signalTransporter != null && signalId != null)
				{
					job.setSignalTransporter(signalTransporter, signalId);
				}
				job.start("Starting...", server, agent);
				dc.commit();
			}
			catch (Throwable ex)
			{}
			finally
			{
				if (dc != null) dc.close();
			}
		}
		
		void setEnded(PluginResponse response)
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				Job job = Job.getById(dc, jobId);
				job.getData().setDryRun(response.isDryRun());

				if (response.getStatus() == Response.Status.ERROR)
				{
					job.doneError(response.getMessage(), response.getErrorList());
				}
				else
				{
					job.doneOk(response.getMessage());
					PluginDefinition pd = null;
					boolean isImportOrExport = false;
					try
					{
						pd = job.getPluginDefinition();
						MainType pluginMainType = pd.getMainType();
						isImportOrExport = pluginMainType.equals(MainType.EXPORT) || (pluginMainType.equals(MainType.IMPORT));
					}
					catch(PermissionDeniedException pex)
					{
						isImportOrExport = false;
					}
					job.setRemoved(job.getRemoveJobWhenFinished() && isImportOrExport);
				}
				dc.commit();
			}
			catch (Throwable ex)
			{}
			finally
			{
				if (dc != null) dc.close();
			}
		}
		
	}

	/**
		Wrapper for the signal transporter that intercepts the ABORT signal
		and changes the job's status to ABORTING. The ABORT signal is dealyed
		until after a successful commit.
		@since 2.6
	*/
	private class JobSignalTransporter
		implements SignalTransporter
	{

		private final SignalTransporter transporter;
		private Signal delayed;
		private JobSignalTransporter(SignalTransporter transporter)
		{
			this.transporter = transporter;
		}
		
		public Collection<Signal> getSupportedSignals()
		{
			return transporter.getSupportedSignals();
		}
		public void init(String params)
		{
			transporter.init(params);
		}
		public void send(Signal signal)
		{
			if (signal == Signal.ABORT)
			{
				getData().setStatus(Job.Status.ABORTING.getValue());
				delayed = signal;
			}
			else
			{
				transporter.send(signal);
			}
		}
		
		void onAfterCommit()
		{
			if (delayed != null) transporter.send(delayed);
		}
		
	}
	

	/**
		The type of job.
	*/
	public enum Type
	{
		/**
			The job is executed by a plugin.
		*/
		RUN_PLUGIN(1, "Plugin"),

		/**
			The job is executed by some other program.
		*/
		OTHER(2, "Other");

		/**
			Maps the integer that is stored in the database.
		*/
		private static final Map<Integer, Type> valueMapping = new HashMap<Integer, Type>();
		
		static
		{
			for(Type type : Type.values())
			{
				Type t = valueMapping.put(type.getValue(), type);
				assert t == null : "Another Job.Type with the value "+type.getValue()+" already exists";
			}
		}
		
		/**
			Get the <code>Job.Status</code> object when you know the integer code.
			
			@param value The integer value.
			@return The type object or null if the value is unknown
		*/
		public static Type fromValue(int value)
		{
			Type type = valueMapping.get(value);
			assert type != null : "type == null for value "+value;
			return type;
		}

		private final int value;
		private final String displayValue;
		private Type(int value, String displayValue)
		{
			this.value = value;
			this.displayValue = displayValue;
		}
		
		public String toString()
		{
			return displayValue;
		}
		
		/**
			Get the integer value that is used when storing a <code>Job.Status</code> 
			to the database.
			@return The intger value of this type 
		*/
		public int getValue()
		{
			return value;
		}
	}


	/**
		The status of a job.
	*/
	public enum Status
	{
		/**
			The job hasn't been configured yet.
		*/
		UNCONFIGURED(0, "Not configured", false),
	
		/**
			The job is waiting to be started.
		*/
		WAITING(1, "Waiting", true),

		/**
			The job is beeing prepared for execution.
		*/
		PREPARED(5, "Preparing", true),

		
		/**
			The job is executing.
		*/
		EXECUTING(2, "Executing", false),

		/**
			The {@link Signal#ABORT} signal has been sent to the
			job, but it has not yet responded.
		*/
		ABORTING(6, "Aborting", false),
		
		/**
			The job has finished successfully.
		*/
		DONE(3, "Done", false),

		/**
			The job has finished with an error.
		*/
		ERROR(4, "Error", false);
		
		/**
			Maps the integer that is stored in the database.
		*/
		private static final Map<Integer, Status> valueMapping = new HashMap<Integer, Status>();
		
		static
		{
			for(Status status : Status.values())
			{
				Status s = valueMapping.put(status.getValue(), status);
				assert s == null : "Another Job.Status with the value "+status.getValue()+" already exists";
			}
		}
		
		/**
			Get the <code>Job.Status</code> object when you know the integer code.
			
			@param value The integer value.
			@return The type object or null if the value is unknown
		*/
		public static Status fromValue(int value)
		{
			Status status = valueMapping.get(value);
			assert status != null : "status == null for value "+value;
			return status;
		}

		private final int value;
		private final String displayValue;
		private final boolean executable;
		private Status(int value, String displayValue, boolean executable)
		{
			this.value = value;
			this.displayValue = displayValue;
			this.executable = executable;
		}
		
		public String toString()
		{
			return displayValue;
		}
		
		/**
			Get the integer value that is used when storing a <code>Job.Status</code> 
			to the database.
			@return The intger value of this type 
		*/
		public int getValue()
		{
			return value;
		}
		
		/**
			If the {@link Job#execute(ProgressReporter, String)} method can be
			called when the job is in this status or not.
		*/
		public boolean isExecutable()
		{
			return executable;
		}
	}
	
	/**
		A rough estimate of the execution time of a job.
	*/
	public enum ExecutionTime
	{
		/**
			The job is fast and should not take more than a minute to execute.
		*/
		SHORTEST(0, "Shortest (< 1 minute)", 1, Thread.NORM_PRIORITY - 1),
	
		/**
			The job shouldn't take more than 10 minutes to execute.
		*/
		SHORT(1, "Short (1-10 minutes)", 1, Thread.NORM_PRIORITY - 1),

		/**
			The job shouldn't take less than 1 hour to execute. 
		*/
		MEDIUM(2, "Medium (< 1 hour)", 2, Thread.NORM_PRIORITY - 2),
		
		/**
			The job takes more than 1 hour to execute.
		*/
		LONG(3, "Long (> 1 hour)", 4, Thread.NORM_PRIORITY - 2);

		/**
			Maps the integer that is stored in the database.
		*/
		private static final Map<Integer, ExecutionTime> valueMapping = new HashMap<Integer, ExecutionTime>();
		
		static
		{
			for(ExecutionTime et : ExecutionTime.values())
			{
				ExecutionTime e = valueMapping.put(et.getValue(), et);
				assert e == null : "Another Job.ExecutionTime with the value "+et.getValue()+" already exists";
			}
		}
		
		/**
			Get the <code>Job.Status</code> object when you know the integer code.
			
			@param value The integer value.
			@return The type object or null if the value is unknown
		*/
		public static ExecutionTime fromValue(int value)
		{
			ExecutionTime et = valueMapping.get(value);
			assert et != null : "ExecutionTime == null for value "+value;
			return et;
		}

		private final int value;
		private final String displayValue;
		private final int defaultSlots;
		private final int defaultPriority;
		
		private ExecutionTime(int value, String displayValue, int defaultSlots, int defaultPriority)
		{
			this.value = value;
			this.displayValue = displayValue;
			this.defaultSlots = defaultSlots;
			this.defaultPriority = defaultPriority;
		}
		
		public String toString()
		{
			return displayValue;
		}
		
		/**
			Get the integer value that is used when storing a <code>Job.ExecutionTime</code> 
			in the database.
			@return The integer value of this type 
		*/
		public int getValue()
		{
			return value;
		}
		
		/**
			The default number of execution slots available for jobs with
			this execution time in the internal job queue.
		*/
		public int getDefaultSlots()
		{
			return defaultSlots;
		}
		
		/**
			The default thread priority to use when executing jobs in
			with this execution time.
		*/
		public int getDefaultPriority()
		{
			return defaultPriority;
		}
	}

}

