/*
	$Id: Metadata.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import net.sf.basedb.util.ToStringComparator;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.BatchableData;
import java.util.Date;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.HashMap;

import org.hibernate.metadata.ClassMetadata;
import org.hibernate.EntityMode;

/**
	Utility methods for retreiving metadata about items.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class Metadata<C>
{

	private static Set<Item> annotatableItems;
	private static Set<Item> shareableItems;
	private static Set<Item> removableItems;
	private static Set<Item> ownableItems;
	private static Set<Item> diskConsumableItems;
	private static Set<Item> fileStoreEnabledItems;

	private static boolean isInitialised = false;
	
	static synchronized void init()
	{
		if (isInitialised) return;
		annotatableItems = Collections.unmodifiableSet(getImplementingItems(Annotatable.class));
		shareableItems = Collections.unmodifiableSet(getImplementingItems(Shareable.class));
		removableItems = Collections.unmodifiableSet(getImplementingItems(Removable.class));
		ownableItems = Collections.unmodifiableSet(getImplementingItems(Ownable.class));
		diskConsumableItems = Collections.unmodifiableSet(getImplementingItems(DiskConsumable.class));
		fileStoreEnabledItems = Collections.unmodifiableSet(getImplementingItems(FileStoreEnabled.class));
		isInitialised = true;
	}
	
	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		isInitialised = false;
		annotatableItems = null;
		shareableItems = null;
		removableItems = null;
		ownableItems = null;
		diskConsumableItems = null;
		fileStoreEnabledItems = null;
	}

	/**
		Get the value of a property of an item. This method can be used
		in place of a regular <code>get</code> method of the item. Ie.
		the following method calls are equivalent:
		<pre class="code">
String name = item.getName();
String name = (String)Metadata.getProperty(item, "name");
</pre>
		You can even use dot notation to navigate to associated items:
		<pre class="code">
String quota = item.getOwner().getQuota().getName();
String quota = (String)Metadata.getProperty(item, "owner.quota.name");
</pre>

		@param item The item object to get the value from. Null is not allowed
		@param propertyName The name of the property to get
	 	@return The value as a java.lang.Object  
	 	@throws BaseException If something goes wrong.
	 	@throws InvalidUseOfNullException If required parameter is null
	*/
	@SuppressWarnings({"unchecked"})
	public static Object getPropertyValue(BasicItem item, String propertyName)
		throws BaseException
	{
		if (item == null) throw new InvalidUseOfNullException("item");
		
		Metadata md = getInstance(item.getClass());
		return md.getPropertyValue(item.getDbControl(), item, propertyName);
	}

	/**
		Get all {@link Item}:s implementing/extinding the specified 
		interface/class.
	*/
	private static Set<Item> getImplementingItems(Class<?> clazz)
	{
		Set<Item> items = new TreeSet<Item>(new ToStringComparator<Item>(false));
		for (Item i : Item.values())
		{
			Class<?> itemClass = i.getItemClass();
			if (itemClass != null && clazz.isAssignableFrom(itemClass))
			{
				items.add(i);
			}
		}
		return items;
	}
	
	/**
		Get a set containing all {@link Annotatable} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getAnnotatableItems()
	{
		return annotatableItems;
	}
	
	/**
		Get a set containing all {@link Shareable} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getShareableItems()
	{
		return shareableItems;
	}

	/**
		Get a set containing all {@link Removable} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getRemovableItems()
	{
		return removableItems;
	}
	
	/**
		Get a set containing all {@link Ownable} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getOwnableItems()
	{
		return ownableItems;
	}
	
	/**
		Get a set containing all {@link DiskConsumable} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getDiskConsumableItems()
	{
		return diskConsumableItems;
	}

	/**
		Get a set containing all {@link FileStoreEnabled} item types. 
		The set cannot be modified.
	*/
	public static Set<Item> getFileStoreEnabledItems()
	{
		return fileStoreEnabledItems;
	}
	
	
	/**
		Cache of Metadata instances. Only one instance for each type
		of item exists.
	*/
	private static final Map<String, Metadata<?>> instances = new HashMap<String, Metadata<?>>();
	
	/**
		Get the metadata instance for the specified type of item.
	 	@param <I> 
		@param itemClass The class of the item
	 	@return a new instance of this class. 
	 	@throws BaseException If a new instance could not be created. 
	*/
	@SuppressWarnings({"unchecked"})
	public static <I> Metadata<I> getInstance(Class<I> itemClass)
		throws BaseException
	{
		Item itemType = null;
		if (BasicItem.class.isAssignableFrom(itemClass))
		{
			itemType = Item.fromItemClass((Class<BasicItem>)itemClass);
		}
		else if (BasicData.class.isAssignableFrom(itemClass))
		{
			itemType = Item.fromDataClass((Class<BasicData>)itemClass);
		}
		if (itemType == null)
		{
			throw new BaseException("Class '"+itemClass+"' is not known as a persistent entity");
		}
		return getInstance(itemClass, itemType.getDataClass().getName());
	}

	/**
		Get the metadata instance for the item with the specified 
		entity name. Use this method where one class is mapped to
		several entities, for example RawData.
	 	@param <I> 
		@param clazz The class of the item, should be either a
			BasicItem or BasicData class
		@param entityName The entity name for the item, normally
			this is <code>clazz.getName()</code> but some entities,
			for example <code>RawData</code>, have multiple names
	 	@return a new instance of this class 
	 	@throws BaseException If a new instance could not be created.
	*/
	@SuppressWarnings({"unchecked"})
	public static <I> Metadata<I> getInstance(Class<I> clazz, String entityName)
		throws BaseException
	{
		Metadata<I> md = (Metadata<I>)instances.get(entityName);
		if (md == null)
		{
			synchronized (instances)
			{
				md = new Metadata<I>(entityName);
				instances.put(entityName, md);
			}
		}
		return md;
	}
	
	private final ClassMetadata classMetadata;
	private final String entityName;
	
	private Metadata(String entityName)
		throws BaseException
	{
		this.entityName = entityName;
		classMetadata = HibernateUtil.getClassMetadata(entityName);
	}
	
	/**
		Get the value of a property of an item. This method can be used
		in place of a regular <code>get</code> method of the item. Ie.
		the following method calls are equivalent:
		<pre class="code">
String name = item.getName();
String name = (String)Metadata.getProperty(item, "name");
</pre>
		You can even use dot notation to navigate to associated items:
		<pre class="code">
String quota = item.getOwner().getQuota().getName();
String quota = (String)Metadata.getProperty(item, "owner.quota.name");
</pre>

		<p>
		TODO - the implementation of this method is a mess. Can it be made
		more robust?

		@param dc The DbControl object used if database access is needed
		@param item The item object to get the value from. Null is not allowed.
		@param propertyName The name of the property to get. Null is not allowed.
	 	@return Property value as a java.lang.Object or 
	 		null if the property has no value. 
	 	@throws BaseException If getting the property value goes wrong.
	 	@throws InvalidUseOfNullException if any of the required arguments is null.
	*/
	public Object getPropertyValue(DbControl dc, C item, String propertyName)
		throws BaseException
	{
		if (item == null) throw new InvalidUseOfNullException("item");
		if (propertyName == null) throw new InvalidUseOfNullException("propertyName");
		String[] propertyPath = propertyName.split("\\.");
		ClassMetadata currentClassMetadata = classMetadata;
		String currentIdProperty = currentClassMetadata.getIdentifierPropertyName();
		org.hibernate.Session session = dc.getHibernateSession();
		Object value = null;
		Object data = null;
		if (item instanceof BasicItem)
		{
			data = ((BasicItem)item).getData();
		}
		else if (item instanceof BasicData)
		{
			data = (BasicData)item;
		}
		else if (item instanceof ReporterScore)
		{
			data = ((ReporterScore)item).getData();
		}
		else
		{
			throw new AssertionError("Item: " + item + " is not a BasicItem, BasicData or ReporterScore");
		}
		
		try
		{
			for (int i = 0; i < propertyPath.length; ++i)
			{
				if (value instanceof BasicData)
				{
					BasicData bData = (BasicData)value;
					data = value;
					Class<?> dataClass = null;
					// If it is a proxy we must reload it using the current session
					if (data instanceof org.hibernate.proxy.HibernateProxy)
					{
						dataClass = org.hibernate.proxy.HibernateProxyHelper.getClassWithoutInitializingProxy(data);
						data = HibernateUtil.loadData(session, dataClass, bData.getId());
					}
					else
					{
						dataClass = data.getClass();
					}
					currentClassMetadata = HibernateUtil.getClassMetadata(dataClass.getName());
					currentIdProperty = currentClassMetadata.getIdentifierPropertyName();
				}
				if (propertyPath[i].equals(currentIdProperty))
				{
					value = currentClassMetadata.getIdentifier(data, EntityMode.POJO);
				}
				else
				{
					value = currentClassMetadata.getPropertyValue(data, propertyPath[i], EntityMode.POJO);
				}
				if (value == null) return null;
			}
		}
		catch (Exception ex)
		{
			throw new BaseException(ex);
		}
		
		if (value instanceof BasicData)
		{
			if (value instanceof BatchableData)
			{
				HibernateUtil.evictData(dc.getHibernateSession(), (BasicData)value);
			}
			else
			{
				value = dc.getItem(BasicItem.class, (BasicData)value);
			}
		}

		// Must make a copy, or a user with READ permission may change the value
		else if (value instanceof Date)
		{
			value = DateUtil.copy((Date)value);
		}
		// We can only return collections if they don't contain BasicData object, 
		// since that would bypass permission checking
		else if (value instanceof Collection)
		{
			Collection<?> c = (Collection<?>)value;
			Object element = null;
			if (!c.isEmpty()) element = c.iterator().next();
			if (element instanceof BasicData)
			{
				throw new PermissionDeniedException(Permission.READ, "collection '"+propertyName+"' of "+item);
			}
			value = Collections.unmodifiableCollection(c);
		}
		else if (value instanceof Map)
		{
			Map<?, ?> m = (Map<?, ?>)value;
			Map.Entry<?, ?> entry = null;
			if (!m.isEmpty()) entry =  m.entrySet().iterator().next();
			if (entry != null && (entry.getKey() instanceof BasicData || entry.getValue() instanceof BasicData))
			{
				throw new PermissionDeniedException(Permission.READ, "collection '"+propertyName+"' of "+item);
			}
			value = Collections.unmodifiableMap(m);
		}
		return value;
	}

	public String toString()
	{
		return "Metadata for entity " + entityName;
	}
	
}
