/*
	$Id $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

/**
	Contains information about a single parameter that the
	plugin needs for a request.

	@author Nicklas
	@version 2.0
	@see RequestInformation
	@see net.sf.basedb.core.PluginRequest
	@base.modified $Date$
*/
public class PluginParameter<T>
{
	/**
		If the parameter should be hidden from user input or not.
		The parameter value will then be the default value.
	*/
	public final boolean hidden;
	
	/**
		The name of the parameter that the plugin use to
		identify the value with.
	*/
	private final String name;

	/**
		A short label used by the client application as in
		the GUI.
	*/
	private final String label;
	
	/**
		A longer help text explaining what the particular parameter
		is needed for.
	*/
	private final String description;

	/**
		A default value that overrides the default set by
		the parameter type.
	*/
	private final T defaultValue;
	
	/**
		Information of the parameter type.
	*/
	private final ParameterType<T> type;


	/**
		Constructor that sets all members except the default value. The default
		value is taken from the parameter types default value.
		@param name The name of the parameter, null is allowed for section headers
		@param label Label that can be used in the GUI for the parameter, or null
			to create a hidden parameter
		@param description A help text explaining what the parameter is used for
		@param type The parameter type, or null to create a section header
	*/
	public PluginParameter(String name, String label, String description, ParameterType<T> type)
	{
		this(name, label, description, null, type);
	}
	
	/**
		Constructor that sets all members. If a default value is specified it override
		the default value specified by the paramter type.
		
		@param name The name of the parameter, null is allowed for section headers
		@param label Label that can be used in the GUI for the parameter, or null
			to create a hidden parameter
		@param description A help text explaining what the parameter is used for
		@param defaultValue The default value of the parameter or null to use
			the default set by the parameter type
		@param type The parameter type, or null to create a section header
	*/
	public PluginParameter(String name, String label, String description, T defaultValue, ParameterType<T> type)
	{
		this.name = name;
		this.label = label;
		this.description = description;
		this.type = type;
		this.defaultValue = defaultValue;
		this.hidden = label == null;
	}

	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		return name;
	}
	// -------------------------------------------

	/**
		Check if the parameter is a hidden parameter or not. A hidden parameter should
		not be displayed for the user. The value should be set to the the parameter types
		default value, ie. <code>getParmeterType().getDefaultValue()</code>
		@return TRUE if the parameter is hidden, FALSE otherwise
	*/
	public boolean isHidden()
	{
		return hidden;
	}
	
	/**
		Get the parameter name. This is the name of the parameter that the plugin 
		use to identify the value with. This is the name the client application
		should use when calling {@link net.sf.basedb.core.PluginRequest#setParameterValue(String, Object)}.
		@return The name
	*/
	public String getName()
	{
		return name;
	}

	/**
		Get the label of the parameter. The label is a short string
		that a client application should display in the GUI.
		@return The label
	*/
	public String getLabel()
	{
		return label;
	}
	
	/**
		Get a description of the parameter. The description may be a longer
		help text explaining why the parameter is needed and what it is for.
		@return The label
	*/
	public String getDescription()
	{
		return description;
	}

	/**
		Get the default value for this parameter or null if no
		default has been specified
		@return The default value
	*/
	public T getDefaultValue()
	{
		return defaultValue != null ? defaultValue : type != null ? type.getDefaultValue() : null;
	}
	
	/**
		Get the parameter type, which tells the client application the
		data type and allowed values for the parameter.
		@return A <code>ParameterType</code> object
	*/
	public ParameterType<T> getParameterType()
	{
		return type;
	}

}
