/*
	$Id: RawDataTypes.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.PlatformData;
import net.sf.basedb.core.data.PlatformVariantData;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.XMLUtil;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.net.URL;
import java.util.Collections;
import java.util.Collection;

import org.jdom.Document;
import org.jdom.Element;

/**
	This class is used for reading the XML configuration file with 
	information about raw data types. All files returned by {@link 
	Application#getRawDataTypeFiles()} are parsed.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class RawDataTypes
{

	/**
		Log core events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.RawDataTypes");

	/**
		A map from raw data type name --> {@link RawDataType}.
	*/
	private static Map<String, RawDataType> rawDataTypes = null;

	private static Map<String, RawDataType> platformTypes = null;
	
	private static Set<RawDataType> all = null;
	
	static PlatformData generic = null;
	
	/**
		The DTD which is used to validate the XML file.
	*/
	private static final URL dtdFile = 
		RawDataTypes.class.getResource("/net/sf/basedb/core/dtd/raw-data-types.dtd");
	
	private static boolean isInitialised = false;

	/**
		Initialize the class.
		@throws BaseException This exception is thrown if there is an error
	*/
	static synchronized void init()
		throws BaseException
	{
		if (isInitialised) return;
		rawDataTypes = new TreeMap<String, RawDataType>();
		all = new TreeSet<RawDataType>();
		loadRawDataTypesFile();
		isInitialised = true;
	}
	
	/**
		Load file-only platforms and create raw data types for them.
		@since 2.5
	*/
	static synchronized void initPlatforms()
	{
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		
		try
		{
			platformTypes = new HashMap<String, RawDataType>();
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			org.hibernate.Query query = HibernateUtil.createQuery(session, 
				"FROM PlatformData pfd WHERE pfd.fileOnly = true");
			for (PlatformData platform : HibernateUtil.loadList(PlatformData.class, query, null))
			{
				RawDataType rdt = new RawDataType(platform);
				platformTypes.put(rdt.getId(), rdt);
				all.add(rdt);
			}
			
			query = HibernateUtil.createQuery(session, 
				"FROM PlatformVariantData pfv WHERE pfv.fileOnly = true");
			for (PlatformVariantData variant : HibernateUtil.loadList(PlatformVariantData.class, query, null))
			{
				RawDataType rdt = new RawDataType(variant);
				platformTypes.put(rdt.getId(), rdt);
				all.add(rdt);
			}

			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_PLATFORM_FOR_EXTERNAL_ID");
			query.setString("externalId", Platform.GENERIC);
			generic = HibernateUtil.loadData(PlatformData.class, query);
		}
		finally
		{
			if (tx != null) HibernateUtil.commit(tx);
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		isInitialised = false;
		if (rawDataTypes != null) rawDataTypes.clear();
		rawDataTypes = null;
		if (platformTypes != null) platformTypes.clear();
		platformTypes = null;
		if (all != null) all.clear();
		all = null;
	}


	/**
		Load and parse the file with raw data types.
		This method will populate the {@link #rawDataTypes} variable.
	*/
	private static synchronized void loadRawDataTypesFile()
		throws BaseException
	{
		List<String> files = Application.getRawDataTypeFiles();
		try
		{
			for (String xmlFile : files)
			{
				Document dom = XMLUtil.getValidatedXml(RawDataTypes.class.getResource(xmlFile), dtdFile);
				loadRawDataTypes(dom, xmlFile);
				log.info("Loaded raw data types from file: " + xmlFile);
			}
		}
		catch (Exception ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Get a collection with all raw data types that has been defined. This method
		only returns both database and file-only raw data types. To
		get file-only raw data types use {@link #getFileOnlyRawDataTypes()}
		@return A <code>Collection</code> containing {@link RawDataType} objects
			where {@link RawDataType#isStoredInDb()} returns true
	*/
	public static Collection<RawDataType> getRawDataTypes()
	{
		return Collections.unmodifiableCollection(all);
	}
	
	/**
		Get the raw data types that are representations of file-only
		{@link Platform}:s.
		@return A <code>Collection</code> containing {@link RawDataType} objects 
			where {@link RawDataType#isStoredInDb()} returns false
		@since 2.5
	*/
	public static Collection<RawDataType> getFileOnlyRawDataTypes()
	{
		return Collections.unmodifiableCollection(platformTypes.values());
	}

	/**
		Get a information about a raw data type.

		@param id The id of the raw data type
		@return A <code>RawDataType</code> object with information about the
			raw data type, or null if the specified id is not found
	*/
	public static RawDataType getRawDataType(String id)
	{
		if (id == null) return null;
		if (id.startsWith("platform."))
		{
			return getPlatformRawDataType(id);
		}
		else if (id.startsWith("variant."))
		{
			return getVariantRawDataType(id);
		}
		else
		{
			return rawDataTypes.get(id);
		}
	}
	
	/**
		Get a information about a raw data type or throw an exception if the
		raw data type is not found.
	
		@param id The id of the raw data type
		@return A <code>RawDataType</code> object with information about the
			raw data type
		@throws ItemNotFoundException If the specified raw data type is not found
		@since 2.6
	*/
	public static RawDataType getSafeRawDataType(String id)
	{
		RawDataType rdt = getRawDataType(id);
		if (rdt == null) throw new ItemNotFoundException("RawDataType[id=" + id + "]");
		return rdt;
	}
	
	private static RawDataType getPlatformRawDataType(String id)
	{
		RawDataType rdt = platformTypes.get(id);
		if (rdt != null) return rdt;
		
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_PLATFORM_FOR_EXTERNAL_ID");
			query.setString("externalId", id.substring(9));
			PlatformData platform = HibernateUtil.loadData(PlatformData.class, query);
			if (platform == null || !platform.isFileOnly()) return null;
			rdt = new RawDataType(platform);
			platformTypes.put(id, rdt);
			all.add(rdt);
		}
		finally
		{
			if (tx != null) HibernateUtil.commit(tx);
			if (session != null) HibernateUtil.close(session);
		}
		return rdt;
	}
	
	private static RawDataType getVariantRawDataType(String id)
	{
		RawDataType rdt = platformTypes.get(id);
		if (rdt != null) return rdt;
		
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_PLATFORMVARIANT_FOR_EXTERNAL_ID");
			query.setString("externalId", id.substring(8));
			PlatformVariantData variant = HibernateUtil.loadData(PlatformVariantData.class, query);
			if (variant == null || !variant.isFileOnly()) return null;
			rdt = new RawDataType(variant);
			platformTypes.put(id, rdt);
			all.add(rdt);
		}
		finally
		{
			if (tx != null) HibernateUtil.commit(tx);
			if (session != null) HibernateUtil.close(session);
		}
		return rdt;	
	}

	/**
		Load the raw data types and put the information in the
		{@link #rawDataTypes} map.
	*/
	@SuppressWarnings({"unchecked"})
	private static void loadRawDataTypes(Document dom, String xmlFile)
	{
		List<Element> rawDataTypeTags = dom.getRootElement().getChildren("raw-data-type");
		DbEngine engine = HibernateUtil.getDbEngine();
		Set<String> usedNames = new HashSet<String>();
		for (Element el : rawDataTypeTags)
		{
			String id = Values.getStringOrNull(el.getAttributeValue("id"));
			String name = Values.getStringOrNull(el.getAttributeValue("name"));
			String description = Values.getStringOrNull(el.getAttributeValue("description"));
			String storage = el.getAttributeValue("storage");
			String table = Values.getStringOrNull(el.getAttributeValue("table"));
			int channels = XMLUtil.getIntAttribute(el, "channels", 2);
			
			if (!ExtendedProperty.isValidName(id))
			{
				throw new InvalidDataException("Invalid id for raw data type '" + 
					id + "' in file '" + xmlFile + "'");
			}
			if (usedNames.contains("id:" + id))
			{
				throw new InvalidDataException("Duplicate id for raw data type '" + 
						id +"' in file '" + xmlFile);
			}
			if (usedNames.contains("name:" + name))
			{
				throw new InvalidDataException("Duplicate name for raw data type '" + 
						id +"' in file '" + xmlFile + "': name=" + name);
			}
			usedNames.add("id:" + id);
			usedNames.add("name:" + name);
			if ("database".equals(storage))
			{
				if (!engine.isValidTableName(table))
				{
					throw new InvalidDataException("Invalid table for raw data type '" + 
							id +"' in file '" + xmlFile + "': table=" + table);
				}
				if (usedNames.contains("table:" + table))
				{
					throw new InvalidDataException("Duplicate table for raw data type '" + 
							id +"' in file '" + xmlFile + "': table=" + table);
				}
				usedNames.add("table:" + table);
			}
			else
			{
				throw new ConfigurationException("Attribute storage='" + storage + 
					"' is no longer supported. Please remove declaration for <rawdatatype id='" + id + 
					"'> from file '" + xmlFile + "'");
			}
			if (channels <= 0)
			{
				throw new InvalidDataException("Number of channels must be > 0 for raw data type '"+
						id +"' in file '" + xmlFile + "': channels=" + channels);
			}
			List<RawDataProperty> properties = loadProperties(el, channels, xmlFile);
			List<IntensityFormula> formulas = loadIntensityFormulas(el, channels, xmlFile);
			RawDataType rdt = new RawDataType(id, name, description, channels, 
					table, properties, formulas);
			rawDataTypes.put(id, rdt);
			all.add(rdt);
		}
	}

	/**
		Load the properties for the specified raw data type node and
		return a list of {@link RawDataProperty} objects.
	*/
	@SuppressWarnings({"unchecked"})
	private static List<RawDataProperty> loadProperties(Element rawDataTypeElement, int channels, String xmlFile)
	{
		List<RawDataProperty> properties = new ArrayList<RawDataProperty>();
		List<Element> children = rawDataTypeElement.getChildren("property");
		String rawDataType = rawDataTypeElement.getAttributeValue("id");
		DbEngine engine = HibernateUtil.getDbEngine();
		Set<String> usedNames = new HashSet<String>();
		for (Element property : children)
		{
			String name = Values.getStringOrNull(property.getAttributeValue("name"));
			if (!ExtendedProperty.isValidName(name))
			{
				throw new InvalidDataException("Invalid property for raw data type '" + 
					rawDataType + "' in file '" + xmlFile + "': name=" + name);
			}
			if (usedNames.contains("name:" + name))
			{
				throw new InvalidDataException("Duplicate property for raw data type '" + 
					rawDataType + "' in file '" + xmlFile + "': name=" + name);
			}
			usedNames.add("name:" + name);
			String title = Values.getStringOrNull(property.getAttributeValue("title"));
			if (title == null) title = name;
			String column = Values.getStringOrNull(property.getAttributeValue("column"));
			if (!engine.isValidColumnName(column))
			{
				throw new InvalidDataException("Invalid column for raw data property '" + 
					rawDataType + "[" + name + "]' in file '" + xmlFile + "': column=" + column);
			}
			if (usedNames.contains("column:" + column))
			{
				throw new InvalidDataException("Duplicate column for raw data property '" + 
					rawDataType + "[" + name + "]' in file '" + xmlFile + "': column=" + column);
			}
			usedNames.add("column:" + column);
			String description = Values.getStringOrNull(property.getAttributeValue("description"));
			Type type = Type.fromValue(property.getAttributeValue("type"));
			int length = XMLUtil.getIntAttribute(property, "length", 255);
			if (type == Type.STRING && length > 255) type = Type.TEXT;
			boolean nullable = XMLUtil.getBooleanAttribute(property, "null", true);
			boolean averagable = XMLUtil.getBooleanAttribute(property, "averagable", type.isNumerical());
			String avgMethod = property.getAttributeValue("averagemethod");
			Formula.AverageMethod averageMethod = null;
			if (avgMethod != null) 
			{
				averageMethod = Formula.AverageMethod.valueOf(avgMethod.toUpperCase());
			}
			else if (averagable) 
			{
				averageMethod = Formula.AverageMethod.ARITHMETIC_MEAN;
			}
			else
			{
				averageMethod = Formula.AverageMethod.NONE;				
			}
			int channel = XMLUtil.getIntAttribute(property, "channel", 0);
			if (channel < 0 || channel > channels)
			{
				throw new InvalidDataException("Channel for raw data property '" + 
						rawDataType+ "[" + name + "]' in file '" + xmlFile + "'" +
						" must be >= 0 and < " + channels + ": channel=" + channel);
			}
			properties.add(new RawDataProperty(name, title, description, column, type, length, nullable, averageMethod, channel));
		}
		return properties;
	}
	
	/**
		Load the intensity formulas for the specified raw data type node and
		return a list of {@link IntensityFormula} objects.
	*/
	@SuppressWarnings({"unchecked"})
	private static List<IntensityFormula> loadIntensityFormulas(Element rawDataTypeElement, int channels, String xmlFile)
	{
		List<IntensityFormula> formulas = new ArrayList<IntensityFormula>();
		List<Element> children = rawDataTypeElement.getChildren("intensity-formula");
		String rawDataType = rawDataTypeElement.getAttributeValue("id");
		Set<String> usedNames = new HashSet<String>();
		for (Element formula : children)
		{
			String name = Values.getStringOrNull(formula.getAttributeValue("name"));
			if (!ExtendedProperty.isValidName(name))
			{
				throw new InvalidDataException("Invalid intensity formula for raw data type '" + 
					rawDataType + "' in file '" + xmlFile + "': name=" + name);
			}
			if (usedNames.contains("name:" + name))
			{
				throw new InvalidDataException("Duplicate intensity formula for raw data type '" + 
					rawDataType + "' in file '" + xmlFile + "': name=" + name);
			}
			usedNames.add("name:" + name);
			String title = Values.getStringOrNull(formula.getAttributeValue("title"));
			if (title == null) title = name;
			String description = Values.getStringOrNull(formula.getAttributeValue("description"));
			String[] expressions = new String[channels];
			for (Element expression :  (List<Element>)formula.getChildren("formula"))
			{
				int channel = XMLUtil.getIntAttribute(expression, "channel", 0);
				String exp = Values.getStringOrNull(expression.getAttributeValue("expression"));
				if (channel <= 0 || channel > channels)
				{
					throw new InvalidDataException("Channel for intensity formula '" + 
							rawDataType+ "[" + name + "]' in file '" + xmlFile + "'"  +
							" must be > 0 and < " + channels + ": channel=" + channel);
				}
				if (expressions[channel-1] != null)
				{
					throw new InvalidDataException("Duplicate expression for intensity formula '" + 
						rawDataType + "[" + name + "]' in file '" + xmlFile + "': channel=" + channel);
				}
				if (exp == null)
				{
					throw new InvalidDataException("Missing expression for formula '" + 
						rawDataType + "[" + name + "]' in file '" + xmlFile + "': channel=" + channel);
				}
				expressions[channel-1] = exp;
			}
			formulas.add(new IntensityFormula(name, title, description, expressions));
		}
		return formulas;
	}
	
}
