/*
	$Id: RealJoin.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import java.util.Collection;

import net.sf.basedb.core.query.Join;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.query.JoinType;

/**
	Represents a join from one virtual table in the dynamic database to
	a table in the regular database.
	<code>virtualTable vt JOIN regularTable rt ON rt.id = vt.leftColumn</code>.
	<p>
	Note that the join is always done to the id of the items in the regular table.

	@author Nicklas
	@version 2.0
	@see RealTable
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
class RealJoin
	implements Join
{
	
	private final JoinType joinType;
	private final VirtualTable left;
	private final RealTable right;
	private final VirtualColumn leftColumn;
	
	/**
		Create a new join to a regular table.
		<code><i>SELECT ... FROM left</i> JOIN right ON right.id = left.column</code>.

		@param joinType The type of join
		@param left The left table to join from, if not specified we join
			from the root table
		@param right The table to join to
		@param leftColumn The column in the left table to join from, we always
			join to the id of the real table
	*/
	RealJoin(JoinType joinType, VirtualTable left, RealTable right, VirtualColumn leftColumn)
	{
		this.joinType = joinType;
		this.left = left;
		this.right = right;
		this.leftColumn = leftColumn;
	}

	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String leftAlias = left == null ? query.getRootAlias() : left.getAlias();
			String rightAlias = right.getAlias();
			
			StringBuilder sb = new StringBuilder();
			sb.append(joinType.getJoinExpression()).append(" ").append(right.getQualifiedTableName());
			sb.append(" AS ").append(rightAlias);
			sb.append(" ON ");
			
			String qualifiedLeftColumn = HibernateUtil.quote(leftColumn.getColumn());
			sb.append(rightAlias).append(".id = ").append(leftAlias).append(".").append(qualifiedLeftColumn);
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("RealJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
	/**
		@return Always null
	*/
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -------------------------------------------
	/*
	 	From the Join interface
	 	-------------------------------------------
	*/
	public String getThetaJoin(Query query, DbControl dc) 
		throws BaseException 
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			StringBuilder sb = new StringBuilder();
			sb.append(", ").append(right.getQualifiedTableName());
			sb.append(" ").append(HibernateUtil.quote(right.getAlias()));
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ThetaJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
		
	public String getThetaJoinCondition(Query query, DbControl dc) 
		throws BaseException 
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String leftAlias = left == null ? query.getRootAlias() : HibernateUtil.quote(left.getAlias());
			String rightAlias = HibernateUtil.quote(right.getAlias());
			StringBuilder sb = new StringBuilder();

			String qualifiedColumn = HibernateUtil.quote(leftColumn.getColumn());
			sb.append(leftAlias).append(".").append(qualifiedColumn);
			sb.append(" = ").append(rightAlias).append(".").append(HibernateUtil.quote("id"));
			if (joinType == JoinType.LEFT) sb.append("(+)");
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ThetaJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	public String toString()
	{
		return joinType.getJoinExpression()+" "+right.getTableName()+" AS "+right.getAlias()+
			" ON "+right.getAlias() + ".id = " + (left == null ? "" : left.getAlias() + ".") + leftColumn.getName();
	}
	// -------------------------------------------

	
}
