/*
	$Id: ReporterList.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Predicate;

import net.sf.basedb.core.data.ReporterListData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.data.ReporterListScoreData;
import net.sf.basedb.info.ReporterListInfo;
import net.sf.basedb.info.ToTransferable;


/**

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class ReporterList
	extends CommonItem<ReporterListData>
		implements ToTransferable<ReporterListInfo>
{
	/**
		The type of item represented by this class.
		@see Item#REPORTERLIST
		@see #getType()
	*/
	public static final Item TYPE = Item.REPORTERLIST;
	
	/**
		The maximum length of the external id variable that can be stored in the
		database. Check the length against this value before calling the
		{@link #setExternalId(String)} method to avoid
		exceptions.
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = ReporterListData.MAX_EXTERNAL_ID_LENGTH;

	/**
		Create a new <code>ReporterList</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>Role</code> item
		@throws BaseException If there is an error
	*/
	public static ReporterList getNew(DbControl dc)
		throws BaseException
	{
		ReporterList rl = dc.newItem(ReporterList.class);
		rl.setName("New reporter list");
		return rl;
	}

	/**
		Get a <code>ReporterList</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>ReporterList</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static ReporterList getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ReporterList rl = dc.loadItem(ReporterList.class, id);
		if (rl == null) throw new ItemNotFoundException("ReporterList[id="+id+"]");
		return rl;
	}

	/**
		Get a new query object for this class.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<ReporterList> getQuery()
	{
		return new ItemQuery<ReporterList>(ReporterList.class);
	}

	/**
		Create a set with the intersection of the reporter ID:s 
		in the given lists.
		@return A set with reporter ID:s. The set is always a new
			set, modifications to it will not affect anything else
		@since 2.8
	*/
	private static Set<Integer> intersection(Collection<ReporterList> lists)
	{
		Set<Integer> intersection = new HashSet<Integer>();
		if (lists != null)
		{
			boolean first = true;
			Iterator<ReporterList> it = lists.iterator();
			while (it.hasNext())
			{
				Set<Integer> other = it.next().getData().getReporterListScores().keySet();
				if (first)
				{
					intersection.addAll(other);
					first = false;
				}
				else
				{
					intersection.retainAll(other);
				}
			}
		}
		return intersection;
	}
	
	/**
		Create a set with the union of the reporter ID:s 
		in the given lists.
		@return A set with reporter ID:s. The set is always a new
			set, modifications to it will not affect anything else
		@since 2.8
	*/
	private static Set<Integer> union(Collection<ReporterList> lists)
	{
		Set<Integer> union = new HashSet<Integer>();
		if (lists != null)
		{
			Iterator<ReporterList> it = lists.iterator();
			while (it.hasNext())
			{
				union.addAll(it.next().getData().getReporterListScores().keySet());
			}
		}
		return union;
	}
	
	/**
		Count the number of times each reporter is present in the
		given lists and return those reporters that are found in
		&gt=minCount and &lt;=maxCount lists.
		<p>
		NOTE! 
		<ul>
			<li>'minCount=1' and 'maxCount=number of lists' is the same
				as {@link #union(Collection)}
			<li>'minCount=maxCount=number of lists' is the same
				as {@link #intersection(Collection)}.
		</ul>

		@param lists The reporter lists to look in
		@param minCount The minimum number of list a reporter
			is required to be present in
		@param maxCount The maximum number of lists a reporter
			is allowed to the present in
		@return A set with reporter ID:s. It is always a new
			set, modifications to it will not affect anything else
		@since 2.8
	*/
	private static Set<Integer> count(Collection<ReporterList> lists, final int minCount, final int maxCount)
	{
		Map<Integer, Integer> union = new HashMap<Integer, Integer>();
		if (lists != null)
		{
			Iterator<ReporterList> it = lists.iterator();
			while (it.hasNext())
			{
				Iterator<Integer> e = it.next().getData().getReporterListScores().keySet().iterator();
				while (e.hasNext())
				{
					Integer id = e.next();
					Integer count = union.get(id);
					count = count == null ? 1 : count+1;
					union.put(id, count);
				}
			}
		}
		CollectionUtils.filter(union.values(), 
			new Predicate()
		{
			@Override
			public boolean evaluate(Object o)
			{
				int count = (Integer)o;
				return minCount <= count && count <= maxCount;
			}
		});
		return union.keySet();
	}
	
	ReporterList(ReporterListData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/**
		Get the external id for this <code>ReporterList</code> item.
		@return A string with the external id of this item
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	/**
		Set the external id for this <code>ReporterList</code> item. Null is allowed, but
		the value must not be longer than the value specified by
		the {@link #MAX_EXTERNAL_ID_LENGTH} constant. The value doesn't have to be unique.
		@param externalId The new external id for this item
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is longer than
			{@link #MAX_EXTERNAL_ID_LENGTH}
	*/
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNullableString(externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}

	/**
		Get the experiment this reporter list belongs to.
		@return An <code>Experiment</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Experiment getExperiment()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Experiment.class, getData().getExperiment());
	}
	
	/**
		Set the experiment. 
	 	@param experiment Experiment to set. Null is not allowed.
	 	@throws PermissionDeniedException If logged in user doesn't have write permission on the reporter list.
	 	@throws InvalidDataException If experiment is null.
	*/
	public void setExperiment(Experiment experiment)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (experiment != null) experiment.checkPermission(Permission.WRITE);
		getData().setExperiment(experiment == null ? null : experiment.getData());
	}

	/**
		Get the number of reporter in this reporter list.
		@since 2.8
	*/
	public int getSize()
	{
		return getData().getSize();
	}
	
	/**
		Add or update the score for a reporter.
	 
		@param reporter The reporter to add to the list
		@param score The score for the reporter, or null
		@return TRUE if the reporter was added, FALSE if it already existed
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the reporter is null
	 */
	public boolean addReporter(ReporterData reporter, Float score)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (reporter == null) throw new InvalidUseOfNullException("reporter");
		ReporterListScoreData rlScore = getData().getReporterListScores().get(reporter.getId());
		if (rlScore == null)
		{
			rlScore = new ReporterListScoreData();
			rlScore.setReporterList(this.getData());
			rlScore.setReporter(reporter);
			rlScore.setScore(score);
			getData().getReporterListScores().put(reporter.getId(), rlScore);
			getData().setSize(getSize() + 1);
			return true;
		}
		else
		{
			rlScore.setScore(score);
			return false;
		}
	}

	/**
		Remove a reporter from the reporter list.
		@param reporter The reporter to remove.
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the reporter is null
	 */
	public void removeReporter(ReporterData reporter)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (reporter == null) throw new InvalidUseOfNullException("reporter");
		if (getData().getReporterListScores().containsKey(reporter.getId()))
		{
			getData().setSize(getSize() - 1);
			getData().getReporterListScores().remove(reporter.getId());
		}
	}
	
	/**
		Add all reporters from another reporter list to this
		list. The new list is the union of the existing list
		and the added list. Scores on existing reporters are
		kept. Scores from new reporters are not transfered.
		
		@param list The reporter list to add reporters from
		@return The number of reporters added (not including 
			reporters that already was in this list)
		@since 2.8
	*/
	public int addAll(ReporterList list)
	{
		checkPermission(Permission.WRITE);
		return addReporters(list.getData().getReporterListScores().keySet());
	}
	
	/**
		Add reporters that are present in at least ONE OF the given lists to
		this list. Eg. the added reporters is the union from 
		the given lists. 
	
		@param lists The lists to add reporters from
		@return The number of reporters added (not including 
			reporters that already was in this list)
		@since 2.8
	*/
	public int addUnion(Collection<ReporterList> lists)
	{
		checkPermission(Permission.WRITE);
		int numAdded = 0;
		if (lists.size() == 1)
		{
			numAdded = addReporters(lists.iterator().next().getData().getReporterListScores().keySet());
		}
		else if (lists.size() > 1)
		{
			numAdded = addReporters(union(lists));
		}
		return numAdded;
	}
	
	/**
		Add reporters that are present in the specified number of lists to 
		this list.
		<ul>
		<li>Delegate to {@link #addUnion(Collection)} if
			<code>minCount &lt;= 1 && maxCount &gt;= size(lists)</code>
		<li>Delegate to {@link #addIntersection(Collection)}
			if <code>minCount = size(lists) && maxCount >= size(lists)</code>. 
		<li>If <code>minCount &gt; size(lists) || minCount > maxCount</code> no 
			reporters will be added.
		</ul>
		
		@param lists The lists to add reporters from
		@param minCount The minimum number of lists a reporter is required to
			be present in
		@param maxCount The maximum number of lists a reporter is allowed to 
			be present in
		@return The number of reporters added (not including 
			reporters that already was in this list)
		@since 2.8
	*/
	public int addIfPresentIn(int minCount, int maxCount, Collection<ReporterList> lists)
	{
		checkPermission(Permission.WRITE);
		int numAdded = 0;
		int numLists = lists.size();
		if (minCount <= 1 && maxCount >= numLists)
		{
			numAdded = addUnion(lists);
		}
		else if (minCount == numLists && maxCount >= numLists)
		{
			numAdded = addIntersection(lists);
		}
		else if (minCount < numLists && minCount <= maxCount)
		{
			numAdded = addReporters(count(lists, minCount, maxCount));
		}
		return numAdded;
	}
	
	/**
		Add reporters that are present in ALL OF the given lists to
		this list. Eg. the added reporters is the intersection from 
		the given lists. 

		@param lists The lists to add reporters from
		@return The number of reporters added (not including 
			reporters that already was in this list)
		@since 2.8
	*/
	public int addIntersection(Collection<ReporterList> lists)
	{
		checkPermission(Permission.WRITE);
		int numAdded = 0;
		if (lists.size() == 1)
		{
			numAdded = addReporters(lists.iterator().next().getData().getReporterListScores().keySet());
		}
		else if (lists.size() > 1)
		{
			numAdded = addReporters(intersection(lists));
		}
		return numAdded;
	}
	
	/**
		@since 2.8
	*/
	private int addReporters(Set<Integer> reporters)
	{
		int numAdded = 0;
		ReporterListData data = getData();
		Map<Integer, ReporterListScoreData> target = data.getReporterListScores();
		for (Integer id : reporters)
		{
			if (!target.containsKey(id))
			{
				ReporterListScoreData rlScore = new ReporterListScoreData();
				rlScore.setReporterList(data);
				rlScore.setReporterId(id);
				target.put(id, rlScore);
				numAdded++;
			}
		}
		data.setSize(getSize() + numAdded);
		return numAdded;
	}
	
	/**
		Remove all reporters from this list that are present
		in the other list. 
		
		@param list The reporter list that contains the reporters
			to remove
		@return The number of reporters removed from this list
		@since 2.8
	*/
	public int removeAll(ReporterList list)
	{
		checkPermission(Permission.WRITE);
		return removeReporters(list.getData().getReporterListScores().keySet());
	}
	
	/**
		Remove reporters that are present in at least ONE OF the given lists from
		this list. Eg. the removed reporters is the union from 
		the given lists. 
	
		@param lists The lists that contains the reporters to remove
		@return The number of reporters removed from this list
		@since 2.8
	*/
	public int removeUnion(Collection<ReporterList> lists)
	{
		checkPermission(Permission.WRITE);
		int numRemoved = 0;
		if (lists.size() == 1)
		{
			numRemoved = removeReporters(lists.iterator().next().getData().getReporterListScores().keySet());
		}
		else if (lists.size() > 1)
		{
			numRemoved = removeReporters(union(lists));
		}
		return numRemoved;
	}
	
	/**
		Remove reporters that are present in the specified number of lists from 
		this list.
		<ul>
		<li>Delegate to {@link #removeUnion(Collection)} if
			<code>minCount &lt;= 1 && maxCount &gt;= size(lists)</code>
		<li>Delegate to {@link #removeIntersection(Collection)}
			if <code>minCount = size(lists) && maxCount >= size(lists)</code>. 
		<li>If <code>minCount &gt; size(lists) || minCount > maxCount</code> no 
			reporters will be removed.
		</ul>
		
		@param lists The lists with the reporters
		@param minCount The minimum number of lists a reporter is required to
			be present in
		@param maxCount The maximum number of lists a reporter is allowed to 
			be present in
		@return The number of reporters removed
		@since 2.8
	*/
	public int removeIfPresentIn(int minCount, int maxCount, Collection<ReporterList> lists)
	{
		checkPermission(Permission.WRITE);
		int numRemoved = 0;
		int numLists = lists.size();
		if (minCount <= 1 && maxCount >= numLists)
		{
			numRemoved = removeUnion(lists);
		}
		else if (minCount == numLists && maxCount >= numLists)
		{
			numRemoved = removeIntersection(lists);
		}
		else if (minCount < numLists && minCount <= maxCount)
		{
			numRemoved = removeReporters(count(lists, minCount, maxCount));
		}
		return numRemoved;
	}
	
	/**
		Remove reporters that are present in ALL OF the given lists from
		this list. Eg. the removed reporters is the intersection from 
		the given lists. 
	
		@param lists The lists that contains the reporters to remove
		@return The number of reporters removed from this list
		@since 2.8
	*/
	public int removeIntersection(Collection<ReporterList> lists)
	{
		checkPermission(Permission.WRITE);
		int numRemoved = 0;
		if (lists.size() == 1)
		{
			numRemoved = removeReporters(lists.iterator().next().getData().getReporterListScores().keySet());
		}
		else if (lists.size() > 1)
		{
			numRemoved = removeReporters(intersection(lists));
		}
		return numRemoved;
	}
	
	/**
		@since 2.8
	*/
	private int removeReporters(Set<Integer> reporters)
	{
		int numRemoved = 0;
		ReporterListData data = getData();
		Map<Integer, ReporterListScoreData> target = data.getReporterListScores();
		
		// We loop over the list that is smallest
		if (reporters.size() <= this.getSize())
		{
			for (Integer id : reporters)
			{
				if (target.containsKey(id))
				{
					target.remove(id);
					numRemoved++;
				}
			}
		}
		else
		{
			Iterator<Integer> it = target.keySet().iterator();
			while (it.hasNext())
			{
				if (reporters.contains(it.next()))
				{
					it.remove();
					numRemoved++;
				}
			}
		}
		data.setSize(getSize() - numRemoved);
		return numRemoved;
	}
	
	/**
		Remove all reporters from this list that are NOT present
		in the other list. The new list is the intersection
		of the existing list and the other list.
		
		@param list The reporter list that contains the reporters
			to be retained
		@return The number of reporters removed from this list
		@since 2.8
	*/
	public int retainAll(ReporterList list)
	{
		checkPermission(Permission.WRITE);
		return retainReporters(list.getData().getReporterListScores().keySet());
	}
	
	/**
		Retain reporters that are present in at least ONE OF the given lists 
		in this list. Eg. reporters that are not found in any of the given
		lists are removed from this list.
	
		@param lists The lists that contains the reporters to retain
		@return The number of reporters removed from this list
		@since 2.8
	*/
	public int retainUnion(Collection<ReporterList> lists)
	{
		checkPermission(Permission.WRITE);
		int numRemoved = 0;
		if (lists.size() == 0)
		{
			getData().getReporterListScores().clear();
			numRemoved = getSize();
			getData().setSize(0);
		}
		else if (lists.size() == 1)
		{
			numRemoved = retainReporters(lists.iterator().next().getData().getReporterListScores().keySet());
		}
		else
		{
			numRemoved = retainReporters(union(lists));
		}
		return numRemoved;
	}
	
	/**
		Retain reporters that are present in ALL OF the given lists 
		in this list. Eg. reporters that are not found in all 
		lists of the given lists are removed from this list.
	
		@param lists The lists that contains the reporters to retain
		@return The number of reporters removed from this list
		@since 2.8
	*/
	public int retainIntersection(Collection<ReporterList> lists)
	{
		checkPermission(Permission.WRITE);
		int numRemoved = 0;
		if (lists.size() == 0)
		{
			getData().getReporterListScores().clear();
			numRemoved = getSize();
			getData().setSize(0);
		}
		else if (lists.size() == 1)
		{
			numRemoved = retainReporters(lists.iterator().next().getData().getReporterListScores().keySet());
		}
		else
		{
			numRemoved = retainReporters(intersection(lists));
		}
		return numRemoved;
	}
	
	/**
		Retain reporters that are present in the specified number of lists in 
		this list.
		<ul>
		<li>Delegate to {@link #retainUnion(Collection)} if
			<code>minCount &lt;= 1 && maxCount &gt;= size(lists)</code>
		<li>Delegate to {@link #retainIntersection(Collection)}
			if <code>minCount = size(lists) && maxCount >= size(lists)</code>. 
		<li>If <code>minCount &gt; size(lists) || minCount > maxCount</code> no 
			reporters will be removed.
		</ul>
		
		@param lists The lists with the reporter
		@param minCount The minimum number of lists a reporter is required to
			be present in
		@param maxCount The maximum number of lists a reporter is allowed to 
			be present in
		@return The number of reporters removed
		@since 2.8
	*/
	public int retainIfPresentIn(int minCount, int maxCount, Collection<ReporterList> lists)
	{
		checkPermission(Permission.WRITE);
		int numRemoved = 0;
		int numLists = lists.size();
		if (minCount <= 1 && maxCount >= numLists)
		{
			numRemoved = retainUnion(lists);
		}
		else if (minCount == numLists && maxCount >= numLists)
		{
			numRemoved = retainIntersection(lists);
		}
		else if (minCount < numLists && minCount <= maxCount)
		{
			numRemoved = retainReporters(count(lists, minCount, maxCount));
		}
		return numRemoved;
	}
	
	/**
		@since 2.8
	*/
	private int retainReporters(Set<Integer> reporters)
	{
		int numRemoved = 0;
		ReporterListData data = getData();
		Map<Integer, ReporterListScoreData> target = data.getReporterListScores();
		
		Iterator<Integer> it = target.keySet().iterator();
		while (it.hasNext())
		{
			if (!reporters.contains(it.next()))
			{
				it.remove();
				numRemoved++;
			}
		}
		data.setSize(getSize() - numRemoved);
		return numRemoved;
	}
	
	/**
		Get the score for the specified reporter.
		@param reporter The reporter
		@return The score as a float value, or null if the reporter isn't found 
			or hasn't been assigned a score
		@throws InvalidDataException
	 */
	public Float getScore(ReporterData reporter)
	{
		ReporterListScoreData score = getData().getReporterListScores().get(reporter.getId());
		return score == null ? null : score.getScore();
	}

	/**
	 	Get a query returning {@link ReporterScore} objects.
	*/
	public ReporterScoreQuery getReporterScores()
	{
		ReporterScoreQuery query = ReporterScore.getQuery(this);
		return query;
	}

	public ReporterListInfo toTransferable(ReporterListInfo info)
	{
		super.toTransferable(info);
		info.setExternalId(this.getExternalId());
		return info;
	}
	
}



