/*
	$Id: RoleKey.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/

package net.sf.basedb.core;

import net.sf.basedb.core.data.RoleKeyData;
import java.util.Set;

/**
	This class is used to represent a role key. A role key
	is, as the name suggests, a key that can be given to roles.
	<p>
	A BASE installation comes with a default set of role
	keys, one for every type of item. The keys cannot be deleted
	and no new ones can be created. The id of a role key for
	an item can be found by calling {@link SystemItems#getRoleKeyId(Item)}.
	<p>
	The permission given to a role is applied to all items
	of the type that the role key is valid for. The following
	example code will give all users the permission to create
	samples.

	<pre class="code">
Role r = Role.getById(dbControl, Role.USER);
RoleKey rk = RoleKey.getById(dbControl, SystemItems.getRoleKeyId(Item.SAMPLE));
rk.setPermission(r, Permission.CREATE);
dbControl.commit();
</pre>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class RoleKey
	extends Key<RoleKeyData>
	implements Nameable
{
	/**
		The type of item represented by this class.
		@see Item#ROLEKEY
		@see #getType()
	*/
	public static final Item TYPE = Item.ROLEKEY;

	/**
		This filter will only return items if the logged in user has
		generic read permission to roles.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = 
		new QueryRuntimeFilterFactory.ChildFilter(TYPE, Item.ROLE);

	/**
		Get a <code>RoleKey</code> item when you know the ID.
		The ID of a role key can be found by calling 
		{@link SystemItems#getRoleKeyId(Item)}.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>RoleKey</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static RoleKey getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		RoleKey rk = dc.loadItem(RoleKey.class, id);
		if (rk == null) throw new ItemNotFoundException("RoleKey[id="+id+"]");
		return rk;
	}
	
	/**
		Get a <code>RoleKey</code> item when you know the {@link Item} type.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param itemType The item type of the role key to load
		@return The <code>RoleKey</code> item
		@throws ItemNotFoundException If a role key for the specified 
			item type is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static RoleKey getByItemType(DbControl dc, Item itemType)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		return getById(dc, SystemItems.getRoleKeyId(itemType));
	}

	/**
		Get a query object configured to retrieve role keys. 
		If the logged in user doesn't have generic permission to roles, no
		role keys are returned.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<RoleKey> getQuery()
	{
		return new ItemQuery<RoleKey>(RoleKey.class, RUNTIME_FILTER);
	}

	RoleKey(RoleKeyData roleKeyData)
	{
		super(roleKeyData);
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setName(StringUtil.setNotNullString(name, "name", Nameable.MAX_NAME_LENGTH));
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setDescription(StringUtil.setNullableString(description, "description", Nameable.MAX_DESCRIPTION_LENGTH));
	}
	// -------------------------------------------
	
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Always return TRUE.
		All role keys are system items and cannot be deleted.
	*/
	public boolean isUsed()
		throws BaseException
	{
		return true;
	}
	/**
		CREATE and DELETE permissions are always denied. READ 
		permission is always granted. Other permissions are inherited 
		from the {@link Role}.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		denied |= Permission.deny(Permission.DELETE, Permission.CREATE);
		granted |= Permission.grant(Permission.READ) | getSessionControl().getRolePermissions(Item.ROLE);
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/**
		Get the {@link Item} type for the type of items this role key applies to. 
		@return An {@link Item} object
	*/
	public Item getItemType()
	{
		return Item.fromValue(getData().getItemType());
	}
	
	/**
		Grant permissions to a role. If {@link Permission#DENIED} is
		set it will override permissions gained from other keys. Ie.
		if a user has a role key with {@link Permission#DENIED} it takes precedence
		over all other permissions. Use an empty <code>Set</code> or null
		to remove all permissions for the role.

		@param role The <code>Role</code>
		@param permissions The permissions to grant
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#WRITE} permission for the role key
		@throws InvalidDataException If the role is null
		@see Permission
	*/
	public void setPermissions(Role role, Set<Permission> permissions)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (role == null) throw new InvalidUseOfNullException("role");
		if (permissions == null || permissions.isEmpty())
		{
			getData().getRoles().remove(role.getData());
		}
		else
		{
			getData().getRoles().put(role.getData(), Permission.grant(permissions));
		}
	}

	/**
		Get the permissions for a role.
		@param role The <code>Role</code> for which we want to get the permission
		@return An <code>EnumSet</code> containing the granted permissions, or an
			empty set if no permissions have been granted
		@throws InvalidDataException If the role is null
		@see Permission
	*/
	public Set<Permission> getPermissions(Role role)
		throws InvalidDataException
	{
		if (role == null) throw new InvalidUseOfNullException("role");
		return Permission.fromInt(IntegerUtil.getInt(getData().getRoles().get(role.getData())));
	}
}
