/*
	$Id: SpotImages.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.core.data.SpotImagesData;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.data.SharedData;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;
import java.util.zip.ZipException;
import java.io.IOException;
import java.io.ByteArrayOutputStream;

import java.awt.image.RenderedImage;
import java.awt.image.DataBuffer;
import java.awt.image.renderable.ParameterBlock;

import javax.media.jai.JAI;
import javax.media.jai.RenderedOp;
import javax.media.jai.TiledImage;
import javax.media.jai.LookupTableJAI;

import com.sun.media.jai.codec.TIFFDirectory;
import com.sun.media.jai.codec.TIFFDecodeParam;
import com.sun.media.jai.codec.SeekableStream;
import com.sun.media.jai.codec.JPEGEncodeParam;


/**
	Holds settings for creating small spot iamges from 1-3
	image files and the raw data information for a single raw bioassay.
	<p>
	For performance and disk space reasons the spot images are placed
	NxN (N given by the {@link #NUM_SPOT_IMAGES} constant) into JPEG
	files, which are put in a zip file.
	
	@author Nicklas
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
	@version 2.0
*/
public class SpotImages
	extends ChildItem<SpotImagesData>
{
	/**
		The type of item represented by this class.
		@see Item#SPOTIMAGES
		@see #getType()
	*/
	public static final Item TYPE = Item.SPOTIMAGES;

	/**
		The number of spot images to place side by side in each file.
		Each file contains NxN spot images.
	*/
	public static final int NUM_SPOT_IMAGES = 16;
	
	/**
		Create a new <code>Image</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param rawBioAssay The <code>RawBioAssay</code> to create spot images for
		@return The new <code>SpotImages</code> item
		@throws BaseException If there is an error
		@see RawBioAssay#getSpotImages()
	*/
	static SpotImages getNew(DbControl dc, RawBioAssay rawBioAssay)
		throws BaseException
	{
		SpotImages si = dc.newItem(SpotImages.class);
		si.setRawBioAssay(rawBioAssay);
		si.setGamma(1.8f);
		si.setQuality(0.8f);
		return si;
	}

	SpotImages(SpotImagesData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	SharedData getSharedParent()
	{
		return getData().getRawBioAssay();
	}
	// -------------------------------------------

	/**
		Get the raw bioassay the spot images belongs to.
		@return The <code>RawBioAssay</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission
		@throws BaseException If there is another error
	*/
	public RawBioAssay getRawBioAssay()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(RawBioAssay.class, getData().getRawBioAssay());
	}
	
	/**
		Set the raw bioassay the spot images belongs to. Can only be set 
		on a new item.
		@throws InvalidDataException If the raw bioassay is null
	*/
	void setRawBioAssay(RawBioAssay rawBioAssay)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (rawBioAssay == null) throw new InvalidUseOfNullException("rawBioAssay");
		getData().setRawBioAssay(rawBioAssay.getData());
	}

	/**
		Get the X-axis scale factor used to convert data coordinates to image
		pixel values: <code>xPixel = (xData - xOffset) / xScale</code>. The scale is
		always a positive integer.
	*/
	public int getXScale()
	{
		return getData().getXScale();
	}
	private void setXScale(int xScale)
		throws InvalidDataException
	{
		if (xScale <= 0) throw new NumberOutOfRangeException("xScale", xScale, 1, false);
		getData().setXScale(xScale);
	}

	/**
		Get the Y-axis scale factor used to convert data coordinates to image
		pixel values: <code>yPixel = (yData - yOffset) / yScale</code>. The scale is
		always a positive integer.
	*/
	public int getYScale()
	{
		return getData().getYScale();
	}
	private void setYScale(int yScale)
		throws InvalidDataException
	{
		if (yScale <= 0) throw new NumberOutOfRangeException("yScale", yScale, 1, false);
		getData().setYScale(yScale);
	}

	/**
		Get the X-axis offset used to convert data coordinates to image
		pixel values: <code>xPixel = (xData - xOffset) / xScale</code>. The offset may be
		positive or negative.
	*/
	public int getXOffset()
	{
		return getData().getXOffset();
	}
	private void setXOffset(int xOffset)
	{
		getData().setXOffset(xOffset);
	}

	/**
		Get the Y-axis offset used to convert data coordinates to image
		pixel values: <code>yPixel = (yData - yOffset) / yScale</code>. The offset may be
		positive or negative.
	*/
	public int getYOffset()
	{
		return getData().getYOffset();
	}
	private void setYOffset(int yOffset)
	{
		getData().setYOffset(yOffset);
	}

	/**
		Get the size of the spots in pixels. The size is a positive integer.
	*/
	public int getSpotsize()
	{
		return getData().getSpotsize();
	}
	private void setSpotsize(int spotsize)
		throws InvalidDataException
	{
		if (spotsize <= 0) throw new NumberOutOfRangeException("spotsize", spotsize, 1, false);
		getData().setSpotsize(spotsize);
	}

	/**
		Get the quality used for the jpeg compression. The quality is a number between
		0.0 and 1.0, with 1.0 indicating highest quality.
	*/
	public float getQuality()
	{
		return getData().getQuality();
	}
	private void setQuality(float quality)
		throws InvalidDataException
	{
		if (quality < 0.0 || quality > 1.0) throw new NumberOutOfRangeException("quality", quality, 0.0, 1.0);
		getData().setQuality(quality);
	}

	/**
		Get the gamma correction value. The value is always greater than or equal to 1.0.
	*/
	public float getGamma()
	{
		return getData().getGamma();
	}
	private void setGamma(float gamma)
		throws InvalidDataException
	{
		if (gamma < 1.0) throw new NumberOutOfRangeException("gamma", gamma, 1.0, false);
		getData().setGamma(gamma);
	}

	/**
		Get the file which was used for the red channel in the
		spot images.
		@return A file object or null if no file was used
		@throws PermissionDeniedException If the logged in user doesn't
			have read permission for the file
		@throws BaseException If there is another error
	*/
	public File getRedImageFile()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(File.class, getData().getRedImageFile());
	}
	private void setRedImageFile(File redImageFile)
		throws PermissionDeniedException
	{
		if (redImageFile != null) redImageFile.checkPermission(Permission.USE);
		getData().setRedImageFile(redImageFile == null ? null : redImageFile.getData());
	}
	
	
	/**
		Get the file which was used for the green channel in the
		spot images.
		@return A file object or null if no file was used
		@throws PermissionDeniedException If the logged in user doesn't
			have read permission for the file
		@throws BaseException If there is another error
	*/
	public File getGreenImageFile()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(File.class, getData().getGreenImageFile());
	}
	private void setGreenImageFile(File greenImageFile)
		throws PermissionDeniedException
	{
		if (greenImageFile != null) greenImageFile.checkPermission(Permission.USE);
		getData().setGreenImageFile(greenImageFile == null ? null : greenImageFile.getData());
	}

	/**
		Get the file which was used for the blue channel in the
		spot images.
		@return A file object or null if no file was used
		@throws PermissionDeniedException If the logged in user doesn't
			have read permission for the file
		@throws BaseException If there is another error
	*/
	public File getBlueImageFile()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(File.class, getData().getBlueImageFile());
	}
	private void setBlueImageFile(File blueImageFile)
		throws PermissionDeniedException
	{
		if (blueImageFile != null) blueImageFile.checkPermission(Permission.USE);
		getData().setBlueImageFile(blueImageFile == null ? null : blueImageFile.getData());
	}

	/**
		Get the file which holds the spot images. The file is a zip file
		with a number of images each holding several spots.
		
		@return A file object or null if no file exists
		@throws PermissionDeniedException If the logged in user doesn't
			have read permission for the file
		@throws BaseException If there is another error
	*/
	public File getSpotImagesFile()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(File.class, getData().getSpotImagesFile());
	}

	/**
		Disconnects the generated spot image file from this item. To delete
		the file you must also call the {@link DbControl#deleteItem(BasicItem)} method.
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
		@throws BaseException If there is an error
	*/
	public void disconnectSpotImages()
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		getData().setSpotImagesFile(null);
	}

	/**
		Creates spot images with 8 bits per channel, from source images. The spot images
		is group together, 16x16, into larger images. The larger images are stored in a
		zip file.

		@param spotImagesFile The file which will hold the created zip file with the spot images.
			Any existing content will be replaced
		@param xScale X-axis scale factor from data values to pixel values, must be > 0
		@param yScale Y-axis scale factor from data values to pixel values, must be > 0
		@param xOffset X offset in data value, can be any value
		@param yOffset Y offset in data value, can be any value
		@param spotsize The size of the spot i pixels, must be > 0
		@param gamma Gamma correction value, must be > 1.0
		@param quality The JPEG compression qualtity, must be a value between 0.0 and 1.0
		@param redImageFile The file containing the image to be used for the red channel
			or null
		@param greenImageFile The file containing the image to be used for the green channel
			or null
		@param blueImageFile The file containing the image to be used for the blue channel
			or null
		@param progress A {@link ProgressReporter} object if the calling client is
			interested in the progress or null if not

		@throws PermissionDeniedException If the logged in user doesn't have write
			permission for the raw bioassay or to the specified spot images file or doesn't have 
			use permission for any of the image files
		@throws InvalidDataException If the directory or file name is null or all image files are
			null, or any of the other values are out of range
		@throws ItemAlreadyExistsException If a file with the same name already exists in
			the directory
		@throws BaseException If there is another error
	*/
	public void createSpotImages(File spotImagesFile, int xScale, int yScale, int xOffset, int yOffset, 
		int spotsize, float gamma, float quality, File redImageFile, File greenImageFile, File blueImageFile,
		ProgressReporter progress)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		ThreadSignalHandler.checkInterrupted();
		checkPermission(Permission.WRITE);
		if (spotImagesFile == null) throw new InvalidUseOfNullException("spotImagesFile");
		spotImagesFile.checkPermission(Permission.WRITE);
		if (redImageFile == null && greenImageFile == null && blueImageFile == null)
		{
			throw new InvalidUseOfNullException("redImageFile, greenImageFile, blueImageFile");
		}
		setXScale(xScale);
		setYScale(yScale);
		setXOffset(xOffset);
		setYOffset(yOffset);
		setSpotsize(spotsize);
		setGamma(gamma);
		setQuality(quality);
		setRedImageFile(redImageFile);
		setGreenImageFile(greenImageFile);
		setBlueImageFile(blueImageFile);
		getData().setSpotImagesFile(spotImagesFile.getData());
		spotImagesFile.setFileType(FileType.getById(getDbControl(), SystemItems.getId(FileType.SPOT_IMAGES)));
		spotImagesFile.setMimeType("application/zip");

		// Keep track of which files we have so we don't have to do it for every spot
		boolean hasRed = redImageFile != null;
		boolean hasGreen = greenImageFile != null;
		boolean hasBlue = blueImageFile != null;

		try
		{
			// Set up the gamma correction table for 16 to 8 bit
			if (progress != null) progress.display(5, "Initialising...");
			byte[] gammaCorrection = new byte[0x10000];
			for (int i = 0; i < 0x10000; ++i) 
			{
				gammaCorrection[i] = (byte)Math.min(255.0, 255.0 * Math.pow(((double)i)/65535.0, gamma));
			}
			
			// Create an empty (black) image which we use for the spot images
			int widthHeight = spotsize * NUM_SPOT_IMAGES;
			Short[] blackColor = new Short[3];
			blackColor[0] = (short) 0;
			blackColor[1] = (short) 0;
			blackColor[2] = (short) 0;
			ParameterBlock blackImageParameters = new ParameterBlock();
			blackImageParameters.add(new Float(widthHeight)); // width
			blackImageParameters.add(new Float(widthHeight)); // height
			blackImageParameters.add(blackColor);
			RenderedOp blackImage = JAI.create("constant", blackImageParameters);
			TiledImage outputImage = new TiledImage(blackImage, widthHeight, widthHeight);
			outputImage.createGraphics();
	
			// Merge the three images into a single source image
			RenderedOp sourceImage = mergeColorImages();
			int imageWidth = sourceImage.getWidth();
			int imageHeight = sourceImage.getHeight();
			
			// Create the zip file that holds the spot images
			ZipOutputStream zipFileStream = new ZipOutputStream(spotImagesFile.getUploadStream(false));
			
			/*
				Get the raw data, sorted by position which is the way we want to
				store the spot images
			*/
			DataQuery<RawData> query = getRawBioAssay().getRawData();
			query.order(Orders.asc(Hql.property("position")));
			final Float floatSpotsize = new Float(spotsize);
			final float halfSpotsize = -spotsize / 2;
			final int spotImagesPerImage = NUM_SPOT_IMAGES * NUM_SPOT_IMAGES;
			final int totalSpots = getRawBioAssay().getNumDbSpots();
			int lastImageNumber = 0;

			// Iterate over each raw data spot and create spot images
			if (progress != null) progress.display(10, "Generating spot images (0 done)...");
			DataResultIterator<RawData> rawDataIterator = query.iterate(getDbControl());
			while (rawDataIterator.hasNext())
			{
				ThreadSignalHandler.checkInterrupted();
				RawData rawData = rawDataIterator.next();
	
				// Get the coordinates
				Float x = rawData.getX();
				Float y = rawData.getY();
				if (x == null || y == null) continue;
				
				// Get the image and tile number for the spot
				int position = rawData.getPosition();
				int tileNumber = position % spotImagesPerImage;
				int imageNumber = position / spotImagesPerImage;
				
				// If we move to a new image, the last one must be saved
				if (imageNumber != lastImageNumber)
				{
					if (progress != null) 
					{
						long spotsDone = spotImagesPerImage*imageNumber;
						int percent = 10 + (int)(90L * spotsDone / totalSpots);
						progress.display(percent, "Generating spot images ("+spotsDone+" done)...");
					}
					RenderedImage eightBitImage = convertTiffImage(outputImage, gammaCorrection);
					saveImage(zipFileStream, "spots"+lastImageNumber+".jpg", eightBitImage, quality);
					// Create a new black output image
					outputImage = new TiledImage(blackImage, widthHeight, widthHeight);
					outputImage.createGraphics();
					lastImageNumber = imageNumber;
				}
				
				// Cut around the spot for each raw data object using the "crop" operation
				float xSource = halfSpotsize + (0.5f + ((float)(x - xOffset)) / xScale);
				float ySource = halfSpotsize + (0.5f + ((float)(y - yOffset)) / yScale);
				if (insideImage(xSource, ySource, spotsize, imageWidth, imageHeight))
				{
					ParameterBlock cropParameters = new ParameterBlock();
					cropParameters.addSource(sourceImage);
					cropParameters.add(new Float(xSource));
					cropParameters.add(new Float(ySource));
					cropParameters.add(floatSpotsize);
					cropParameters.add(floatSpotsize);
					RenderedOp croppedImage = JAI.create("crop", cropParameters);
		
					// Rescale the colors in the spot image
					RenderedOp recoloredImage = rescaleColors(croppedImage, hasRed, hasGreen, hasBlue);
					
					// Move the spot to the correct location in the destination image using the "translate" operation
					ParameterBlock translateParameters = new ParameterBlock();
					translateParameters.addSource(recoloredImage);
					translateParameters.add(new Float(spotsize * (tileNumber % NUM_SPOT_IMAGES) - xSource));
					translateParameters.add(new Float(spotsize * (tileNumber / NUM_SPOT_IMAGES) - ySource));
					RenderedOp translatedImage = JAI.create("translate", translateParameters);
		
					// Copy the final image into the output image
					outputImage.set(translatedImage);
					
					// Clean up
					croppedImage.dispose();
					recoloredImage.dispose();
					translatedImage.dispose();
				}
			}
	
			if (progress != null) progress.display(100, "Generating spot images ("+totalSpots+" done)...");
			RenderedImage eightBitImage = convertTiffImage(outputImage, gammaCorrection);
			saveImage(zipFileStream, "spots"+lastImageNumber+".jpg", eightBitImage, quality);
			zipFileStream.close();
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
	}
	
	/**
		Get a spot as a PNG image. It is returned as a <code>SpotImage</code>.
		@param data A <code>RawData</code> object representing the spot
		@return <code>SpotImage</code> as a PNG image
		@throws InvalidDataException If the data object is null
		@throws ItemNotFoundException If a spot image for the position is not found
		@throws BaseException If there is another error
	*/
	public SpotImage getSpotImage(RawData data)
		throws InvalidDataException, ItemNotFoundException, BaseException
	{
		if (data == null) throw new InvalidUseOfNullException("data");

		ArrayList<RawData> l = new ArrayList<RawData>();
		l.add(data);
		List<SpotImage> images = getSpotImages(l);
		SpotImage spotImage = null;
		if (images.size() > 0) 
		{
			spotImage = images.get(0);
		}
		if (spotImage == null)
		{
			throw new ItemNotFoundException("SpotImage[postion="+data.getPosition()+"]");
		}
		return spotImage;
	}
	
	/**
		Get a list of spot PNG images from a list of <code>RawData</code>.
		It is returned as a list of <code>SpotImage</code> objects. The returned list
		will be mapped to the incomming list. If a spot image is not found, null will
		be put at its position in the list. That will make both list to always have the same size.
		It is more efficient having the list sorted by the <code>RawData</code>'s position.

		@param data A list of <code>RawData</code> objects representing the spots
		@return List of <code>SpotImage</code> objects as PNG images
		@throws InvalidDataException If the list of data is null
		@throws ItemNotFoundException If the file with the spot images is not found
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the file continaing the spot images
		@throws BaseException If there is another error
	*/
	public List<SpotImage> getSpotImages(List<RawData> data)
		throws InvalidDataException, ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (data == null)  throw new InvalidUseOfNullException("data");
		File spotImagesFile = getSpotImagesFile();
		if (spotImagesFile == null) throw new ItemNotFoundException("No spot images has been generated yet.");
		
		// Array to hold the generated spot images object
		ArrayList<SpotImage> spotImages = new ArrayList<SpotImage>(data.size());
		
		ZipFile zipFile = null;
		try
		{
			// The zip file containing the files with NxN spot images
			zipFile = new ZipFile(new java.io.File(Application.getUserFilesDirectory(), spotImagesFile.getData().getInternalName()));
			int lastImageNumber = -1;
			// Holds a single NxN image
			RenderedOp image = null;
			final int spotsize = getSpotsize();
			final int spotImagesPerImage = NUM_SPOT_IMAGES * NUM_SPOT_IMAGES;
			for (RawData rawData : data)
			{
				if (rawData == null)
				{
					spotImages.add(null);
				}
				else
				{
					// Get the image and tile number for the spot
					int position = rawData.getPosition();
					int tileNumber = position % spotImagesPerImage;
					int imageNumber = position / spotImagesPerImage;
					
					// Open another image if needed
					if (imageNumber != lastImageNumber)
					{
						// Get the zip file entry of the image
						ZipEntry zipEntry = zipFile.getEntry("spots" + imageNumber + ".jpg" );
						if (zipEntry == null)
						{
							// If it doesn't exist
							image = null;
						}
						else
						{
							// Open it as RenderedOp object
							SeekableStream stream = SeekableStream.wrapInputStream(zipFile.getInputStream(zipEntry), true);
							ParameterBlock jpegParameters = new ParameterBlock();
							jpegParameters.add(stream);
							image = JAI.create("jpeg", jpegParameters);
						}
						lastImageNumber = imageNumber;
					}
					
					if (image == null)
					{
						spotImages.add(null);
					}
					else
					{
						ParameterBlock cropParameters = new ParameterBlock();
						cropParameters.addSource(image);
						float xCoord = spotsize * (tileNumber % NUM_SPOT_IMAGES);
						float yCoord = spotsize * (tileNumber / NUM_SPOT_IMAGES);
						cropParameters.add(new Float(xCoord));
						cropParameters.add(new Float(yCoord));
						cropParameters.add(new Float(spotsize));
						cropParameters.add(new Float(spotsize));
						RenderedOp croppedImage = JAI.create("crop", cropParameters);

						// Move image to origin
						ParameterBlock translateParameters = new ParameterBlock();
						translateParameters.addSource(croppedImage);
						translateParameters.add(new Float(-xCoord));
						translateParameters.add(new Float(-yCoord));
						RenderedOp translatedImage = JAI.create("translate", translateParameters);

						// Encode to PNG
						ByteArrayOutputStream pngStream = new ByteArrayOutputStream(2048);
						ParameterBlock pngParameters = new ParameterBlock();
						pngParameters.addSource(translatedImage);
						pngParameters.add(pngStream);
						pngParameters.add("png");
						JAI.create("encode", pngParameters); 
						spotImages.add(new SpotImage(pngStream.toByteArray()));
					}
				}
			}
		}
		catch (ZipException ex)
		{
			throw new BaseException(ex);
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			try
			{
				if (zipFile != null) zipFile.close();
			}
			catch (IOException ex)
			{}
		}
		return spotImages;
	}

	/**
		Merge the three color images into a single image. If any of the
		color files are missing an empty band will be used for that color.
		@return The merged image
	*/
	private RenderedOp mergeColorImages()
		throws BaseException
	{
		// The width and height of the image(s)
		int width = 0;
		int height = 0;

		// Holds the three source TIFF images
		RenderedOp[] sourceImages = new RenderedOp[3];

		int numImages = 0;
		File red = getRedImageFile();
		File green = getGreenImageFile();
		File blue = getBlueImageFile();
		
		/*
			Keep track of which band we have used from each file
			so we for example can use the same three channel TIFF file 
			for all colors
		*/
		Map<File, Integer> nextBand = new HashMap<File, Integer>();
		nextBand.put(red, 0);
		nextBand.put(green, 0);
		nextBand.put(blue, 0);
		
		// Load the red image file
		if (red != null)
		{
			int band = nextBand.get(red);
			sourceImages[0] = loadTiffImage(red, band);
			nextBand.put(red, band+1);
			if (numImages == 0)
			{
				width = sourceImages[0].getWidth();
				height = sourceImages[0].getHeight();
				numImages++;
			}
		}
		
		// Load the green image file
		if (green != null)
		{
			int band = nextBand.get(green);
			sourceImages[1] = loadTiffImage(green, band);
			nextBand.put(green, band+1);
			if (numImages == 0)
			{
				width = sourceImages[1].getWidth();
				height = sourceImages[1].getHeight();
				numImages++;
			}
		}

		// Load the blue image file
		if (blue != null)
		{
			int band = nextBand.get(blue);
			sourceImages[2] = loadTiffImage(blue, band);
			nextBand.put(blue, band+1);
			if (numImages == 0)
			{
				width = sourceImages[2].getWidth();
				height = sourceImages[2].getHeight();
				numImages++;
			}
		}

		// Create a black band to to use for the missing channels, if any
		RenderedOp blackImage = null;
		if (numImages != 3)
		{
			Short[] blackColor = new Short[1];
			blackColor[0] = (short) 0;
			ParameterBlock blackParameters = new ParameterBlock();
			blackParameters.add(new Float(width));
			blackParameters.add(new Float(height));
			blackParameters.add(blackColor);
			blackImage = JAI.create("constant", blackParameters);
		}
		
		// Do the merge
		ParameterBlock mergeParameters = new ParameterBlock();
		for (int i = 0; i < 3; ++i)
		{
			if (sourceImages[i] != null)
			{
				mergeParameters.addSource(sourceImages[i]);
			}
			else
			{
				mergeParameters.addSource(blackImage);
			}
		}
		RenderedOp image = JAI.create("bandmerge", mergeParameters);
		return image;
	}

	
	/**
		Rescale the colors of the images so we use the whole 16-bit color range. 
		The empty bands are not include in the calculation.
		@param image Image to rescale
		@param red TRUE if the red channel isn't empty and should be included in the calculation
		@param green TRUE if green red channel isn't empty and should be included in the calculation
		@param blue TRUE if the blue channel isn't empty and should be included in the calculation
		@return Rescaled image
	*/
	private RenderedOp rescaleColors(RenderedOp image, boolean red, boolean green, boolean blue)
	{
		double max = Double.MIN_VALUE;
		double min = Double.MAX_VALUE;

		// Find the max and min intensity values in the image using the "extrema" operation
		ParameterBlock extremaParameters = new ParameterBlock();
		extremaParameters.addSource(image);
		RenderedOp extrema = JAI.create("extrema", extremaParameters);

		// First index: 0 =  min value, 1 = max value
		// Second index: 0, 1, 2 - color channel
		double[][] minMax = (double[][])extrema.getProperty("extrema");
		if (red )
		{
			max = Math.max(minMax[1][0], max);
			min = Math.min(minMax[0][0], min);
		}
		
		if (green)
		{
			max = Math.max(minMax[1][1], max);
			min = Math.min(minMax[0][1], min);
		}
		
		if (blue)
		{
			max = Math.max(minMax[1][2], max);
			min = Math.min(minMax[0][2], min);
		}

		// Calculate the scale and offset to strech the current max and min between 0 and 65535
		double scale = 65535.0 / (max - min);
		double offset = 65535.0 * min / (min - max);
		double[] scales = new double[3];
		double[] offsets = new double[3];
		// Each color channel can have individual scales and offsets but we want to use the same for all
		if (red)
		{
			scales[0] = scale;
			offsets[0] = offset;
		}
		if (green)
		{
			scales[1] = scale;
			offsets[1] = offset;
		}
		if (blue)
		{
			scales[2] = scale;
			offsets[2] = offset;
		}
		
		// Rescale the image
		ParameterBlock rescaleParameters = new ParameterBlock();
		rescaleParameters.addSource(image);
		rescaleParameters.add(scales);
		rescaleParameters.add(offsets);
		return JAI.create("rescale", rescaleParameters);
	}

	/**
		Loads a TIFF file and returns a <code>RenderedOp</code>.
		@param file A Base <code>File</code> object pointing to the image to be loaded
		@param band The index of the TIFF directory to load
	*/
	private RenderedOp loadTiffImage(File file, int band)
		throws BaseException
	{
		/*
			Create an input stream from the specified file to be
			used with the TIFF decoder.
		*/
		SeekableStream stream = SeekableStream.wrapInputStream(file.getDownloadStream(0), true);

		/*
			Store the input stream in a ParameterBlock to be sent to
			the operation registry, and eventually to the TIFF
			decoder
		*/
		ParameterBlock params = new ParameterBlock();
		params.add(stream);

		/*
			Specify the TIFF decoder to decode images as they are and
			not to convert unsigned short images to byte images.
		*/
		TIFFDecodeParam decodeParam = new TIFFDecodeParam();
		decodeParam.setDecodePaletteAsShorts(true);

		// Find the offset of the specified band in the TIFF file
		try
		{
			TIFFDirectory tiffDir = new TIFFDirectory(stream, band);
			decodeParam.setIFDOffset(tiffDir.getIFDOffset());
		}
		catch (IOException ex)
		{
			throw new BaseException(ex);
		}
		
		// Create a rendering operator to decode the TIFF file.
		params.add(decodeParam);
		RenderedOp image = JAI.create("tiff", params);
		return image;
	}

	/**
		Convert TIFF image from 16 bits to 8 bits if needed
		@param image TIFF image to convert
		@param conversionTable Conversion table
		@return A <code>RenderedOp</code> object of the image
		@throws BaseException If there is an error
	*/
	private RenderedImage convertTiffImage(RenderedImage image, byte[] conversionTable)
		throws BaseException
	{
		RenderedImage convertedImage = null;
		// Get the data type which tells us the number of bits per color
		int dataType = image.getSampleModel().getDataType();
		if (dataType == DataBuffer.TYPE_BYTE) // Already 8 bits
		{
			convertedImage = image;
		} 
		else if (dataType == DataBuffer.TYPE_USHORT) // Convert 16->8 bits
		{
			LookupTableJAI table = new LookupTableJAI(conversionTable);
			convertedImage = JAI.create("lookup", image, table);
		}
		else 
		{
			throw new BaseException("Can't convert TIFF image. Unknown data type: "+dataType);
		}
		return convertedImage;
	}

	/**
		Saving an image as jpeg and putting it in a zip file
		@param zipFileStream Destination stream to where the JPEG is saved to
		@param fileName The filename of the image
		@param image The image to be saved
		@param quality The quality setting of the JPEG image
	*/
	private void saveImage(ZipOutputStream zipFileStream, String fileName, RenderedImage image, float quality)
		throws IOException
	{
		zipFileStream.putNextEntry(new ZipEntry(fileName));
		JPEGEncodeParam jpegParam = new JPEGEncodeParam();
		jpegParam.setQuality(quality);
		ParameterBlock encodeParameters = new ParameterBlock();
		encodeParameters.addSource(image);
		encodeParameters.add(zipFileStream);
		encodeParameters.add("jpeg");
		encodeParameters.add(jpegParam);
		JAI.create("encode", encodeParameters); 
	}
	
	/**
		Check if entire spot is inside the soure image.
		@param x Spot x coordinate
		@param y Spot y coordinate
		@param spotsize Spot width and height
		@param width Source image width
		@param height Source image height
	 */
	private boolean insideImage(float x, float y, int spotsize, int width, int height)
	{
		return (x >= 0) && ((x+spotsize) <= width) && (y >= 0) && ((y+spotsize) <= height);
	}
	
}
