/*
	$Id: Transformation.java 4479 2008-09-05 15:16:27Z jari $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.TransformationData;
import net.sf.basedb.core.data.RawBioAssayData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

import java.util.Collection;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;

/**
	A transformation represents an analysis step for an {@link Experiment}. There
	are two types of transformations. A root transformation is used to calculate
	the initial set of intensities from a set of {@link RawBioAssay}:s. The result
	of a transformation is usually one or more {@link BioAssaySet}:s. The other
	type of transformation takes a bioassayset as the source and recalculates the
	intensities, applies a filter or does something else.
	<p>
	A transformation is created by either the {@link Experiment#newTransformation(Job, Collection)}
	method or the {@link BioAssaySet#newTransformation(Job)} method.
	
	@author Nicklas
	@version 2.0
	@see Experiment#newTransformation(Job, Collection)
	@see BioAssaySet#newTransformation(Job)
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class Transformation
	extends ChildItem<TransformationData>
	implements Nameable, Removable
{
	/**
		The type of item represented by this class.
		@see Item#TRANSFORMATION
		@see #getType()
	*/
	public static final Item TYPE = Item.TRANSFORMATION;

	/**
		Create a new <code>Transformation</code> item. No source
		has been specified. A caller should use either:
		{@link #setSource(BioAssaySet)} or {@link #setRawSources(Collection)}.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param experiment The experiment 
		@param job The job that executed the transformation
		@return The new <code>Transformation</code> item
		@throws BaseException If there is another error
		@see Experiment#newTransformation(Job, Collection)
		@see BioAssaySet#newTransformation(Job)
	*/
	static Transformation getNew(DbControl dc, Experiment experiment, Job job)
		throws BaseException
	{
		Transformation t = dc.newItem(Transformation.class);
		t.setName("New transformation");
		t.setExperiment(experiment);
		t.setJob(job);
		return t;
	}

	/**
		Get a <code>Transformation</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Transformation</code> item
		@throws ItemNotFoundException If an item with the specified id 
			is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Transformation getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Transformation t = dc.loadItem(Transformation.class, id);
		if (t == null) throw new ItemNotFoundException("Transformation[id=" + id + "]");
		return t;
	}
	
	/**
		Get a query configured to retrieve transformations for
		a given experiment.
		@param experiment The experiment to retreive transformations for, null
			is not allowed
		@return An {@link ItemQuery} object
	 	@throws InvalidDataException If required parameter is null. 
	*/
	public static ItemQuery<Transformation> getQuery(Experiment experiment)
		throws InvalidDataException
	{
		if (experiment == null) throw new InvalidUseOfNullException("experiment");
		ItemQuery<Transformation> query = new ItemQuery<Transformation>(Transformation.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("experiment"),
				Hql.entity(experiment)
			)
		);
		return query;
	}

	/**
		Get a query configured to retrieve transformations using the
		given bioassayset as source.
		
		@param bioAssaySet The source bioassayset to retreive transformations for, null
			is not allowed
		@return An {@link ItemQuery} object
	 	@throws InvalidDataException If required parameter is null. 
	*/
	public static ItemQuery<Transformation> getQuery(BioAssaySet bioAssaySet)
		throws InvalidDataException
	{
		if (bioAssaySet == null) throw new InvalidUseOfNullException("bioAssaySet");
		ItemQuery<Transformation> query = new ItemQuery<Transformation>(Transformation.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("source"),
				Hql.entity(bioAssaySet)
			)
		);
		return query;
	}

	// Keep reference to job so we can call Job.setExperiment at end of transaction
	private Job job;
	
	/**
	 	Creates a new experiment item from the given data.
		@param data the data
	*/
	Transformation(TransformationData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------

	/*
		From BasicItem
		-------------------------------------------
	*/
	/**
		Delete the product bioassaysets when deleting this transction.
		Set the experiment for the job when creating this transction.
	*/
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Transactional.Action.DELETE)
		{
			deleteProducts();
		}
		else if (action == Transactional.Action.CREATE)
		{
			if (job != null)
			{
				getDbControl().refreshItem(job);
				job.setExperiment(getExperiment());
			}
		}
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	/**
		Get the experiment.
	*/
	SharedData getSharedParent()
	{
		return getData().getExperiment();
	}
	/**
		USE permission is enough to be able to manage transformations.
	*/
	Permission getPermissionForWriteDeleteAndCreate()
	{
		return Permission.USE;
	}
	// -------------------------------------------

	/**
		Get the experiment this transformation belongs to.
		@return An <code>Experiment</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Experiment getExperiment()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Experiment.class, getData().getExperiment());
	}
	
	/**
		Set the experiment. This cannot be changed after the transformation
		has been created.
	*/
	private void setExperiment(Experiment experiment)
	{
		getData().setExperiment(experiment.getData());
	}

	/**
		Get the source bioassayset used by this transformation.
		
		@return A <code>BioAssaySet</code> item, or null if this is a root
			transformation
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
		@see #getRawSources()
	*/
	public BioAssaySet getSource()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(BioAssaySet.class, getData().getSource());
	}

	/**
		Set the source bioassayset. This cannot be changed after the transformation
		has been created.
		@see BioAssaySet#newTransformation(Job)
	*/
	void setSource(BioAssaySet bioAssaySet)
	{
		getData().setSource(bioAssaySet.getData());
	}
	
	/**
		Get a query that returns the raw bioassay sources used by this
		transformation. It is only a root transformation that has raw bioassays
		as sources, other transformations has a source {@link BioAssaySet}.
		@return An <code>ItemQuery</code> object
		@see #getSource()
	*/
	public ItemQuery<RawBioAssay> getRawSources()
	{
		ItemQuery<RawBioAssay> query = RawBioAssay.getQuery();
		query.joinPermanent(Hql.innerJoin("transformations", Item.TRANSFORMATION.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.TRANSFORMATION.getAlias()),
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Set the raw bioassay sources used in this transformation. Can only be specified
		for a root transformation. All raw bioassays must be members of the experiment.
		@throws InvalidDataException If any of the raw bioassays are not part
			of the experiment
		@see Experiment#newTransformation(Job, Collection)
		@see Experiment#isUsing(RawBioAssay)
	*/
	void setRawSources(Collection<RawBioAssay> rawBioAssays)
		throws InvalidDataException
	{
		Experiment e = getExperiment();
		Set<RawBioAssayData> rawSources = getData().getRawSources();
		for (RawBioAssay rba : rawBioAssays)
		{
			if (!e.isUsing(rba)) 
			{
				throw new InvalidDataException("Raw bioassay "+rba+" is not a part of the experiment: "+e);
			}
			rawSources.add(rba.getData());
		}
	}

	/**
		Get the job that executed this transformation.
		@return A <code>Job</code> object, ot null if this transformation
			wasn't executed by a job
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Job getJob()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Job.class, getData().getJob());
	}
	
	/**
		Set the job that executed this transformation. Once set it can't 
		be changed.
	*/
	private void setJob(Job job)
	{
		this.job = job;
		getData().setJob(job == null ? null : job.getData());
	}

	/**
		Internal cache of named cubes used by the newProduct method.
	*/
	private Map<String, DataCube> namedCubes;
	
	/**
		Internal cache of named layers used by the newProduct method.
	*/
	private Map<String, DataCubeLayer> namedLayers;
	
	/**
		Create a new product bioassayset from this transformation. The transformation
		mustn't have been committed to the database. The bioassayset can use
		either the same data cube/layer as the source bioassayset is using, a new
		data cube/layer or a data cube/layer that has previously been created by
		this method. The names specified here are only valid for the current session
		and are only needed if the transformation must create more than one bioassayset
		in a different data cube than it's source bioassayset.
		<pre class="code">
Transformation t = ...

// Use same data cube, but a new layer; automatically create child bioassays
BioAssaySet bas = t.newProduct(null, "new", true);

// Use same data cube and layer; automatically create child bioassays
BioAssaySet filtered = t.newProduct(null, null, true);

// Use new data cube and layer; automatic creation of 
// child bioassays is not supported
BioAsssySet merged = t.newProduct("new", "new", false);

// Create four new bioassaysets using two datacubes, one
// with two layers and one with one layer
BioAssaySet bas1 = t.newProduct("cube1", "layer1", false);
BioAssaySet bas2 = t.newProduct("cube1", "layer2", false);
BioAssaySet bas3 = t.newProduct("cube2", "layer3", false);
BioAssaySet bas4 = t.newProduct("cube2", "layer3", false);
</pre>

		Note that the name of the layer is checked first, so we can't have the
		layers with the same name in two different cubes. For example the following
		call won't work:
		<pre class="code">
// Don't! We bas3 is now using the same cube and layer as bas1
BioAssaySet bas3 = t.newProduct("cube2", "layer1", false);
</pre>

		@param namedCube The name of the data cube where the new bioassayset should
			store/find it's data. If no data cube with the specified name exists a new
			data cube is created. Specify 'null' to use the same data cube as
			the source bioassayset.
		@param namedLayer The name of the data cube layer where the new bioassayset
			should store/find it's data. If no layer with the specified name exists
			a new layer is created. Specify 'null' to use the same layer as the
			source bioassayset.
		@param createChildBioAssays TRUE to create one child bioassay for 
			each bioassay in the source bioassay set. This requires that 
			the product is created in the same data cube as the source
			bioassay set, ie. namedCube should be null
		@return The new bioassayset
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission for the experiment this transformation belongs to
			or if this transformation has already been committed to the database
		@throws InvalidDataException If the layer is null or not part
			of the same experiment as this transformation
		@throws BaseException If there is another error
	*/
	public BioAssaySet newProduct(String namedCube, String namedLayer, boolean createChildBioAssays)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		if (isInDatabase())
		{
			throw new PermissionDeniedException("Transformation has already been committed: "+this);
		}
		if (namedCubes == null)
		{
			namedCubes = new HashMap<String, DataCube>();
			namedLayers = new HashMap<String, DataCubeLayer>();
			BioAssaySet source = getSource();
			if (source != null)
			{
				namedCubes.put(null, source.getDataCube());
				namedLayers.put(null, source.getDataCubeLayer());
			}
		}
		DataCubeLayer layer = namedLayers.get(namedLayer);
		DbControl dc = getDbControl();
		DataCube cube = null;
		if (layer == null)
		{
			cube = namedCubes.get(namedCube);
			if (cube == null)
			{
				cube = getExperiment().getVirtualDb().newDataCube();
				namedCubes.put(namedCube, cube);
			}
			layer = cube.newLayer();
			namedLayers.put(namedLayer, layer);
		}
		BioAssaySet newBas = BioAssaySet.getNew(dc, this, layer);
		if (cube != null && !cube.isInDatabase()) dc.saveItemIf(newBas, cube, true);
		if (!layer.isInDatabase()) dc.saveItemIf(newBas, layer, true);
		
		if (namedCube == null && createChildBioAssays)
		{
			BioAssaySet source = getSource();
			for (BioAssay parentBa : source.getBioAssays().list(dc))
			{
				BioAssay childBa = newBas.newBioAssay(parentBa);
				dc.saveItemIf(newBas, childBa, false);
			}
		}
		return newBas;
	}

	/**
		Get a query that return all bioassaysets created by this 
		transformation.
		@return An <code>ItemQuery</code> object
		@see BioAssaySet#getQuery(Transformation)
	*/
	public ItemQuery<BioAssaySet> getProducts()
	{
		return BioAssaySet.getQuery(this);
	}
	
	/**
		Delete all product bioassaysets. Called by onBeforeCommit when
		the transformation is deleted.
	*/
	private void deleteProducts()
	{
		DbControl dc = getDbControl();
		ItemResultIterator<BioAssaySet> products = getProducts().iterate(dc);
		while (products.hasNext())
		{
			dc.deleteItem(products.next());
		}
	}
	
}
