/*
	$Id: UserDefaultSetting.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.UserDefaultSettingData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;


/**
	This class is used to represent a setting valid for a specific user
	using any client application.

	@author Nicklas
	@version 2.0
*/
public class UserDefaultSetting
	extends Setting<UserDefaultSettingData>
{
	/**
		The type of item represented by this class.
		@see Item#USERDEFAULTSETTING
		@see #getType()
	*/
	public static final Item TYPE = Item.USERDEFAULTSETTING;

	/**
		This filter will only return items if the logged in user has
		generic read permission to users.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = 
		new QueryRuntimeFilterFactory.ChildFilter(TYPE, Item.USER);

	/**
		Create a new <code>UserDefaultSetting</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param user The <code>User</code> which this should be a default setting for
		@param name The name of the setting
		@param value The value of the setting
		@return The new <code>UserDefaultSetting</code> item
		@throws BaseException If there is an error
		@see User#newDefaultSetting(String,String)
	*/
	public static UserDefaultSetting getNew(DbControl dc, User user, String name, String value)
		throws BaseException
	{
		UserDefaultSetting uds = dc.newItem(UserDefaultSetting.class);
		uds.setUser(user);
		uds.setName(name);
		uds.setValue(value);
		return uds;
	}

	/**
		Get a <code>UserDefaultSetting</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>UserDefaultSetting</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static UserDefaultSetting getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		UserDefaultSetting uds = dc.loadItem(UserDefaultSetting.class, id);
		if (uds == null) throw new ItemNotFoundException("UserDefaultSetting[id="+id+"]");
		return uds;
	}

	/**
		Get a query configured to retrieve default settings for a user.
		
		@param user The user to retreive settings for, null is allowed if
			the logged in user has generic READ permission for users in which case
			all settings will be returned
		@return An {@link ItemQuery} object
		@see User#getDefaultSettings()
	*/
	public static ItemQuery<UserDefaultSetting> getQuery(User user)
	{
		ItemQuery<UserDefaultSetting> query = null;
		if (user != null)
		{
			query = new ItemQuery<UserDefaultSetting>(UserDefaultSetting.class, null);
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("user"),
					Hql.entity(user)
				)
			);
		}
		else
		{
			query = new ItemQuery<UserDefaultSetting>(UserDefaultSetting.class, RUNTIME_FILTER);
		}
		return query;
	}

	UserDefaultSetting(UserDefaultSettingData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		READ permission is granted if the logged in user is the owner of this
		setting. WRITE, CREATE and DELETE permissions are granted if the logged in user
		is the owner and not is a multiuser account. WRITE, DELETE and CREATE permissions are 
		granted if the logged in user have generic WRITE permission for users.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		SessionControl sc = getSessionControl();
		UserData user = getData().getUser();
		if (sc.getLoggedInUserId() == user.getId())
		{
			if (user.isMultiuserAccount())
			{
				granted |= Permission.grant(Permission.READ);
			}
			else
			{
				granted |= Permission.grant(Permission.READ, Permission.WRITE, Permission.DELETE, Permission.CREATE);
			}
		}
		int rolePermissions = sc.getRolePermissions(Item.USER);
		if (Permission.hasPermission(rolePermissions, Permission.READ))
		{
			granted |= Permission.grant(Permission.READ);
		}
		if (Permission.hasPermission(rolePermissions, Permission.WRITE))
		{
			granted |= Permission.grant(Permission.WRITE, Permission.DELETE, Permission.CREATE);
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/**
		Get the user this setting is valid for.
		@return The <code>User</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission
		@throws BaseException If there is another error
	*/
	public User getUser()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(User.class, getData().getUser());
	}
	
	/**
		Set the user this setting is valid for. Can only be set on a
		new item.
		@throws InvalidDataException If the user is null
	*/
	void setUser(User user)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (user == null) throw new InvalidUseOfNullException("user");
		getData().setUser(user.getData());
	}
	
}

