/**
	$Id: DbEngine.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2006 Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA. 
*/
package net.sf.basedb.core.dbengine;

import java.util.Set;

import org.hibernate.dialect.Dialect;

/**
	Information about database-specific options that are not covered by the
	Hibernate {@link org.hibernate.dialect.Dialect} objects.
	
	
	@author nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
	@see net.sf.basedb.core.dbengine.EngineFactory#createEngine(Dialect)
*/
public interface DbEngine
{
	
	/**
		If the database uses theta join or ansi join.
		@return TRUE for theta join, FALSE for ansi join
	*/
	public boolean useThetaJoin();
	
	/**
		Generate SQL to create an index.

		@param catalog The name of the catalog (database) where the table is
			located, or null if it is located in the current catalog
		@param schema The name of the schema where the table is located, or
			null if is located in the current schema
		@param table The name of the table
		@param name The name of the index to be created
		@param columns The name of the columns in the index
		@param unique If the index must contain unique values or not
		@return The SQL to execute
	*/
	public String getCreateIndexSql(String catalog, String schema, String table, String name, Set<String> columns, boolean unique);
	
	/**
		Generate SQL to drop an index.

		@param catalog The name of the catalog (database) where the table is
			located, or null if it is located in the current catalog
		@param schema The name of the schema where the table is located, or
			null if is located in the current schema
		@param table The name of the table
		@param name The name of the index
		@param unique If the index contains unique values or not
		@return The SQL to execute
	*/
	public String getDropIndexSql(String catalog, String schema, String table, String name, boolean unique);
	
	/**
		If TRUE, a batcher that inserts, updates or delets a lot
		of rows in a table should execute the SQL returned by 
		{@link #getAnalyzeTableSql(String, String, String)} for that table.
		<p>
		Postgres, for example, almost always require that an analysis is run
		since subsequent queries otherwise may take a long time to execute
		if indexes can't be used properly.
	*/
	public boolean analyzeAfterBatchOperation();
	
	/**
	 	Generate SQL to optimize a table. The meaning of this may depend
	 	on the actual database used. Typical operations include updating index
	 	statistics, removing unused records, etc. Here are some example:
	 	<pre class="code">
MySQL: OPTIMIZE TABLE &lt;table&gt;
Postgres: VACCUUM &lt;table&gt;
</pre>
		
		NOTE! The SQL returned by this method may impact the performance on
		the database. It is only intended to be executed when there is a need
		to clean up the database. How often this is needed depends on the 
		database.

		@param catalog The name of the catalog (database) where the table is
			located, or null if it is located in the current catalog
		@param schema The name of the schema where the table is located, or
			null if is located in the current schema
	 	@param table The name of the table to optimize
	 	@return The SQL to execute
	 	@see #getAnalyzeTableSql(String, String, String)
	*/
	public String getOptimizeTableSql(String catalog, String schema, String table);
	
	/**
		Generate SQL to update the index statistics for a table.
		Here are some example:
	 	<pre class="code">
MySQL: ANALYZE TABLE &lt;table&gt;
Postgres: ANALYZE &lt;table&gt;
</pre>

		NOTE! The SQL returned by this method must be able to run within
		a transaction and from multiple transactions on the same table
		at the same time. If no such SQL statement exists null should be returned.

		@param catalog The name of the catalog (database) where the table is
			located, or null if it is located in the current catalog
		@param schema The name of the schema where the table is located, or
			null if is located in the current schema
		@param table The name of the table to analyze
		@return The SQL to execute, or null if not supported
		@see #getOptimizeTableSql(String, String, String)
	 */
	public String getAnalyzeTableSql(String catalog, String schema, String table);
	
	/**
		Create a "safe" CREATE TABLE query string that will not fail if the table already
	 	exists. Implementors should inject proper SQL code in a way that is supported in the
	 	underlying database. The MySQL engine, for example, replaces CREATE TABLE with
	 	CREATE TABLE IF NOT EXISTS. If the underlying database doesn't support this kind
	 	of check the original SQL should be returned umodified.
	 	
		@param sql The original SQL
		@param catalog The name of the catalog (database) where the table is
			being create, or null if it is located in the current catalog
		@param schema The name of the schema where the table is being created, or
			null if is located in the current schema
		@param table The name of the table that is being created
	 	@return The modified or unmodified SQL statement
	 	@since 2.6
	*/
	public String makeSafeCreateTable(String sql, String catalog, String schema, String table);
	
	/**
		If the database does case sensitive or case insensitive string
		comparison in expressions.
		@return FALSE if comparisons are case sensitive (ABC != abc), TRUE if they
			are case insensitive (ABC = abc)
		@since 2.4
	*/
	public boolean caseInsensitiveComparison();
	
	/**
		If the current database supports column aliases in the where part of 
		a query or not. For example: 
		<code>SELECT tbl.blah as foo, ... FROM Table tbl WHERE foo = 2</code>
		<p>
		If not supported the entire expression will be used again.

		@return TRUE if column aliases are supported, FALSE otherwise
	*/
	public boolean supportsColumnAliasInWhere();
	
	/**
		If the current database supports column aliases in the order by part of 
		a query or not. For example: 
		<code>SELECT tbl.blah as foo, ... FROM Table tbl ORDER BY foo</code>
		<p>
		If not supported the entire expression will be used again.

		@return TRUE if column aliases are supported, FALSE otherwise
	*/
	public boolean supportColumnAliasInOrderBy();
	
	/**
		If the current database supports column aliases in the group by part of 
		a query or not. For example: 
		<code>SELECT tbl.blah as foo, ... FROM Table tbl GROUP BY foo</code>
		<p>
		If not supported the entire expression will be used again.

		@return TRUE if column aliases are supported, FALSE otherwise
	*/
	public boolean supportColumnAliasInGroupBy();

	/**
		If the current database supports column aliases in the order by part of 
		a query or not. For example: 
		<code>SELECT count(tbl.blah) as foo, ... FROM Table tbl 
		GROUP BY ... HAVING foo = 2</code>
		<p>
		If not supported the entire expression will be used again.

		@return TRUE if column aliases are supported, FALSE otherwise
	*/
	public boolean supportColumnAliasInHaving();
	
	/**
		If the database requires that columns appearing in the
		ORDER BY part of the query must also be part of the SELECT
		list when using the DISTINCT keyword.
		@return TRUE if order by columns must be selected, FALSE otherwise
	*/
	public boolean selectOrderByColumnsIfDistinct();
	
	/**
		If we need to check for invalid arguments to numerical
		functions to avoid exceptions from the database. Postgres,
		for example, throws exceptions if trying to divide by zero,
		while MySQL returns null. To get the same behavior we need
		to add a <code>CASE...WHEN</code> statement to the query for
		Postgres. Example:
		<pre class="code">
CASE WHEN denominator = 0 THEN null ELSE numerator / denominator
</pre>
		@return TRUE if we need to check the numbers, FALSE otherwise
	*/
	public boolean checkForInvalidNumberOperation();
	
	/**
		Get the function call that takes the natural logarithm
		of a value. For example: <code>LN(value)</code>
		@param value The value to take the logarithm of
		@return The function call that calculates the logartihm 
		@throws UnsupportedOperationException If the database doesn't support 
			logarithms
	*/
	public String ln(String value);
	
	/**
		Get the function call that takes the absolute of a value. 
		For example: <code>ABS(value)</code>
		@param value The value to use in the calculation
		@return The function call that calculates the absolute value 
		@throws UnsupportedOperationException If the database doesn't support 
			this function
	*/
	public String abs(String value);

	/**
		Get the function call that calculates the exponential of a value. 
		For example: <code>EXP(value)</code>
		@param value The value to use in the calculation
		@return The function call that calculates the exponential
		@throws UnsupportedOperationException If the database doesn't support 
			this function
	*/
	public String exp(String value);

	/**
		Get a function/expression that matches the left value against
		the right value which should be a regular expression. If the
		database doesn't support regular expressions, the returned
		expression should use equality testing.
		
		@param value The left value 
		@param regexp The regular expression as a string
		@return The function call/expression that matches the value
			agains the regular expression
		@since 2.8
	*/
	public String rlike(String value, String regexp);
	
	/**
		Check if a given string is valid to be used as a table name in the
		current database.
		@param tableName The string to check
		@return TRUE if the name is valid, FALSE if not
		@since 2.4
	*/
	public boolean isValidTableName(String tableName);
	
	/**
		Check if a given string is valid to be used as a column name in the
		current database.
		@param columnName The string to check
		@return TRUE if the name is valid, FALSE if not
		@since 2.4
	*/
	public boolean isValidColumnName(String columnName);
}
