/*
	$Id:CelFileHandler.java 3820 2007-10-12 10:03:18Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core.filehandler;

import affymetrix.fusion.cdf.FusionCDFData;
import affymetrix.fusion.cel.FusionCELData;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreUtil;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidRelationException;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.StringUtil;

/**
	File handler that can validate and extract metadata from
	Affymetrix CEL files.
	
	@author Nicklas
	@version 2.5
	@base.modified $Date:2007-10-12 12:03:18 +0200 (Fri, 12 Oct 2007) $
*/
public class CelFileHandler
	extends AffymetrixFileHandler
	implements DataFileValidator, DataFileMetadataReader
{
	private FusionCELData celData;

	public CelFileHandler()
	{}
	
	/*
		From the DataFileValidator interface
		----------------------------------------
	*/
	/**
		Check that the raw bioassay is using the Affymetrix platform 
		({@link Platform#AFFYMETRIX}) and that the file selected
		as the CEL file really is a CEL file.
	*/
	public void validate(DbControl dc)
	{
		RawBioAssay rba = (RawBioAssay)getItem();
		// Check platform
		if (!isAffymetrix(rba))
		{
			throw new InvalidDataException("Raw bioassay '" + rba.getName() + 
				"' is not using the Affymetrix platform");
		}
		
		// Get the member for CEL file type
		FileSetMember celMember = getMember(DataFileType.AFFYMETRIX_CEL);
		if (celMember != null)
		{
			// Verify CEL file
			File celFile = celMember.getFile();
			celData = loadCelFile(celFile);
		}
		
		// The CEL file is valid, a failure below will result in InvalidRelationException
		try
		{
			ArrayDesign design = rba.getArrayDesign();
			if (design == null) 
			{
				throw new InvalidDataException("Raw bioassay " + rba.getName() + 
					" has no array design");
			}
			if (!design.isPlatform(Platform.AFFYMETRIX))
			{
				throw new InvalidDataException("The array design " + design.getName() + 
					" is not an Affymetrix design");
			}

			// Verify that CEL chip type match CDF type
			File cdfFile = FileStoreUtil.getDataFile(dc, design, DataFileType.AFFYMETRIX_CDF);
			FusionCDFData cdfData = loadCdfFile(cdfFile);
			validateCelAndCdf(celData, cdfData, cdfFile.getName());
		}
		catch (Throwable t)
		{
			throw new InvalidRelationException(t);
		}
		
	}
	// -------------------------------------------
	
	/*
		From the DataFileMetadataReader interface
		----------------------------------------
	*/
	public void extractMetadata(DbControl dc)
	{
		if (celData == null)
		{
			FileSetMember celMember = getMember(DataFileType.AFFYMETRIX_CEL);
			if (celMember != null) celData = loadCelFile(celMember.getFile());
		}
		if (celData != null)
		{
			RawBioAssay rba = (RawBioAssay)getItem();
			setHeader(rba, "Algorithm", celData.getAlg());
			for (int i = 0; i < celData.getNumberAlgorithmParameters(); ++i)
			{
				String name = celData.getAlgorithmParameterTag(i);
				setHeader(rba, "Algorithm parameter: " + name, 
						celData.getAlgorithmParameter(name));
			}
			setHeader(rba, "Dat header", celData.getDatHeader());
			setHeader(rba, "Chip type", celData.getChipType());
			
			// Set number of spots
			rba.setNumFileSpots(celData.getCells());
		}
	}
	
	/**
		Reset the number of file spots to 0, and remove all headers.
	*/
	public void resetMetadata(DbControl dc)
	{
		RawBioAssay rba = (RawBioAssay)getItem();
		rba.setNumFileSpots(0);
		rba.removeHeaders();
	}
	// -------------------------------------------

	private void setHeader(RawBioAssay rba, String name, String value)
	{
		if (name == null || value == null) return;
		value = StringUtil.trimString(value, RawBioAssay.MAX_HEADER_VALUE_LENGTH);
		rba.setHeader(name, value);
	}

	private static void validateCelAndCdf(FusionCELData cel, FusionCDFData cdf, String cdfChipType)
		throws InvalidDataException
	{
		String celChipType = cel.getChipType();
		if (!cdfChipType.startsWith(celChipType))
		{
			throw new InvalidDataException("CEL chip type (" + celChipType + 
				") doesn't match CDF chip type (" + cdfChipType + ")");
		}
		
		if (cel.getRows() != cdf.getHeader().getRows() || cel.getCols() != cdf.getHeader().getRows())
		{
			throw new InvalidDataException("CEL size (rows=" + cel.getRows() + ", cols=" + cel.getCols() +
				") doesn't match CDF size (rows=" + 
				cdf.getHeader().getRows() + ", cols=" + cdf.getHeader().getCols()+")");
		}		
	}

	
}