/*
	$Id: Plugin.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core.plugin;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.SessionControl;

import java.util.Collection;
import java.util.Map;
import java.util.HashMap;

/**
	This interface must be implemented by all plugins to BASE.
	A plugin must also define a default public no-argument
	constructor.

	<p>
	When a plugin is installed into BASE the core will use
	the following sequence to store all meta-information about the plugin
	in the database:
	<ol>
	<li>Create a plugin object, using the default constructor
	<li>Call the {@link #getMainType()} and {@link #getAbout()} methods
	<li>Check if the plugin inmplements the {@link InteractivePlugin}
		and if it does, call the {@link InteractivePlugin#getGuiContexts()}
		method.
	<li>Check which other interfaces the plugin implements, and
		link to {@link net.sf.basedb.core.PluginType} as appropriate
	</ol>

	<p>
	When a plugin is executed using a the request/response scheme the core will
	use the following sequence:
	<ol>
	<li>Create a plugin object, using the default constructor
	<li>Call the {@link #init(SessionControl, ParameterValues, ParameterValues)} method
		which gives the plugin access to it's configuration and job-specific parameters
	<li>Call the {@link #run(Request, Response, ProgressReporter)} method to let
		the plugin do it's work
	<li>Call {@link #done()} to let the plugin clean up after itself
	<li>After <code>done</code> has been called, the plugin instance is not
		reused again
	</ul>

	@author Nicklas, Enell
	@version 2.0
	@base.modified $Date$
*/
public interface Plugin
{
	/**
		Get the type of the plugin. This method must always
		return the same value.
		@return One of the defined types
	*/
	public MainType getMainType();

	/**
		Get information about the plugin, such as
		name, version, authors, etc.
		@return An {@link About} object
	*/
	public About getAbout();

	/**
		Return TRUE if the plugin supports different configurations when
		executing a job.
		@return TRUE or FALSE
	*/
	public boolean supportsConfigurations();
	
	/**
		Return TRUE if the plugin requires a configuration when executing
		a job. The core will not start a job without a configuration if this
		method returns TRUE. Note! If the {@link #supportsConfigurations()}
		method returns FALSE, the return value of this method is expected 
		to also be FALSE.
		@return TRUE or FALSE
	*/
	public boolean requiresConfiguration();
	
	/**
		Get the permissions this plugin requires to function as expected.
		The plugin may specify permissions that is always granted to the
		plugin even if the logged in user doesn't have that permission and
		permissions that are only granted if the logged in user has the required
		permission. Permissions that are not specified are always denied.
		If null or an empty collection is returned, the permissions are not
		used and the plugin will always get the same permissions as the logged
		in user.
		
		@return A collection of permissions or null to not use permissions
	*/
	public Collection<Permissions> getPermissions();
	
	/**
		This method is called right after the plugin object has been constructed 
		to pass the configuration and job parameters to the plugin. The 
		<code>ParameterValues</code> parameters can be null if no parameters exists.

		@param sc A <code>SessionControl</code> object that the plugin
			can use to communicate with the core.
		@param configuration The configuration parameters for the plugin
		@param job The job parameters for the plugin
		@throws BaseException if there is an error.
	*/
	public void init(SessionControl sc, ParameterValues configuration, ParameterValues job)
		throws BaseException;

	/**
		Run the plugin. If a progress reporter object is passed it is 
		recommended that the plugin makes use of it.

		@param request Request object with the command and parameters
		@param response Response object in for the plugin to response
			through
		@param progress A {@link ProgressReporter} where the plugin can report
			its progess, can be null
	*/
	public void run(Request request, Response response, ProgressReporter progress);
	
	/**
		This method is called when the core is finished
		with the plugin object. The plugin should clean up
		and close any resources it has aquired.
	*/
	public void done();

	public enum MainType
	{
		IMPORT(1, "Import"),
		EXPORT(2, "Export"),
		INTENSITY(3, "Intensity"),
		ANALYZE(4, "Analyze"),
		OTHER(5, "Other");

		/**
			Maps an integer to a type.
		*/
		private static final Map<Integer, MainType> valueMapping = new HashMap<Integer, MainType>();
		static
		{
			for(MainType type : MainType.values())
			{
				MainType t = valueMapping.put(type.getValue(), type);
				assert t == null : "Another type with the value "+type.getValue()+" already exists";
			}
		}
		
		public static MainType fromValue(int value)
		{
			MainType type = valueMapping.get(value);
			assert type != null : "type == null for value "+value;
			return type;
		}
		
		private final int value;
		private final String displayValue;
		private MainType(int value, String displayValue)
		{
			this.value = value;
			this.displayValue = displayValue;
		}

		public int getValue()
		{
			return value;
		}

		public String toString()
		{
			return displayValue;
		}
	}
}
