/*
	$Id: Response.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core.plugin;

import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.ProgressReporter;

import java.util.List;

/**
	The response interface used by the core.

	@author Nicklas, Enell
	@version 2.0
	@base.modified $Date$
*/
public interface Response
{
	/**
		This method will continue the configuration of the job with 
		a new command to the plugin. This response is only allowed 
		while configuring a plugin or a job, not after executing the plugin.
		
		@param nextCommand the next command to the plugin
		@throws PermissionDeniedException If another command isn't allowed,
			ie. after an {@link Request#COMMAND_EXECUTE} command
	*/
	public void setContinue(String nextCommand)
		throws PermissionDeniedException;
	
	/**
		This will end the configuration of a job and set the estimated
		execution time for it.
		@param message An optional message
		@param estimatedExecutionTime An estimated execution time for the job,
			null is the same as <code>ExecutionTime.LONG</code>
	*/
	public void setDone(String message, Job.ExecutionTime estimatedExecutionTime);

	/**
		This will end the configuration of a job, and start to execute it
		immediately if the plugin is permitted to do so. If it is not possible
		to execute the job immediately it will be added to the job queue for 
		later execution unless the <code>failIfDenied</code> flag is set to true,
		in which case an exception will be thrown.
		<p>
		If immediate execution is possible the client application must continue to 
		use the same instance of the plugin object. This means that immediate execution
		always happens in the same virtual machine. Plugins should be aware of
		this and avoid excessive use of resources such as memory and CPU time.
		<p>
		The client application may execute the job in another thread.
		@param message An optional message
		@param estimatedExecutionTime An estimated execution time for the job,
			null is the same as <code>ExecutionTime.LONG</code>. This value is only used
			if immediate execution isn't possible and the job has to be added to the
			job queue.
		@param failIfDenied TRUE if a PermissionDeniedException should be thrown if
			immediate exection isn't possible, FALSE to add the job to the job queue 
			for later exection
	 	@throws PermissionDeniedException If logged in user doesn't have write permission on the job. 
		@since 2.2
	*/
	public void setExecuteImmediately(String message, Job.ExecutionTime estimatedExecutionTime, 
		boolean failIfDenied)
		throws PermissionDeniedException;
	
	/**
		This will end the configuration of a <b>export</b> job, and start to execute it
		for immediate download. This method is similar to the {@link 
		#setExecuteImmediately(String, Job.ExecutionTime, boolean)} but:
		<ul>
		<li>The plugin must implement the {@link ImmediateDownloadExporter} interface.
		<li>The core will call {@link ImmediateDownloadExporter#doExport(ExportOutputStream, ProgressReporter)}
			instead of the {@link Plugin#run(Request, Response, ProgressReporter)} method.
		</ul>
		
		@param message An optional message
		@param estimatedExecutionTime An estimated execution time for the job,
			null is the same as <code>ExecutionTime.LONG</code>. This value is only used
			if immediate execution isn't possible and the job has to be added to the
			job queue.
		@param failIfDenied TRUE if a PermissionDeniedException should be thrown if
			immediate exection isn't possible, FALSE to add the job to the job queue 
			for later exection
	 	@throws PermissionDeniedException If the logged in user doesn't have write permission to the job. 
		@since 2.2
	 */
	public void setDownloadImmediately(String message, Job.ExecutionTime estimatedExecutionTime, 
		boolean failIfDenied)
		throws PermissionDeniedException;
	
	/**
		This will end the job execution and report it as a success.
		@param message A message containing any relevant information about
			what the plugin did, or null
	*/
	public void setDone(String message);
	
	/**
		A plug-in may call this method in it's configuration phase or
		after it has been executed to indicate that the job should be
		or was executed as a dry-run job. A dry-run job is a job that 
		does all the things a normal job would do except committing the
		changes to the database. The dry-run job may output a 
		log file with detailed information about what should have happened
		if it hadn't been a dry-run.
		<p>
		NOTE! Setting this value to true doesn't change the behaviour
		of the core. It is the responsibility of the plug-in to check
		this value(eg. {@link Job#isDryRun()} and take appropriate action.
		
		@param dryRun TRUE to indicate that the job should be or was
			executed as a dry-run
		@since 2.8
	*/
	public void setDryRun(boolean dryRun);
	
	/**
		This will end the job execution or configuration and report it as a 
		failure. If in a configuration sequence the client application may try 
		the same command again with different parameters.
		
		@param message Information about the reason for the error
		@param errorList Detailed information about the error 
	*/
	public void setError(String message, List<? extends Throwable> errorList);
	
	public enum Status
	{
		DONE(true), 
		CONTINUE(false), 
		ERROR(true), 
		EXECUTE_IMMEDIATELY(true), 
		DOWNLOAD_IMMEDIATELY(true);
		
		private final boolean allowSaveParameters;
		
		private Status(boolean allowSaveParameters)
		{
			this.allowSaveParameters = allowSaveParameters;
		}
		
		public boolean allowSaveParameters()
		{
			return allowSaveParameters;
		}
		
	}
}
