/*
	$Id: ReporterListExpression.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2006 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core.query;

import java.util.Collection;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.HibernateUtil;
import net.sf.basedb.core.RealTable;
import net.sf.basedb.core.ReporterList;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.QueryType;

/**
	Represents a property of a reporter list in a dynamic query:
	<code>alias.property</code>.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class ReporterListExpression
	implements Expression
{
	
	private final boolean hqlMode;
	private final ReporterList reporterList;
	private final String tableAlias;
	private String tableColumn;
	private String joinAlias;
	private String joinProperty;
	private String property;
	private final JoinType autoJoinType;
	
	/**
		Create a new expression for the specified reporter list.
		
		@param reporterList The reporter list
		@param tableColumn The name of the column we should select
		@param autoJoinType If an automatic join is done use this type, the join
			type depends on if we are checking if a reporter is part of or not part
			of a reporter list
		@see net.sf.basedb.core.query.Dynamic#isPartOf(ReporterList)
		@see net.sf.basedb.core.query.Dynamic#isNotPartOf(ReporterList)
		@see net.sf.basedb.core.query.Dynamic#score(ReporterList)
	*/
	ReporterListExpression(ReporterList reporterList, String tableColumn, JoinType autoJoinType)
	{
		this.hqlMode = false;
		this.reporterList = reporterList;
		this.tableAlias = RealTable.REPORTER_LIST_SCORES.getAlias(reporterList);
		this.tableColumn = tableColumn;
		this.autoJoinType = autoJoinType;
	}
	
	ReporterListExpression(ReporterList reporterList, String joinAlias, String joinProperty, String property, JoinType autoJoinType)
	{
		this.hqlMode = true;
		this.reporterList = reporterList;
		this.tableAlias = RealTable.REPORTER_LIST_SCORES.getAlias(reporterList);
		this.joinAlias = joinAlias;
		this.joinProperty = joinProperty;
		this.property = property;
		this.autoJoinType = autoJoinType;
	}
	
	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.SQL && !hqlMode)
		{
			String alias = HibernateUtil.quote(tableAlias);
			return alias + "." + HibernateUtil.quote(tableColumn);
		}
		else if (query.getQueryType() == QueryType.HQL && hqlMode)
		{
			return tableAlias + "." + property;
		}
		else
		{
			throw new UnsupportedOperationException("ReporterListExpression operation not supported by "+query.getQueryType()+" query.");
		}
	}
	/**
		@return Always null
	*/
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	public String toString()
	{
		if (!hqlMode)
		{
			return tableAlias + "." + tableColumn;
		}
		else
		{
			return tableAlias + "." + property;
		}
	}
	// -------------------------------------------

	/**
		Get the reporter list.
	*/
	public ReporterList getReporterList()
	{
		return reporterList;
	}
	
	public String getJoinAlias()
	{
		return joinAlias;
	}
	
	public String getJoinProperty()
	{
		return joinProperty;
	}
	
	/**
		Get the join type for automatic joins. 
		@return A join type or null if the default should be used
	*/
	public JoinType getAutoJoinType()
	{
		return autoJoinType;
	}
	
}
