/**
	$Id: SocketSignalReceiver.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.core.signal;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.UnknownHostException;
import java.nio.channels.ClosedByInterruptException;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.Application;
import net.sf.basedb.util.QueryParameters;
import net.sf.basedb.util.SocketUtil;
import net.sf.basedb.util.Values;

/**
	A signal receiver implementation that listens on a socket for incoming signals.
	Accordingly, this receiver supports sending signals between different virtual
	machines running on different servers. The initialisation string for this  
	class should be of the format:
	
	<p>
	<code>port=xx&proxy=host:port&forward=xx&allow=ip-address&allow=ip-address...</code>

	<p>
	where:

	<ul>
	<li>
		The <code>port</code> is the port number the signal receiver will listen on.
		This is optional and if not given a random non-used port will be used.
	
	<li>
		The <code>proxy</code> is the address of a proxy host that will route
		incoming messages to this signal receiver. The proxy will automatically
		be added to the list of allowed hosts. This value is optional.
		
	<li>
		The <code>forward</code> attribute is a boolean option (1 or 0).
		If 1, this signal receiver will act as a proxy and forward signal
		messages which are of the form:
		<p>
		<code><code>signal://handlerID@host:port/?SIGNAL#forward.to.host:port</code>
		<p>
		Before the message is forwarded it is re-built like this:
		<p>
		<code>signal://handlerId@forward.to.host:port/?SIGNAL</code>
		
	<li>
		The <code>allow</code> parts are the ip name or numbers of hosts that are 
		allowed to send signals to the receiver. Except for the special case 
		<code>allow=*</code>, which allows any remote host to send signals an 
		exact match is required. The local host is always allowed to send signals.
		<p>
		If no <code>allow</code> tags are given only allow connections from the
		local host (and the proxy if given) are allowed.	
	</ul>
	
	@author nicklas
	@version 2.6
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class SocketSignalReceiver
	extends AbstractSignalReceiver
{
	/**
		Log signals processing.
	*/
	private static final org.apache.log4j.Logger logger = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.signal.SocketSignalReceiver");

	private InetAddress ip;
	private Thread listener;
	private int port;
	private String proxy;
	private boolean allowForward;
	private Set<InetAddress> allow;
	private boolean allowAll;
	
	/**
		Create a new socket signal receiver. Before it can be used it must
		be initialised with {@link #init(String)}.
	*/
	public SocketSignalReceiver()
	{}

	/*
		From the SignalReceiver interface
		-------------------------------------------
	*/
	/**
		Initialise the receiver. The given parameter should be the port number the 
		receiver should listen on. If the parameter can't be parsed as a numeric value,
		a random free port will be used. This method will open the socket and 
		start a separate thread that listens for incoming signals.
	*/
	public void init(String params)
	{
		logger.info("Initialising: params=" + params);
		parseInitParameters(params);
		listenOnSocket();
		super.init(Application.getHostName() + ":" + port);
	}

	/**
		@return {@link SocketSignalTransporter}
	*/
	public Class<? extends SignalTransporter> getSignalTransporterClass()
	{
		return SocketSignalTransporter.class;
	}

	/**
		Close this receiver and the socket it is listening on.
	*/
	public void close(int wait)
	{
		if (listener != null) 
		{
			logger.info("Interrupting socket signal receiver on port: " + port);
			listener.interrupt();
		}
		super.close(wait);
	}
	// -------------------------------------------

	/*
		From the AbstractSignalReceiver interface
		-------------------------------------------
	*/
	/**
		Override ID generation so we can add proxy information if needed.
	*/
	@Override
	protected String getGlobalSignalId(SignalHandler handler)
	{
		String globalId = super.getGlobalSignalId(handler);
		if (proxy != null)
		{
			try
			{
				URI uri = new URI(globalId);
				globalId = "signal://" + uri.getUserInfo() + "@" + proxy + 
					"/?" + uri.getQuery() + "#" + uri.getHost();
				if (uri.getPort() >= 0) globalId += ":" + uri.getPort();
			}
			catch (URISyntaxException ex)
			{
				// Should not happen
			}
		}
		return globalId;
	}
	/**
		Process the incoming message. This class will check if the
		message contains any forwarding information. If so, the 
		message will be forwarde to the new destination. Otherwise, 
		the message will be processed as normal.
	*/
	@Override
	protected void processSignalMessage(String message)
	{
		logger.info("Processing signal message: " + message);
		boolean forwarded = false;
		if (allowForward) forwarded = forwardSignalMessageIfNeeded(message);
		if (!forwarded) super.processSignalMessage(message);
	}
	// -------------------------------------------

	/**
		Parse the initialisation parameters. Called from the {@link #init(String)}
		method.
		@param params Parameter string passed to {@link #init(String)} method
	*/
	protected void parseInitParameters(String params)
	{
		QueryParameters qp = QueryParameters.parseQueryString(params);
		// Get the port number we listen on
		port = Values.getInt(qp.getValue("port"), 0);
		logger.info("port=" + port);

		// Are we allowed to forward the messages
		allowForward = Values.getBoolean(qp.getValue("forward"));
		logger.info("forward=" + allowForward);
		
		// Get allowed hosts
		allow = new HashSet<InetAddress>();
		allowAll = false;
		List<String> allowIp = qp.getValues("allow");
		if (allowIp != null)
		{
			for (String ip : allowIp)
			{
				logger.info("allow=" + ip);
				if ("*".equals(ip)) 
				{
					logger.debug("Allow all hosts enabled!");
					allowAll = true;
				}
				else
				{
					try
					{
						logger.debug("Adding host " + ip + " to list of allowed hosts.");
						allow.add(InetAddress.getByName(ip));
					}
					catch (UnknownHostException ex)
					{
						logger.warn("Unknown host: " + ip, ex);
					}
				}
			}
		}
		
		// Get proxy
		proxy = qp.getValue("proxy");
		logger.info("proxy=" + proxy);
		if (proxy != null)
		{
			try
			{
				URI proxyURI = new URI("singal://" + proxy);
				allow.add(InetAddress.getByName(proxyURI.getHost()));
			}
			catch (Exception ex)
			{
				throw new SignalException("Invalid proxy: " + proxy, ex);
			}
		}
	}	
	
	/**
		Start listening for incoming signals.
	*/
	protected void listenOnSocket()
	{
		logger.info("Starting socket signal receiver on port: " + port);
		try
		{
			ServerSocketChannel channel = ServerSocketChannel.open();
			ServerSocket socket = channel.socket();
			socket.bind(new InetSocketAddress(port));
			port = socket.getLocalPort();
			ip = socket.getInetAddress();
			listener = new Thread(new ListenerThread(channel), "ListenerThread."+port);
			listener.setDaemon(true);
			listener.start();
			logger.info("Socket signal receiver is listening on port: " + port);
		}
		catch (IOException ex)
		{
			logger.error("Could not initialise socket signal receiver", ex);
			throw new SignalException(ex);
		}
	}
	
	/**
		If the message has forwarding information, forward the message to
		the other host. 
		@return TRUE if the message was forwarded, FALSE otherwise
	*/
	protected boolean forwardSignalMessageIfNeeded(String message)
	{
		if (message == null) return false;
		try
		{
			URI uri = new URI(message);
			String fragment = uri.getFragment();
			if (fragment != null)
			{
				Signal signal = Signal.getSignal(uri.getQuery());
				String init = "signal://" + uri.getUserInfo() + "@" + fragment;
				logger.info("Forwarding signal message: message=" + message + "; to=" + init);
				SignalTransporter forwardTo = new SocketSignalTransporter();
				forwardTo.init(init);
				forwardTo.send(signal);
				return true;
			}
		}
		catch (Exception ex)
		{
			// Ignore invalid messages
			logger.warn("Could not process signal message: " + message, ex);
		}
		return false;
	}
	
	/**
		Checks if the specified remote host is allowed to send
		signals to this receiver.
		@param remoteHost The remote host that is sending the signal
		@return TRUE if the host is allowed, FALSE otherwise
	*/
	protected boolean isAllowedHost(InetAddress remoteHost)
	{
		return allowAll || SocketUtil.isLocalHost(remoteHost) || allow.contains(remoteHost);
	}
	
	
	/**
		This class is used for listening to the specified socket for
		incoming connections. 
	*/
	private class ListenerThread
		implements Runnable
	{
		
		private final ServerSocketChannel socket;
		
		private ListenerThread(ServerSocketChannel socket)
		{
			this.socket = socket;
		}
		
		/*
			From the Runnable interface
			-------------------------------------------
		*/
		public void run()
		{
			// We listen until someone tells us not to!
			boolean interrupted = false;
			while (!interrupted)
			{
				try
				{
					SocketChannel incoming = socket.accept();
					Socket inSocket = incoming.socket();
					InetAddress remoteHost = inSocket.getInetAddress();
					
					logger.debug("Incoming signal: remote host=" + remoteHost);
					
					// Check if we are allowed to accept connections from the remote host
					if (isAllowedHost(remoteHost))
					{
						// Read the incoming message
						String message = SocketUtil.read(inSocket, true);
						logger.debug("Incoming signal accepted: remote host=" + remoteHost +
							"; message=" + message);
						processSignalMessage(message);
					}
					else
					{
						logger.debug("Incoming signal rejected: remote host=" + remoteHost);
					}
					
					interrupted = Thread.interrupted();
				}
				catch (ClosedByInterruptException ex)
				{
					logger.info("Shutting down socket signal receiver on port: " + port);
					interrupted = true;
				}
				catch (IOException ex)
				{
					logger.warn("Error on socket signal receiver on port: " + port, ex);
				}
			}
			SocketUtil.close(socket);
		}
		// -------------------------------------------
	}

	
}
