/*
	$Id: FileUtil.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2005, 2006 Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.util;

import net.sf.basedb.core.AbsoluteProgressReporter;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.signal.ThreadSignalHandler;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PushbackInputStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
	This class collects some useful methods for file and stream
	handling.

	@author Nicklas
	@version 2.0
	@base.modified $Date$
*/
public class FileUtil
{
	private static final int BUFFER_SIZE = 512;

	/**
		Copy from the input stream to the output stream
		until end of file is reached.
		@param in The <code>InputStream</code> to read from
		@param out The <code>OutputStream</code> to write to
		@return The number of bytes copied
		@throws IOException This exception is thrown if there is an error
	*/
	public static long copy(InputStream in, OutputStream out)
		throws IOException
	{
		return copy(in, out, null);
	}

	/**
		Copy from the input stream to the output stream
		until end of file is reached.
		@param in The <code>InputStream</code> to read from
		@param out The <code>OutputStream</code> to write to
		@param progress An optional progress reporter
		@return The number of bytes copied
		@throws IOException This exception is thrown if there is an error
		@since 2.5
	*/
	public static long copy(InputStream in, OutputStream out, AbsoluteProgressReporter progress)
		throws IOException
	{
		int bytes = 0;
		long totalBytes = 0;
		byte[] buffer = new byte[BUFFER_SIZE];
	
		while (bytes != -1) // -1 = end of stream
		{
			ThreadSignalHandler.checkInterrupted();
			bytes = in.read(buffer, 0, buffer.length);
			if (bytes > 0)
			{
				totalBytes += bytes;
				out.write(buffer, 0, bytes);
				if (progress != null) progress.displayAbsolute(totalBytes, null);
			}
		}
		return totalBytes;
	}

	
	/**
		Read from the input stream until the end is reached.
		
		@param in The <code>InputStream</code> to read from
		@return The number of bytes copied
		@throws IOException This exception is thrown if there is an error
		@since 2.4
	*/
	public static long read(InputStream in)
		throws IOException
	{
		int bytes = 0;
		long totalBytes = 0;
		byte[] buffer = new byte[BUFFER_SIZE];
	
		while (bytes != -1) // -1 = end of stream
		{
			ThreadSignalHandler.checkInterrupted();
			bytes = in.read(buffer, 0, buffer.length);
			if (bytes > 0)
			{
				totalBytes += bytes;
			}
		}
		return totalBytes;
	}

	/**
		Get a buffered <code>InputStream</code> object reading from
		the specified file.

		@param file The path to the file to read from
		@return A buffered <code>InputStream</code>
		@throws IOException If the stream cannot be opened
	*/
	public static InputStream getInputStream(File file)
		throws IOException
	{
		return new BufferedInputStream(new FileInputStream(file));
	}
	
	/**
		Get a buffered <code>OutputStream</code> object writing to
		the specified file.
	
		@param file The path to the file to write to
		@return A buffered <code>OutputStream</code>
		@throws IOException If the stream cannot be opened
	*/
	public static OutputStream getOutputStream(File file)
		throws IOException
	{
		return new BufferedOutputStream(new FileOutputStream(file));
	}
	
	/**
		Take a peek at an input stream and check if the first few bytes
		matches the bCheck parameter. After the check the bytes are
		pushed back to the stream again. This method is useful to check
		the file format of a file, which can often be done by checing
		the first few bytes at the start of the file. For example
		to check if the file is gzip file:
		<p>
		<code>bCheck = new byte[] { 0x1f, (byte)0x8b };</code>
		
		@param pin The input stream to read from which must have
			large enough buffer to be able to unread the bytes
		@param bCheck The byte values to use for comparison
		@return TRUE if the file matches the specified bytes, 
			FALSE otherwise
		@throws IOException If there is an error reading or unreading
			the bytes
		@throws NullPointerException If pin or bCheck is null
		@throws IllegalArgumentException If bCheck is zero length
		@since 2.4
	*/
	public static boolean checkMagicNumber(PushbackInputStream pin, byte[] bCheck)
		throws IOException
	{
		if (pin == null) throw new NullPointerException("pin");
		if (bCheck == null) throw new NullPointerException("bCheck");
		if (bCheck.length == 0) throw new IllegalArgumentException("bCheck has zero length");
		byte[] b = new byte[bCheck.length];
		pin.read(b);
		boolean match = Arrays.equals(b, bCheck);
		pin.unread(b);
		return match;
	}
	
	/**
		Get the complete tree of sub-directories from a given directory.
	 	@param dc DbControl used to access the database. 
		@param directory The directory to start with
		@return A <code>Map</code> which maps a directory to a 
			list of it's sub-directories. Directories without sub-directories
			are not included in the map.
		@throws InvalidDataException If the directory is null
		@throws BaseException If there is another error
	*/
	public static Map<Directory, List<Directory>> getDirectoryTree(DbControl dc, Directory directory)
		throws InvalidDataException, BaseException
	{
		ItemQuery<Directory> query = Directory.getQuery();
		query.restrict(
			Restrictions.in(
				Hql.property("parent"),
				Expressions.parameter("parents")
			)
		);
		query.order(Orders.asc(Hql.property("name")));
		query.include(Include.MINE, Include.SHARED, Include.OTHERS, Include.NOT_REMOVED, Include.REMOVED);
		
		Map<Directory, List<Directory>> tree = new HashMap<Directory, List<Directory>>();
		List<Integer> parents = new LinkedList<Integer>();
		parents.add(directory.getId());
		while (parents.size() > 0)
		{
			query.setParameter("parents", parents, Type.INT);
			List<Directory> children = query.list(dc);
			parents.clear();
			for (Directory child : children)
			{
				Directory parent = child.getParent();
				if (!tree.containsKey(parent))
				{
					tree.put(parent, new LinkedList<Directory>());
				}
				tree.get(parent).add(child);
				parents.add(child.getId());
			}
		}
		return tree;
	}
	
	/**
	 	Find files matching a given filter in a directory. If the filter
	 	returns sub-directories, this method will check those directories too.
	 	Use, for example, the {@link RegexpFileFilter} to look for files
	 	matching a regular expression.
	 	
	  	@param directory The parent directory to search in
	  	@param filter A filter which specifies which files to find, or null
	  		to find all files
	  	@return A List with the files in directory
	  		that matches the filter, or null if the specified directory doesn't
	  		exists or isn't a directory
	  	@since 2.5
	  	@see RegexpFileFilter
	*/
	public static List<File> findFiles(File directory, FileFilter filter)
	{	
		if (!directory.exists() || !directory.isDirectory())
		{
			return null;
		}
		
		// To hold the result
		List<File> searchResult = new LinkedList<File>();
		
		// Temporary holds the directories too look in
		List<File> directories = new LinkedList<File>();
		directories.add(directory);
		
		while (directories.size() > 0)
		{
			ThreadSignalHandler.checkInterrupted();
			File dir = directories.remove(0);
			File[] files = dir.listFiles(filter);
			for (File f : files)
			{
				ThreadSignalHandler.checkInterrupted();
				if (f.isDirectory())
				{
					directories.add(f);
				}
				else
				{
					searchResult.add(f);
				}
			}
		}
		return searchResult;
	}
}
