/**
	$Id: Node.java 4479 2008-09-05 15:16:27Z jari $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.util.overview;

import java.util.LinkedList;
import java.util.List;

import net.sf.basedb.core.BasicItem;

/**
 	A node in an overview. There are two types of nodes: folder nodes
 	and item nodes. The only difference is that an item node has a link
 	to a {@link BasicItem} (see {@link #getItem()} but a folder node does not.
 	Both node types can have child nodes.
 	<p>
 	Each node has a unique ID ({@link #getId()}) which is based on the system 
 	hashcode. Each node also has a name which should be unique among the
 	child nodes. If the name isn't unique the {@link #getChild(String)} returns the
 	first matching node.
 	<p>
 	

	@author Nicklas
	@version 2.2
	@base.modified $Date: 2008-09-05 17:16:27 +0200 (Fri, 05 Sep 2008) $
*/
public class Node
{
	private String name;
	private String title;
	private BasicItem item;
	private Node parent;
	private Type type;
	private List<Node> children;
	private int numWarnings;
	private int numErrors;
	private int childWarnings;
	private int childErrors;
	
	/**
		Create an item-type root node.

		@param name The name of the node
		@param title The display title of the node
		@param item The item the node references
		@see ExperimentOverview#getRootNode()
	*/
	public Node(String name, String title, BasicItem item)
	{
		this(name, title, null, item, Type.ITEM);
	}
	
	/**
		Create a folder-type child node.
		@param name The name of the child node
		@param title The display title of the node
		@param parent The parent node
	*/
	public Node(String name, String title, Node parent)
	{
		this(name, title, parent, null, Type.FOLDER);
	}
	
	/**
		Create an item-type child node
		@param name The name of the child node
		@param title The display title of the node
		@param parent The parent node
		@param item The item this node references
	 */
	public Node(String name, String title, Node parent, BasicItem item)
	{
		this(name, title, parent, item, Type.ITEM);
	}
	
	private Node(String name, String title, Node parent, BasicItem item, Type type)
	{
		this.name = name;
		this.title = title;
		this.parent = parent;
		this.item = item;
		this.type = type;
		if (parent != null)
		{
			parent.addChild(this);
		}
	}

	/**
		Get a unique ID for this node. The ID is based on the hashcode and should be unique
		for a given virtual machine.
	*/
	public String getId()
	{
		return "N" + hashCode();
	}
	
	/**
		Get the node type of this node. Note that for folder-type nodes
		the {@link #getItem()} always return null. It it returns null for 
		an item-type node it usually means that access was denied to that
		particular item. 
	*/
	public Type getNodeType()
	{
		return type;
	}
	
	/**
		Get the name of this node. If the name is unique among the child
		nodes for a parent the call <code>getParent().getChild(this.getName())</code>
		should return this node.
	*/
	public String getName()
	{
		return name;
	}
	
	/**
		Get the display title of this node.
	*/
	public String getTitle()
	{
		return title;
	}
	
	/**
		Get the item that is related to this node. Folder-type nodes always
		return null. If an item-type node returns null it usually means that
		access was denied to the item.
	*/
	public BasicItem getItem()
	{
		return item;
	}
	
	/**
		Get the parent of this node. Returns null if this is the root node.
	*/
	public Node getParent()
	{
		return parent;
	}
	
	/**
		Get all child nodes of this node. May return null or an empty
		list if this node has no children.
	*/
	public List<Node> getChildren()
	{
		return children;
	}
	
	/**
		Get the child node with a specific name. If more than one
		node has the same name only the first one found is returned.
		@param name The name of the child node.
	*/
	public Node getChild(String name)
	{
		if (children != null)
		{
			for (Node child : children)
			{
				if (name.equals(child.getName()))
				{
					return child;
				}
			}
		}
		return null;
	}
	
	/**
		Add a child node to this one.
	*/
	protected void addChild(Node child)
	{
		if (children == null) children = new LinkedList<Node>();
		children.add(child);
	}
	
	/**
		Set the number of warnings that was found while validating this
		node. This method automatically updates the number of child
		warnings reported by parent nodes.
		@param numWarnings The number of warnings
		@see #addWarnings(int)
	*/
	public void setWarnings(int numWarnings)
	{
		if (parent != null) parent.addChildWarnings(numWarnings-this.numWarnings);
		this.numWarnings = numWarnings;
	}
	
	/**
		Add more warnings to this node. This method automatically updates the 
		number of child warnings reported by parent nodes.
		@param numWarnings The numer of warnings to add
		@see #setWarnings(int)
	*/
	public void addWarnings(int numWarnings)
	{
		this.numWarnings += numWarnings;
		if (parent != null) parent.addChildWarnings(numWarnings);
	}
	
	/**
		Add child warnings to this node. This method is called by child 
		nodes when the warnings are updated on them.
	*/
	protected void addChildWarnings(int numWarnings)
	{
		this.childWarnings += numWarnings;
		if (parent != null) parent.addChildWarnings(numWarnings);
	}
	
	/**
		Set the number of errors that was found while validating this
		node. This method automatically updates the number of child
		errors reported by parent nodes.
		@param numErrors The number of errors
		@see #addErrors(int)
	*/
	public void setErrors(int numErrors)
	{
		if (parent != null) parent.addChildErrors(numErrors-this.numErrors);
		this.numErrors = numErrors;
	}
	
	/**
		Add more errors to this node. This method automatically updates the 
		number of child errors reported by parent nodes.
		@param numErrors The numer of errors to add
		@see #setErrors(int)
	*/
	public void addErrors(int numErrors)
	{
		this.numErrors += numErrors;
		if (parent != null) parent.addChildErrors(numErrors);
	}
	
	/**
		Add child errors to this node. This method is called by child 
		nodes when the errors are updated on them.
	*/
	protected void addChildErrors(int numErrors)
	{
		this.childErrors += numErrors;
		if (parent != null) parent.addChildErrors(numErrors);
	}
	
	/**
		Get the number of warnings on this node.
		@see #getChildWarnings()
	*/
	public int getNumWarnings()
	{
		return numWarnings;
	}
	
	/**
		Get the number of errors on this node.
		@see #getChildErrors()
	*/
	public int getNumErrors()
	{
		return numErrors;
	}
	
	/**
		Get the number of warnings to children of this node.
		@see #getNumWarnings()
	*/
	public int getChildWarnings()
	{
		return childWarnings;
	}
	
	/**
		Get the number of errors to children of this node.
		@see #getNumErrors()
	*/
	public int getChildErrors()
	{
		return childErrors;
	}
	
	/**
		Clear all warnings, errors on this node and on all child nodes.
	*/
	public void clearFailures()
	{
		numErrors = 0;
		numWarnings = 0;
		childWarnings = 0;
		childErrors = 0;
		if (getChildren() != null)
		{
			for (Node child : getChildren())
			{
				child.clearFailures();
			}
		}
	}

	/**
		The type of node.
	*/
	public enum Type
	{
		FOLDER, ITEM;
	}

}
