/**
$Id: PackedFileExporter.java 4480 2008-09-05 15:16:58Z jari $

Copyright (C) 2007 Nicklas Nordborg, Martin Svensson

This file is part of BASE - BioArray Software Environment.
Available at http://base.thep.lu.se/

BASE is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

BASE is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.
*/

package net.sf.basedb.plugins;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Location;
import net.sf.basedb.core.Nameable;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.Job.ExecutionTime;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AbstractExporterPlugin;
import net.sf.basedb.core.plugin.ExportOutputStream;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.zip.FilePacker;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
	Plug-in that packs one or more selected files and directories into a single file. 
	The actual packing is performed by classes implementing the {@link FilePacker} 
	interface. To add new formats, create a class that implements the interface and,
	register the new class with a {@link net.sf.basedb.core.PluginConfiguration}
	item.
	<p>
	The compressed file can either be downloaded immediately to the local computer or
	stored into BASE's file system.
	
	@author Martin, Nicklas
	@version 2.4
	@base.modified $Date: 2007-07-11 15:45:01 +0200 (Mon, 02 Jul 2007) $
*/
public class PackedFileExporter 
	extends AbstractExporterPlugin 
	implements InteractivePlugin, SignalTarget
{

	private static final About about = new AboutImpl
		(
				"Packed file exporter",
				"This plugin is used to pack one or more files/directories " +
				"into a single file. Items should be selected before starting the plug-in and " +
				"they must all be located in the same parent directory. " +
				"The compressed file can be downloaded immediately " +
				"or stored in the BASE file system. The user can choose to have the selected files and " +
				"directories removed after they've been packed.",
				Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
				"2007, Base 2 development team",
				null,
				null,
				"http://base.thep.lu.se"
		);
	
	private static final GuiContext fileGui = new GuiContext(Item.FILE, GuiContext.Type.LIST);
	private static final GuiContext directoryGui = new GuiContext(Item.DIRECTORY, GuiContext.Type.LIST);
	private static final Set<GuiContext> guiContexts =  
		Collections.unmodifiableSet(new HashSet<GuiContext>(Arrays.asList(fileGui, directoryGui)));
	
	private static final Set<Permissions> permissions = new HashSet<Permissions>();
	
	// Parameters
	private RequestInformation configureJob;
	private RequestInformation configurePlugin;
	private PluginParameter<Boolean> removeFilesParameter;
	private PluginParameter<String> packerParameter;
	
	private int numFiles;
	private long totalBytes;
	private List<Nameable> selectedFilesAndDirs;
	
	private ThreadSignalHandler signalHandler;
	
	/**
		Create new instance.
	*/
	public PackedFileExporter()
	{}
	
	/*
	 	From the Plugin interface
	 	--------------------------------
	*/
	public About getAbout()
	{
		return about;
	}
	public boolean supportsConfigurations()
	{
		return true;
	}
	public boolean requiresConfiguration()
	{
		return true;
	}
	/**
	 	Request create and delete access to File:s and Directory:s.
	*/
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.CREATE, Permission.DELETE)));
			permissions.add(new Permissions(Item.DIRECTORY, null, EnumSet.of(Permission.DELETE)));
		}
		return permissions;
	}
	// -------------------------------------------
	
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	/**
		Check that at least one file or directory has been selected.
	*/
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;
		ItemContext file = sc.getCurrentContext(Item.FILE);
		ItemContext directory = sc.getCurrentContext(Item.DIRECTORY);
		if (file.getSelected().isEmpty() && directory.getSelected().isEmpty())
		{
			message = "You must select at least one file or directory.";
		}
		return message;
	}
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigurePluginParameters();
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context, null);
		}
		return requestInformation;
	}
	@SuppressWarnings("unchecked")
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
			{
				RequestInformation ri = getConfigurePluginParameters();
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameters were found in the request", errors);
					return;
				}				
				
				// Verify that class implements FilePacker
				String packerClass = (String)request.getParameterValue("packer");
				FilePacker packer = null;
				try
				{
					packer = getPacker(packerClass);
				}
				catch (Throwable t)
				{
					response.setError("Class '" + packerClass + "' can't be used for packing", Arrays.asList(t));
					return;
				}
				storeValue(configuration, request, ri.getParameter("packer"));
				
				// Done
				response.setDone("Configuration complete for packer " + packerClass +
						" (" + packer.getDescription() + ")");
			}
			
			else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context, !request.isAllowedImmediateExecution());
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameters were found in the request", errors);
					return;
				}
	
				FilePacker packer = getPacker();
				String extension = packer.getFileExtension();
				
				// Verify SAVE_AS
				String saveAs = (String)request.getParameterValue(SAVE_AS);
				if (saveAs == null)
				{
					// No filename is given, check if immediate execution is allowed
					if (!request.isAllowedImmediateExecution())
					{
						response.setError("Immediate download is not allowed. Please specify a filename.", null);
						return;
					}
				}
				else
				{
					if (!saveAs.endsWith(extension)) saveAs += extension;
					Object parameterValue = request.getParameterValue(OVERWRITE); 
					boolean overwrite = parameterValue != null ? (Boolean)parameterValue : false;
					if (!pathCanBeUsed(saveAs, overwrite))
					{
						response.setError("File exists: " + saveAs, null);
						return;
					}
				}
				
				// Verify that all selected files and directories are children to the root directory
				Directory root = (Directory)request.getParameterValue("root");
				DbControl dc = sc.newDbControl();
				try
				{
					for (Nameable item : selectedFilesAndDirs)
					{
						dc.reattachItem((BasicItem)item);
						Directory dir = null;
						if (item.getType() == Item.FILE)
						{
							dir = ((File)item).getDirectory();
						}
						else
						{
							dir = ((Directory)item).getParent();
						}
						if (!root.equals(dir))
						{
							response.setError(item.getType().toString() + " is not located in root directory: " + root.getPath().toString(), null);
							return;
						}
					}
				}
				finally
				{
					if (dc != null) dc.close();
				}
				
				// Store the selected files and directories		
				storeValue(job, request, ri.getParameter("root"));
				ParameterType<Integer> filesAndDirectories = new IntegerParameterType(null, null, null, true, 0, 0, 0, null);
				Set<Integer> selectedFiles = sc.getCurrentContext(Item.FILE).getSelected();
				if (selectedFiles.size() > 0)
				{
					job.setValues("files", filesAndDirectories, new ArrayList<Integer>(selectedFiles));
				}
				Set<Integer> selectedDirectories = sc.getCurrentContext(Item.DIRECTORY).getSelected();
				if (selectedDirectories.size() > 0)
				{
					job.setValues("directories", filesAndDirectories, new ArrayList<Integer>(selectedDirectories));
				}
				
				// Other options
				storeValue(job, request, ri.getParameter("removeItems"));
				if (saveAs != null)
				{
					storeValue(job, request, ri.getParameter(OVERWRITE));
					storeValue(job, (PluginParameter<String>)ri.getParameter(SAVE_AS), saveAs);
					response.setDone("Job configuration is done", ExecutionTime.SHORT);
				}
				else
				{
					response.setDownloadImmediately("Pack selected files and directories", ExecutionTime.SHORTEST, true);
				}
				
			}
		}
		catch (Throwable th)
		{
			response.setError(th.getMessage(), Arrays.asList(th));
		}
	}
	// -------------------------------------------
	
	/*
	 	From the AbstractExporterPlugin class
	 	--------------------------------------------
	*/
		
	private DbControl dc;
	/*
	  	@see net.sf.basedb.core.plugin.AbstractExporterPlugin#begin(net.sf.basedb.core.DbControl)
	*/
	@Override
	protected void begin(DbControl dc)
	{
		this.dc = dc;
	}
	
	@Override
	@SuppressWarnings("unchecked")
	protected void performExport(ExportOutputStream out, ProgressReporter progress) 
		throws IOException
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		
		// Get parameters
		Directory rootDir = (Directory)job.getValue("root");
		rootDir = Directory.getById(dc, rootDir.getId());
		List<Integer> files = (List<Integer>)job.getValues("files");				
		List<Integer> directories = (List<Integer>)job.getValues("directories");
		boolean removeItems = Boolean.TRUE.equals(job.getValue("removeItems"));
		FilePacker packer = getPacker();

		if (progress != null)
		{
			progress.display(0, "Checking selected files and directories...");
		}
		
		// Preload files
		List<Nameable> itemsToPack = new LinkedList<Nameable>();
		loadFilesAndDirectories(dc, itemsToPack, files, directories, true);
		numFiles = itemsToPack.size();
		int currentFile = 0;
		long completedBytes = 0;
		float progressFactor = totalBytes == 0 ? 0 : 100.0f / totalBytes;
		
		// Prepare for compression
		String rootPath = rootDir.getPath().toString();
		if (!rootPath.endsWith("/")) rootPath += "/";
		out.setMimeType(packer.getMimeType());
		out.setFilename(generateDefaultName(packer, rootDir, itemsToPack));
		packer.setOutputStream(out);
		
		// Compress the files and directories
		for (Nameable item : itemsToPack)
		{
			checkInterrupted();
			currentFile++;
			String entryPath = null;
			Date lastModified = null;
			long size = 0;
			InputStream in = null;
			if (item.getType() == Item.FILE)
			{
				File f = (File)item;
				entryPath = f.getPath().toString();
				lastModified = f.getLastUpdate();
				in = f.getDownloadStream(0);
				if (f.getLocation() == Location.PRIMARY) size = f.getSize();
				if (removeItems && f.hasPermission(Permission.DELETE)) f.setRemoved(true);
			}
			else
			{
				Directory d = (Directory)item;
				entryPath = d.getPath().toString();
				if (removeItems && d.hasPermission(Permission.DELETE)) d.setRemoved(true);
			}
			
			// Remove rootPath from entryPath
			entryPath = entryPath.substring(rootPath.length());
			
			// Update progress
			if (progress != null)
			{
				int percent = (int)(completedBytes * progressFactor);
				progress.display(percent, "Compressing '" + entryPath + "'; " + 
					currentFile + " of " + numFiles + " file(s) (" +
					Values.formatBytes(completedBytes) + " of " + Values.formatBytes(totalBytes) + ")");
			}

			// Send to FilePacker
			packer.pack(entryPath, in, size, lastModified == null ? 0 : lastModified.getTime());
			if (in != null) in.close();
			completedBytes += size;
		}
		
		packer.close();
		
		if (progress != null) 
		{
			progress.display(100, "Done. " + numFiles + " file(s) and directories compressed successfully");
		}
	}
	protected void end(boolean success)
	{
		this.dc = null;
	}
	
	protected String getSuccessMessage()
	{
		return numFiles + " file(s) and directories compressed successfully.";
	}
	// -------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------


	private RequestInformation getConfigureJobParameters(GuiContext context, Boolean requireFile)
	{
		if (configureJob == null)
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

				// Get info about current directory and selected files and directories
				FilePacker packer = getPacker();
				ItemContext fileContext = sc.getCurrentContext(Item.FILE);
				ItemContext dirContext = sc.getCurrentContext(Item.DIRECTORY);
				Directory currentDir = null;
				try
				{
					currentDir = Directory.getById(dc, dirContext.getId());
				}
				catch (Throwable t)
				{}
				
				selectedFilesAndDirs = new LinkedList<Nameable>();
				loadFilesAndDirectories(dc, selectedFilesAndDirs, fileContext.getSelected(), dirContext.getSelected(), false);
				
				// Save as and overwrite parameter
				if (requireFile == null)
				{
					PluginDefinition pd = job.getPluginDefinition();
					requireFile = !pd.getAllowImmediateExecution();
				}
				String defaultPath = generateDefaultPath(packer, currentDir, selectedFilesAndDirs);
				
				// Root directory
				parameters.add(new PluginParameter<Directory>(
					"root",
					"Root directory",
					"The root directory which must be the parent of all selected files and directories",
					new ItemParameterType<Directory>(Directory.class, null, true, 1, null))
				);
				
				// Save as and overwrite
				parameters.add(getSaveAsParameter(null, null, defaultPath, requireFile));
				parameters.add(getOverwriteParameter(null, null));

				// Remove parameter
				removeFilesParameter = new PluginParameter<Boolean>
					(
						"removeItems",
						"Remove files/directories",
						"Select if selected files and directories should be flagged for " +
						"deletion or not, after they've been packed",
						new BooleanParameterType(null, false)
					);
				parameters.add(removeFilesParameter);
				
				configureJob = new RequestInformation
					(
						Request.COMMAND_CONFIGURE_JOB,
						"Select compression format",
						"Select the compression format to use and a few other options.",
						parameters
					);
			}
			finally
			{
				if (dc != null)
				{
					dc.close();
				}
			}
		}		
		return configureJob;
	}
	
	private RequestInformation getConfigurePluginParameters()
	{
		if (configurePlugin == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			packerParameter = new PluginParameter<String>(
				"packer", 
				"Packer class", 
				"Enter the name of the class that is responsible for packing the files. It " +
				"must be a class that implements the FilePacker interface.",
				new StringParameterType(255, null, true)
			);
			parameters.add(packerParameter);
			
			configurePlugin = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_PLUGIN,
				"Select packer class",
				"Enter the name of the FilePacker implementation.",
				parameters
			);
		}
		return configurePlugin;
	}
	

	/**
		Load specified files and directories and put them in <code>loadedFiles</code>.
		This method also calculates the total size of all files and stores it in 
		{@link #totalBytes}.

		@param dc The DbControl to use for loading
		@param loadedFiles The files and directories will be added to this collection
		@param files The ID of the files to load
		@param directories The ID of the directories to load
		@param recursive If we should recurse into subdirectories or not
	*/
	private void loadFilesAndDirectories(DbControl dc, Collection<Nameable> loadedFiles, Collection<Integer> files, Collection<Integer> directories, boolean recursive)
	{
		totalBytes = 0;
		// Load selected files
		if (files != null)
		{
			for (Integer fileId : files)
			{
				checkInterrupted();
				try
				{
					File f = File.getById(dc, fileId);
					loadedFiles.add(f);
					if (f.getLocation() == Location.PRIMARY) totalBytes += f.getSize();
				}
				catch (Throwable t)
				{}
			}
		}
		// Load selected directories and files if recursive
		if (directories != null)
		{
			for (Integer dirId : directories)
			{
				checkInterrupted();
				try
				{
					Directory dir = Directory.getById(dc, dirId);
					loadedFiles.add(dir);
					if (recursive) loadFiles(dc, loadedFiles, dir);
				}
				catch (Throwable t)
				{}
			}
		}
	}
	
	/**
		Load files and subdirectories in the given directory.
		@param dc The DbControl to use for loading
		@param loadedFiles The files and directories will be added to this collection
		@param directory The parent directory
	*/
	private void loadFiles(DbControl dc, Collection<Nameable> loadedFiles, Directory directory)
	{
		ItemQuery<File> fileQuery = directory.getFiles();
		fileQuery.include(Include.MINE, Include.OTHERS, Include.SHARED, Include.IN_PROJECT);
		fileQuery.order(Orders.asc(Hql.property("name")));
		
		for (File f : fileQuery.list(dc))
		{
			checkInterrupted();
			loadedFiles.add(f);
			if (f.getLocation() == Location.PRIMARY) totalBytes += f.getSize();
		}
		
		ItemQuery<Directory> dirQuery = directory.getSubDirectories();
		dirQuery.include(Include.MINE, Include.OTHERS, Include.SHARED, Include.IN_PROJECT);
		dirQuery.order(Orders.asc(Hql.property("name")));
		
		for (Directory dir : dirQuery.list(dc))
		{
			checkInterrupted();
			loadedFiles.add(dir);
			loadFiles(dc, loadedFiles, dir);
		}
	}
	
	/**
		Get the FilePacker that has been configured in the current configuration.
	*/
	private FilePacker getPacker()
	{
		String packerClass = (String)configuration.getValue("packer");
		FilePacker packer = null;
		try
		{
			packer = getPacker(packerClass);
		}
		catch (Throwable t)
		{
			throw new BaseException("Class '" + packerClass + "' can't be used for packing", t);
		}
		return packer;
	}
	
	/**
		Create a FilePacker from the given class.
		@param packerClass The class name of a class that implements FilePacker
	*/
	private FilePacker getPacker(String packerClass)
		throws Exception
	{
		return (FilePacker)Class.forName(packerClass).newInstance();
	}
	
	/**
		Generate a default path for the packed file. If the selection contains only one
		file or directory, the name of that item is used, otherwise the name of the root
		directory is used. The file extension from  {@link FilePacker#getFileExtension()} is 
		automatically appeneded.
		
		@param packer The current FilePacker
		@param rootDir The root directory
		@param selectedItems The selected items (not including subdirectories)
	*/
	private String generateDefaultPath(FilePacker packer, Directory rootDir, Collection<Nameable> selectedItems)
	{
		String defaultPath = rootDir == null ? "~/" : rootDir.getPath().toString();
		if (!defaultPath.endsWith("/")) defaultPath += "/";
		defaultPath +=	generateDefaultName(packer, rootDir, selectedItems);
		return defaultPath;
	}

	/**
		Generate a default name without a path for the packed file.
		@see #generateDefaultPath(FilePacker, Directory, Collection)
	*/
	private String generateDefaultName(FilePacker packer, Directory rootDir, Collection<Nameable> selectedItems)
	{
		String defaultName = "";
		if (selectedItems.size() == 1)
		{
			defaultName += selectedItems.iterator().next().getName();
		}
		else if (rootDir != null)
		{
			defaultName += rootDir.getName();
		}
		else
		{
			defaultName += "packed";
		}
		defaultName += "."+packer.getFileExtension();
		return defaultName;
	}
	
}
