/*
$Id: PluginConfigurationExporter.java 4480 2008-09-05 15:16:58Z jari $

Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg, Martin Svensson
Copyright (C) 2007 Nicklas Nordborg, Martin Svensson

This file is part of BASE - BioArray Software Environment.
Available at http://base.thep.lu.se/

BASE is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

BASE is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.
*/

package net.sf.basedb.plugins;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.ParameterInfo;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.Job.ExecutionTime;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AbstractExporterPlugin;
import net.sf.basedb.core.plugin.ExportOutputStream;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.jdom.DocType;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

/**
	This plugin exports parameters of a pluginconfiguration into an XML-file. 
	Value and class for each parameter are stored in the file. How the data is saved in the
	XML-file is set by the DTD-file plugin-configuration-file.dtd
	  
	@author Martin, Nicklas
	@version 2.0
	@base.modified $Date$
*/

public class PluginConfigurationExporter
	extends AbstractExporterPlugin
	implements InteractivePlugin, SignalTarget
{
	
	private static final About about =
		new AboutImpl
	(
			"Plugin configuration exporter",
			"This plugin is used to export selected plugin configurations to an XML-file. The plugin " +
			"configurations can later be imported by the corresponding import plugin.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2006, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
	);
	
	private static final GuiContext configList = new GuiContext(Item.PLUGINCONFIGURATION, GuiContext.Type.LIST);
	private static final Set<GuiContext> guiContexts = new HashSet<GuiContext>(Arrays.asList(configList));
	private static final Set<Permissions> permissions = new HashSet<Permissions>();
	
	private RequestInformation configureJob;
	
	private int numSelected;
	
	private ThreadSignalHandler signalHandler;
	
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}

	public String isInContext(GuiContext context, Object item)
	{
		return null;
	}

	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (Request.COMMAND_CONFIGURE_JOB.equals(command))
		{
			requestInformation = getConfigureJobParameters(null);
		}
		return requestInformation;
	}

	/**
		Store configuration settings for {@link Request#COMMAND_CONFIGURE_JOB}.
	*/
	public void configure(GuiContext context, Request request, Response response)
	{		
		String command = request.getCommand();		
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(!request.isAllowedImmediateExecution());
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameters were found in the request",errors);
					return;
				}

				// Store selected items
				Set<Integer> selectedItems = 
					sc.getCurrentContext(Item.PLUGINCONFIGURATION).getSelected();
				job.setValues("items", new IntegerParameterType(null, null, null, true, 0, 0, 0, null), 
					new ArrayList<Integer>(selectedItems));
				
				if (request.getParameterValue(SAVE_AS) == null)
				{
					if (!request.isAllowedImmediateExecution())
					{
						response.setError("Immediate download is not allowed. Please specify a filename.", null);
						return;
					}
					response.setDownloadImmediately("Export " + selectedItems.size() + " plugin configurations " + 
						"pluginconfigurations.export.xml", ExecutionTime.SHORTEST, true);
				}
				else
				{
					Object parameterValue = request.getParameterValue(OVERWRITE); 
					boolean overwrite = parameterValue != null ? (Boolean)parameterValue : false;
					if (!pathCanBeUsed((String)request.getParameterValue(ri.getParameter(SAVE_AS).getName()), overwrite))
					{
						response.setError("File exists: " + (String)request.getParameterValue(ri.getParameter(SAVE_AS).getName()), null);
						return;
					}
					storeValue(job, request, ri.getParameter(SAVE_AS));
					storeValue(job, request, ri.getParameter(OVERWRITE));
					response.setDone("The job configuration is complete", ExecutionTime.SHORTEST);
				}
			}
		}
		catch(Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}

	/**
	 	Returns information about this plugin
	*/
	public About getAbout()
	{
		return about;
	}
	
	/**
	 	Always false
	*/
	public boolean supportsConfigurations()
	{
		return false;
	}
	
	/**
	 	Always false
	*/
	public boolean requiresConfiguration()
	{
		return false;
	}

	/**
	 	Request create and write access to File:s and Directory:s and read access to
	 	PluginConfigurations:s, PluginDefinition:s and User:s.
	*/
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.DIRECTORY, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.USER, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.PLUGINCONFIGURATION, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.PLUGINDEFINITION, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}
	// ----------------------------------------------------------------
	
	/*
	 	From the AbstractExporterPlugin class
	 	--------------------------------------------
	*/
	private DbControl dc;
	private int numExported;
	protected void begin(DbControl dc)
	{
		this.dc = dc;
	}
	@SuppressWarnings("unchecked")
	protected void performExport(ExportOutputStream out, ProgressReporter progress)
		throws IOException
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		numExported = 0;
		List<Integer> selectedItems = (List<Integer>)job.getValues("items");
		numSelected = selectedItems != null ? selectedItems.size() : 0;
		if (numSelected > 0)
		{
			out.setMimeType("application/xml");
			out.setFilename("pluginconfigurations.export.xml");
			
			// Get the PluginConfigurations that shall be exported	
			ItemQuery<PluginConfiguration> configQuery = PluginConfiguration.getQuery();
			configQuery.include(Include.MINE, Include.OTHERS, Include.SHARED, Include.IN_PROJECT);
			configQuery.setFirstResult(0);
			configQuery.setMaxResults(0);
			
			numSelected = selectedItems.size();
			configQuery.restrict(
				Restrictions.in(
					Hql.property("id"),
					Expressions.parameter("_selected_")
				)
			);
			configQuery.setParameter("_selected_", selectedItems, Type.INT);
			checkInterrupted();
			ItemResultIterator<PluginConfiguration> configurations = configQuery.iterate(dc);
			
			// Initialize the XML-document			
			XMLOutputter xmlOut = new XMLOutputter(Format.getPrettyFormat().setEncoding("UTF-8"));
			Writer configurationExportStream = new OutputStreamWriter(out, "UTF-8");
			configurationExportStream.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
			xmlOut.output(new DocType("configfile", "plugin-configuration-file.dtd"), configurationExportStream);

			Element configurationRoot = new Element("configfile");
			while (configurations.hasNext())
			{
				checkInterrupted();
				PluginConfiguration pluginConfig = configurations.next();				
				
				Element configElement = new Element("configuration");
				String pluginClassName = pluginConfig.getPluginDefinition().getClassName();
				configElement.setAttribute("pluginClassName", pluginClassName);
				
				Element nameElement = new Element("configname");
				nameElement.setText(pluginConfig.getName());
				configElement.addContent(nameElement);
				
				Element descriptionElement = new Element("description");
				descriptionElement.setText(pluginConfig.getDescription());
				configElement.addContent(descriptionElement);				
				
				// Add each parameter to configElement in the XML
				Set<String>parameters = pluginConfig.getParameterNames();
				for (String parameter : parameters)
				{
					ParameterInfo parameterInfo = pluginConfig.getParameterInfo(parameter);
					Element newParameterElement = new Element("parameter");
					Element paramNameElement = new Element("name");
					paramNameElement.setText(parameter);
					newParameterElement.addContent(paramNameElement);
					
					Element labelElement = new Element("label");				
					labelElement.setText(parameterInfo.getLabel());
					newParameterElement.addContent(labelElement);
					Element parameterDescription = new Element("description");
					parameterDescription.setText(parameterInfo.getDescription());
					newParameterElement.addContent(parameterDescription);
					List<?> values = parameterInfo.getValues();
					Element classElement = new Element("class");				
					
					// Add the parameter values to the parameter element					
					if (values.isEmpty())
					{	
						newParameterElement.addContent(classElement);
						newParameterElement.addContent(new Element("value"));					
					}
					else
					{
						classElement.setText(values.get(0).getClass().getName().toString());
						newParameterElement.addContent(classElement);
						for (Object o : values)
						{											
							Element valueElement = new Element("value");
							valueElement.setText(o.toString());						
							newParameterElement.addContent(valueElement);

						}
					}				
					configElement.addContent(newParameterElement);
				}	
				configurationRoot.addContent(configElement);
				numExported++;
			}
						
			//Finish the XML document, close the stream and commit to DB
			xmlOut.output(configurationRoot, configurationExportStream);
			configurationExportStream.flush();
		}
	}
	protected void end(boolean success)
	{
		this.dc = null;
	}
	protected String getSuccessMessage()
	{
		String msg = null;
		if (numSelected == 0)
		{
			msg = "There were no selected configurations to export";
		}
		else
		{
			msg = numExported + " of " + numSelected + " configurations were successfully exported";
		}
		return msg;
	}
	// ----------------------------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	
	private RequestInformation getConfigureJobParameters(Boolean requireFile)
	{
		if (configureJob == null)
		{			
			Set<Integer> itemIds = sc.getCurrentContext(Item.PLUGINCONFIGURATION).getSelected();
			if (itemIds.size() == 0)
			{
				throw new InvalidDataException("Please select at least one configuration to export");
			}
			
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			if (requireFile == null)
			{
				PluginDefinition pd = job.getPluginDefinition();
				requireFile = pd == null ? false : !pd.getAllowImmediateExecution();
			}

			String defaultPath = requireFile ? "~/pluginconfigurations.export.xml" : null;
			parameters.add(getSaveAsParameter(null, null, defaultPath, requireFile));
			parameters.add(getOverwriteParameter(null, null));
			
			StringBuilder jobText = new StringBuilder();
			jobText.append("Set the XML-file to save the selected configurations in and if existing file should be overwritten.\n");
			jobText.append(itemIds.size() + " selected configuration(s).");
			
			configureJob = new RequestInformation
			(
					Request.COMMAND_CONFIGURE_JOB,
					"Plugin configuration exporter options",
					jobText.toString(),
					parameters
			);
		}
		return configureJob;
	}
	
}
