/**
	$Id: ArraySlideImporter.java 4480 2008-09-05 15:16:58Z jari $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.LabeledExtract;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing array slide items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2008-09-05 17:16:58 +0200 (Fri, 05 Sep 2008) $
*/
public class ArraySlideImporter
	extends AbstractItemImporter<ArraySlide>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.ARRAYSLIDE, GuiContext.Type.LIST));

	private static final About about =
		new AboutImpl
	(
			"Array slide importer",
			"Imports and updates array slides in a batch.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2008, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
	);

	protected static final PluginParameter<String> arrayBatchColumnMapping = new PluginParameter<String>(
			"arrayBatchColumnMapping",
			"Array batch",
			"Mapping that picks the name or ID of the array batch from the data columns. " +
			"The plug-in will first try to find a batch with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Array batch\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> barcodeColumnMapping = new PluginParameter<String>(
			"barcodeColumnMapping",
			"Barcode",
			"Mapping that picks the barcode of the slide from the data columns. This " +
			"parameter is required if the 'Barcode' identification method is selected, " +
			"otherwise it is optional. Example: \\Barcode\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> destroyedColumnMapping = 
		new PluginParameter<String>(
				"destroyedColumnMapping",
				"Destroyed",
				"Mapping that picks the data column specifying if a slide is destroyed or not. " +
				"Recognised values are empty, 0, 1, true, false, yes and no. Other values will be " +
				"treated as destroyed. Example: \\Destroyed\\",
				optionalColumnMapping
		);

	protected static final PluginParameter<String> indexColumnMapping = 
		new PluginParameter<String>(
				"indexColumnMapping",
				"Batch index",
				"Mapping that picks the (1-based) index of the slide among the slides " +
				"created in the same batch. The value a positive integer  and must be in " +
				"the numeric format specified by the 'Number format' parameter." +
				"Example: \\Batch index\\",
				optionalColumnMapping
		);

	public ArraySlideImporter()
	{}

	/*
		From the Plugin interface
		------------------------------------
	*/
	@Override
	public About getAbout()
	{
		return about;
	}
	// ------------------------------------

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper arrayBatchMapper;
	private Mapper indexMapper;
	private Mapper barcodeMapper;
	private Mapper destroyedMapper;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link PropertyIdMethod#BARCODE}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME,  
				PropertyIdMethod.BARCODE,
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	/**
		Adds column mappings for name, description,
		array batch, batch index, barcode and destroyed
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(arrayBatchColumnMapping);
		parameters.add(indexColumnMapping);
		parameters.add(barcodeColumnMapping);
		parameters.add(destroyedColumnMapping);
	}
	
	/**
		Calls {@link ArraySlide#getQuery()}.
	*/
	@Override
	protected ItemQuery<ArraySlide> createItemQuery()
	{
		return ArraySlide.getQuery();
	}
	
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? LabeledExtract.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? LabeledExtract.MAX_DESCRIPTION_LENGTH : null, null);
		arrayBatchMapper = getMapper(ffp, (String)job.getValue("arrayBatchColumnMapping"), null, null);
		indexMapper = getMapper(ffp, (String)job.getValue("indexColumnMapping"), null, null);
		barcodeMapper = getMapper(ffp, (String)job.getValue("barcodeColumnMapping"), 
				cropStrings ? ArraySlide.MAX_BARCODE_LENGTH : null, null);
		destroyedMapper = getMapper(ffp, (String)job.getValue("destroyedColumnMapping"), null, null);
	}
	
	@Override
	protected ArraySlide createItem(DbControl dc, FlatFileParser.Data data)
	{
		ArrayBatch batch = null;
		if (arrayBatchMapper != null) 
		{
			batch = findArrayBatch(dc, FallbackIdMethod.NAME_OR_ID, arrayBatchMapper.getValue(data));
		}

		ArraySlide slide = ArraySlide.getNew(dc, batch);
		updateItem(dc, slide, data);
		return slide;
	}
	
	@Override
	protected void updateItem(DbControl dc, ArraySlide slide, FlatFileParser.Data data)
	{
		if (nameMapper != null) slide.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) slide.setDescription(descriptionMapper.getValue(data));
		if (barcodeMapper != null) slide.setBarcode(barcodeMapper.getValue(data));
		if (indexMapper != null) 
		{
			Integer index = indexMapper.getInt(data);
			if (index != null) slide.setBatchIndex(index);
		}
		if (destroyedMapper != null)
		{
			slide.setDestroyed(Values.getBoolean(destroyedMapper.getValue(data)));
		}
	}
	// ------------------------------------
}
