import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginConfigurationRequest;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.PluginResponse;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.plugin.Response;

/*
 * base2
 * Copyright (C) 2006 Johan Enell, Nicklas Nordborg
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

public class TestMedianRatioNormalization
{
	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing MedianRatioNormalization");
		int experimentId = TestAnalyzePluginUtil.test_createExperiment("data/test.rawdata.import.txt", 2, 100);
		int formulaId = TestFormula.test_create(null, Formula.Type.INTENSITY_EXPRESSION,
				"genepix", new String[] { "raw('ch1FgMedian')", "raw('ch2FgMedian')" }, false);
		TestAnalyzePluginUtil.test_createRootBioAssaySet(experimentId, formulaId);
		int pluginDefinitionId = TestPluginDefinition.test_get("net.sf.basedb.plugins.MedianRatioNormalization");
		int pluginConfigurationId = test_create_configuration(pluginDefinitionId);
		int jobId = test_create_job(pluginConfigurationId, experimentId, 100, 5, 5, 1);
		TestJob.test_execute(jobId);
		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		
		// Cleanup
		TestExperiment.test_delete(experimentId);
		TestJob.test_delete(jobId);
		TestPluginConfiguration.test_delete(pluginConfigurationId);
		TestFormula.test_delete(formulaId);
		return ok;
	}
	
	static int test_create_configuration(int pluginDefinitionId)
	{
		if (pluginDefinitionId == 0 || !TestUtil.hasPermission(Permission.CREATE, Item.PLUGINCONFIGURATION)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();

			PluginDefinition pd = PluginDefinition.getById(dc, pluginDefinitionId);
			PluginConfiguration pc = pd.newPluginConfiguration();
			pc.setName(pd.getName());

			PluginConfigurationRequest request = pc.configure();
			
			PluginResponse response = request.invoke();
			if (response.getStatus() == Response.Status.DONE)
			{
				response.saveParameters(dc);
				dc.saveItem(pc);
				dc.commit();
			}
			else
			{
				throw new BaseException(response.getMessage(), response.getErrorList().get(0));
			}
			id = pc.getId();
			write("--Create configuration for Median Ratio Normalization OK");
		}
		catch (Throwable ex)
		{
			write("--Create configuration for Median Ratio Normalization FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static int test_create_job(int pluginConfigurationId, int experimentId, float minIntensity, float lowExclude, float highExclude, int blockGroup)
	{
		if (pluginConfigurationId == 0 || !TestUtil.hasPermission(Permission.CREATE, Item.JOB)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();

			Experiment e = Experiment.getById(dc, experimentId);
			BioAssaySet bioAssaySet = e.getBioAssaySets().iterate(dc).next();
			
			write("---"+bioAssaySet);
			
			PluginConfiguration pc = PluginConfiguration.getById(dc, pluginConfigurationId);
			Job j = pc.newJob(e);
			j.setName(pc.getName());
			
			PluginConfigurationRequest request = j.configure(null);
			request.setParameterValue("bioAssaySet", bioAssaySet);
			request.setParameterValue("minIntensity", minIntensity);
			request.setParameterValue("lowExclude", lowExclude);
			request.setParameterValue("highExclude", highExclude);
			request.setParameterValue("blockGroup", blockGroup);

			write_request_information(request.getRequestInformation());
			
			PluginResponse response = request.invoke();
			if (response.getStatus() == Response.Status.DONE)
			{
				response.saveParameters(dc);
				dc.saveItem(j);
				dc.commit();
			}
			else
			{
				throw new BaseException(response.getMessage(), response.getErrorList().get(0));
			}
			id = j.getId();
			write("--Create job for Median Ratio Normalization OK");
		}
		catch (Throwable ex)
		{
			write("--Create job for Median Ratio Normalization FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void write_request_information(RequestInformation ri)
	{
		if (TestUtil.getSilent()) return;

		write("Request information for command: "+ri.getCommand());
		write("\tTitle:\t"+ri.getTitle());
		write("\tDescription:\t"+ri.getDescription());
		for (PluginParameter<?> pp : ri.getParameters())
		{
			ParameterType pt = pp.getParameterType();
			System.out.println("\tParameter:\t"+pp.getName() + "\t" + pp.getLabel() + "\t" + pt );
		}
	}
}
