/**
	$Id: RawBioAssayService.java 4480 2008-09-05 15:16:58Z jari $

	Copyright (C) 2007 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.ws.server;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.FileSet;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.info.AnnotationInfo;
import net.sf.basedb.info.ArrayDesignInfo;
import net.sf.basedb.info.DataFileTypeInfo;
import net.sf.basedb.info.QueryOptions;
import net.sf.basedb.info.RawBioAssayInfo;

import org.apache.axiom.om.OMElement;

/**
 	Web service class for 
 	{@link net.sf.basedb.core.RawBioAssay}
  	@author Martin
 	@version 2.5
 */
public class RawBioAssayService 
	extends AbstractRPCService
{
	/**
	 	
	 */
	public RawBioAssayService()
	{}
	
	/**
		Get information about a raw bioassay that you know the ID of.
		@param ID Id to identify the session.
		@param itemId The ID of the raw bioassay
		@return The raw bioassay information
	*/
	public RawBioAssayInfo getById(String ID, int itemId)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();	
		try
		{
			RawBioAssay rba = RawBioAssay.getById(dc, itemId);
			return rba.toTransferable(new RawBioAssayInfo());
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	/**
	 	Gets an array with info about raw bioassays
	 	available for the active session.
	  	@param ID Id for the session to use. 
	  	@param qOpt QueryOptions to put on the query
	  	@return An RawBioAssayInfo array.
	 */
	public RawBioAssayInfo[] getRawBioAssays(String ID, QueryOptions qOpt)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		try
		{	
			ItemQuery<RawBioAssay> query= RawBioAssay.getQuery();		
			query = util.getConfiguredItemQuery(query, qOpt);
			return util.listToInfo(query.list(dc), RawBioAssayInfo.class);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	/**
	 	Gets information about the array design used in a rawbioassay.
	  	@param ID To identify the active session with
	  	@param rawBioAssayId ArrayDesign from RawBioAssay with this id. 
	  	@return ArrayDesignInfo object containing information about the used arraydesign.
	 */
	public ArrayDesignInfo getArrayDesign(String ID, int rawBioAssayId)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		ArrayDesignInfo info = null;
		try
		{
			RawBioAssay rawBioAssay = RawBioAssay.getById(dc, rawBioAssayId);
			ArrayDesign arrayDesign = rawBioAssay.getArrayDesign();
			if (arrayDesign != null) info = arrayDesign.toTransferable(new ArrayDesignInfo());
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return info;
	}
	
	/**
	 	Returns a rawdata file of a certain type attached to an OMElement.
	  	@param ID Used to identify an active session with.
	  	@param rawBioAssayId Id of the rawbio assay the file belongs to.
	  	@param dataFileType The external ID of the file type.
	  	@return OMElement with the file attached, 
	  		or NULL if no file was found.
	  	@see DataFileType
	*/
	public OMElement downloadRawDataByType(String ID, 
			int rawBioAssayId, String dataFileType)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		OMElement toDownload = null;
		try
		{
			RawBioAssay rawBioAssay = RawBioAssay.getById(dc, rawBioAssayId);
			if (rawBioAssay.hasFileSet())
			{
				FileSet fileSet = rawBioAssay.getFileSet();
				DataFileType type = DataFileType.getByExternalId(dc, dataFileType);
				if (fileSet.hasMember(type))
				{
					FileSetMember member = fileSet.getMember(type);
					BaseFileDataSource source = new BaseFileDataSource(member.getFile());
					toDownload = attachFile(source, "downloadFile");
				}
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}		
		return toDownload;
	}
	
	/**
	 	Gets if the rawdata files can be downloaded or not.
	  	@param ID Used to identify an active session with
	  	@param rawBioAssayId Id of the rawbio assay it concerns
	  	@return TRUE if there are files to download, FALSE otherwise.
	 */
	public Boolean hasDownloadableData(String ID, int rawBioAssayId)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		RawBioAssay rawBioAssay = null;
	
		try
		{
			rawBioAssay = RawBioAssay.getById(dc, rawBioAssayId);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return rawBioAssay != null ? rawBioAssay.hasFileSet() : false;
	}
	
	/**
	 	Returns information objects for each DataFileType represented in the FileSet
	 	of a RawBioAssay.
	  	@param ID Id of the active session.
	  	@param rawBioAssayId Id of the rawbioassay it concerns
	 	@param qOpt Restrictions and includes to put on the query
	  	@return An array with DataFileTypeInfo:s.
	 */
	public DataFileTypeInfo[] getDataFileTypes(String ID, int rawBioAssayId, QueryOptions qOpt)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		try
		{
			RawBioAssay rawBioAssay = RawBioAssay.getById(dc, rawBioAssayId);
			if (rawBioAssay.hasFileSet())
			{			
				ItemQuery<DataFileType> query = rawBioAssay.getFileSet().getMemberTypes();
				query = util.getConfiguredItemQuery(query, qOpt);
				return util.listToInfo(query.list(dc), DataFileTypeInfo.class);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return new DataFileTypeInfo[0];
	}
	
	/**
		Get the annotation values for one or more annotation types.
		@param ID Id of the active session.
		@param rawBioAssayId Id of the rawbioassay it concerns
		@param annotationTypes Array with the ID of the annotation types
		@param findInherited TRUE to look among inherited annotations, FALSE to only look
			at primary annotations
		@return An array with AnnotationInfo object, each element in the array 
			corresponds to the annotation type with the same index in the annotationTypes array
	*/
	public AnnotationInfo[] getAnnotations(String ID, int rawBioAssayId, Integer[] annotationTypes, boolean findInherited)
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		try
		{
			RawBioAssay rawBioAssay = RawBioAssay.getById(dc, rawBioAssayId);
			return util.getAnnotations(dc, rawBioAssay, annotationTypes, findInherited);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
}