/**
	$Id: ReporterService.java 4480 2008-09-05 15:16:58Z jari $

	Copyright (C) 2007 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330,
	Boston, MA  02111-1307, USA.
*/
package net.sf.basedb.ws.server;

import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.plugins.ExportedProperty;
import net.sf.basedb.clients.web.plugins.SimpleExport;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ExtendedProperties;
import net.sf.basedb.core.ExtendedProperty;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Reporter;
import net.sf.basedb.core.ReporterList;
import net.sf.basedb.core.ReporterScoreQuery;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.plugin.ExportOutputStream;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.info.ExtendedPropertyInfo;
import net.sf.basedb.info.QueryOptions;
import net.sf.basedb.util.formatter.MultiFormatter;
import net.sf.basedb.util.formatter.ToStringFormatter;

import java.io.File;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.activation.DataSource;

import org.apache.axiom.om.OMElement;

/**
 	Services for Reporters in Base. 
 	Information about reporters can only be accessed by downloading it in a file first.
 	This service provide two methods to download reporters in a file, {@link #downloadReporters(String, ExtendedPropertyInfo[], QueryOptions)}
 	or {@link #downloadReportersInList(String, int, ExtendedPropertyInfo[], QueryOptions)} 
  	@author Martin
 	@since 2.5
 */
public class ReporterService 
	extends AbstractRPCService
{
	/**
	 	Empty constructor
	 */
	public ReporterService()
	{}
	
	/**
	 	Gets information about each extended property for reporters.
	  	@return an ExtendedPropertyInfo array.
	 */
	public ExtendedPropertyInfo[] getExtendedProperties()
	{
		List<ExtendedProperty> properties = ExtendedProperties.getProperties("ReporterData");
		if (properties != null)
		{
			return util.listToInfo(properties, ExtendedPropertyInfo.class);
		}
		else
		{
			return new ExtendedPropertyInfo[0];
		}
	}
	
	/**
	 	Gets an input stream to exported data from a reporter list.
	  	@param ID Id of an active session.
	  	@param reporterListId Id of the reporter list to download reporters information from.
	  	@param extPropInfos The extended property to include in the download.
	  	@param qOpt Query options to set on the query that select data from the database.
	  	@return java.io.InputStream object to the exported data.
	  	@throws Exception If something goes wrong with the export.
	 */
	public OMElement downloadReportersInList(String ID, int reporterListId, ExtendedPropertyInfo[] extPropInfos, QueryOptions qOpt)
		throws Exception
	{		
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		
		ExportOutputStream eos = null;
		File exportFile = null;		
		OMElement elementWithAttachment = null;
		String elementName = "downloadFile";
		String format = "text";
		Item itemType = Item.REPORTERSCORE;
		
		boolean annotatable = false;
		MultiFormatter formatter = new MultiFormatter(new ToStringFormatter(), true);
		formatter.registerFormatter(Date.class, FormatterFactory.getDateFormatter(sc));
		List<ExportedProperty> exportedProperties = new ArrayList<ExportedProperty>(extPropInfos.length);		
		//Fill the list with properties to export, name, external id and reporter type are always included.
		exportedProperties.add(ExportedProperty.parse(dc, "reporter.name:Name", annotatable, formatter));
		exportedProperties.add(ExportedProperty.parse(dc, "reporter.externalId:External ID", annotatable, formatter));
		exportedProperties.add(ExportedProperty.parse(dc, "reporter.reporterType.name:Reporter type", annotatable, formatter));
		exportedProperties.add(ExportedProperty.parse(dc, "score:Score", annotatable, formatter));
		for (int i=0; i<extPropInfos.length; i++)
		{
			String propertyString = "reporter."+extPropInfos[i].getName() + ":" + extPropInfos[i].getTitle();
			ExportedProperty ep = ExportedProperty.parse(dc, propertyString, annotatable, formatter);
			exportedProperties.add(ep);
		}	
		
		//Get and configure the query that select the reporters in the reporter list.
		ReporterList reporterList = ReporterList.getById(dc, reporterListId);
		ReporterScoreQuery query = reporterList.getReporterScores();
		if (qOpt != null && qOpt.getRestriction() != null)
		{
			query.restrict(Hql.restriction(qOpt.getRestriction(), qOpt.getRestrictionPrefix()));
		}
		
		//Do the export to a temporary file and attach this file to the message that will be returned to the client.
		try
		{
			SimpleExport exporter = new SimpleExport();
			exportFile = File.createTempFile("ReporterListFile", null);
			eos = new ExportOutputStream(new FileOutputStream(exportFile));			
			exporter.exportProperties(dc, eos, format, exportedProperties, itemType, query, null);
			eos.close();
			
			DataSource source = new TempFileDataSource(exportFile, "text/plain");
			elementWithAttachment = attachFile(source, elementName);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return elementWithAttachment; 
	}
	
	/**
	 	Attach a file with exported reporter properties so it can be send through web services.	 	
	  	@param ID The id of an active session. Used to verify the logged in user and it's permissions.
	  	@param extPropInfos A list with the extended properties to export. 
	  		A reporters name, external id and reporter type are always exported.
	  	@param qOpt Options to use on the query that select data.
	  	@return An OMElement to return through web services.
	  	@throws Exception If anything goes wrong with the export.
	 */
	public OMElement downloadReporters(String ID, ExtendedPropertyInfo[] extPropInfos, QueryOptions qOpt)
		throws Exception
	{
		SessionControl sc = getSessionControl(ID);
		DbControl dc = sc.newDbControl();
		
		ExportOutputStream eos = null;
		File exportFile = null;		
		OMElement elementWithAttachment = null;
		String elementName = "downloadFile";
		String format = "text";
		Item itemType = Item.REPORTER;
		
		boolean annotatable = false;
		MultiFormatter formatter = new MultiFormatter(new ToStringFormatter(), true);
		formatter.registerFormatter(Date.class, FormatterFactory.getDateFormatter(sc));
		List<ExportedProperty> exportedProperties = new ArrayList<ExportedProperty>(extPropInfos.length);		
		//Fill the list with properties to export, name, external id and reporter type are always included.
		exportedProperties.add(ExportedProperty.parse(dc, "name:Name", annotatable, formatter));
		exportedProperties.add(ExportedProperty.parse(dc, "externalId:External ID", annotatable, formatter));
		exportedProperties.add(ExportedProperty.parse(dc, "reporterType.name:Reporter type", annotatable, formatter));
		for (int i=0; i<extPropInfos.length; i++)
		{
			String propertyString = extPropInfos[i].getName() + ":" + extPropInfos[i].getTitle();
			ExportedProperty ep = ExportedProperty.parse(dc, propertyString, annotatable, formatter);
			exportedProperties.add(ep);
		}	
		
		Query query = Reporter.getQuery();		
		if (qOpt != null && qOpt.getRestriction() != null)
		{
			query.restrict(Hql.restriction(qOpt.getRestriction(), qOpt.getRestrictionPrefix()));
		}
		
		//Do the export to a temporary file and attach this file to the message that will be returned to the client.
		try
		{
			SimpleExport exporter = new SimpleExport();
			exportFile = File.createTempFile("ReporterFile", null);
			eos = new ExportOutputStream(new FileOutputStream(exportFile));			
			exporter.exportProperties(dc, eos, format, exportedProperties, itemType, query, null);
			eos.close();			
			DataSource source = new TempFileDataSource(exportFile, "text/plain");
			elementWithAttachment = attachFile(source, elementName);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return elementWithAttachment;
	}
}
